{
TGMBaseInfoWindow class

  ES: ventana de informacin o balloon
  EN: information windows or balloon

=========================================================================
History:

ver 0.1.5
  ES:
    - error: la funcin GetConvertedString ahora tiene en cuenta la comilla simple
  EN:
    - buf: GetConvertedString function now control the single quote

ver 0.1:
  ES: primera versin
  EN: first version
=========================================================================
IMPORTANTE PROGRAMADORES: Por favor, si tienes comentarios, mejoras,
  ampliaciones, errores y/o cualquier otro tipo de sugerencia, envame un correo a:
  gmlib@cadetill.com

IMPORTANT PROGRAMMERS: please, if you have comments, improvements, enlargements,
  errors and/or any another type of suggestion, please send me a mail to:
  gmlib@cadetill.com
=========================================================================

Copyright () 2011, by Xavier Martinez (cadetill)

@author Xavier Martinez (cadetill)
@web  http://www.cadetill.com
}
unit GMBaseInfoWindow;

interface

uses
  Classes,
  GMSize;

type
  TBaseInfoWindow = class;

  { ****************************************************************************
    internal class
  **************************************************************************** }
  TIWSize = class(TGMSize)
  private
    FInfoWin: TBaseInfoWindow;
  protected
    procedure SetHeight(const Value: Integer); override;
    procedure SetWidth(const Value: Integer); override;
  public
    constructor Create(InfoWin: TBaseInfoWindow); reintroduce; virtual;
  end;

  { ****************************************************************************
    https://developers.google.com/maps/documentation/javascript/reference?hl=en#InfoWindow
  **************************************************************************** }
  {
    ES: clase para ser usada desde otros componentes
    EN: class to be used from other components
  }
  TBaseInfoWindow = class(TPersistent)
  private
    FOwner: TPersistent;
    FHTMLContent: string;
    FDisableAutoPan: Boolean;
    FMaxWidth: Integer;
    FPixelOffset: TIWSize;
    FCloseOtherBeforeOpen: Boolean;
    FOnMaxWidthChange: TNotifyEvent;
    FOnCloseOtherBeforeOpenChange: TNotifyEvent;
    FOnDisableAutoPanChange: TNotifyEvent;
    FOnPixelOffsetChange: TNotifyEvent;
    FOnHTMLContentChange: TNotifyEvent;
    procedure SetCloseOtherBeforeOpen(const Value: Boolean);
    procedure SetDisableAutoPan(const Value: Boolean);
    procedure SetHTMLContent(const Value: string);
    procedure SetMaxWidth(const Value: Integer);
    procedure SetPixelOffset(const Value: TIWSize);
  public
    constructor Create(aOwner: TPersistent); virtual;
    destructor Destroy; override;

    procedure Assign(Source: TPersistent); override;

    function GetConvertedString: string;

    property OnHTMLContentChange: TNotifyEvent read FOnHTMLContentChange write FOnHTMLContentChange;
    property OnDisableAutoPanChange: TNotifyEvent read FOnDisableAutoPanChange write FOnDisableAutoPanChange;
    property OnMaxWidthChange: TNotifyEvent read FOnMaxWidthChange write FOnMaxWidthChange;
    property OnPixelOffsetChange: TNotifyEvent read FOnPixelOffsetChange write FOnPixelOffsetChange;
    property OnCloseOtherBeforeOpenChange: TNotifyEvent read FOnCloseOtherBeforeOpenChange write FOnCloseOtherBeforeOpenChange;
  published
    property HTMLContent: string read FHTMLContent write SetHTMLContent;
    property DisableAutoPan: Boolean read FDisableAutoPan write SetDisableAutoPan;
    property MaxWidth: Integer read FMaxWidth write SetMaxWidth;  // 0 = no MaxWidth
    property PixelOffset: TIWSize read FPixelOffset write SetPixelOffset; // (0,0) = no pixelOffset
    property CloseOtherBeforeOpen: Boolean read FCloseOtherBeforeOpen write SetCloseOtherBeforeOpen;
  end;

implementation

uses
  SysUtils,
  GMMap;

{ TBaseInfoWindow }

procedure TBaseInfoWindow.Assign(Source: TPersistent);
begin
  if Source is TBaseInfoWindow then
  begin
    HTMLContent := TBaseInfoWindow(Source).HTMLContent;
    DisableAutoPan := TBaseInfoWindow(Source).DisableAutoPan;
    MaxWidth := TBaseInfoWindow(Source).MaxWidth;
    PixelOffset := TBaseInfoWindow(Source).PixelOffset;
    CloseOtherBeforeOpen := TBaseInfoWindow(Source).CloseOtherBeforeOpen;
  end
  else
    inherited Assign(Source);
end;

constructor TBaseInfoWindow.Create(aOwner: TPersistent);
begin
  FHTMLContent := '';
  FDisableAutoPan := False;
  FMaxWidth := 0;
  FPixelOffset := TIWSize.Create(Self);
  FCloseOtherBeforeOpen := True;
  FOwner := aOwner;
end;

destructor TBaseInfoWindow.Destroy;
begin
  if Assigned(FPixelOffset) then FreeAndNil(FPixelOffset);

  inherited;
end;

function TBaseInfoWindow.GetConvertedString: string;
begin
  Result := StringReplace(FHTMLContent, #$D#$A, '', [rfReplaceAll]);
  Result := StringReplace(Result, '''', '~$~&~', [rfReplaceAll]);
end;

procedure TBaseInfoWindow.SetCloseOtherBeforeOpen(const Value: Boolean);
begin
  if FCloseOtherBeforeOpen = Value then Exit;

  FCloseOtherBeforeOpen := Value;

  if Assigned(FOnCloseOtherBeforeOpenChange) then FOnCloseOtherBeforeOpenChange(Self);
end;

procedure TBaseInfoWindow.SetDisableAutoPan(const Value: Boolean);
begin
  if FDisableAutoPan = Value then Exit;

  FDisableAutoPan := Value;

  if Assigned(FOnDisableAutoPanChange) then FOnDisableAutoPanChange(Self);
end;

procedure TBaseInfoWindow.SetHTMLContent(const Value: string);
begin
  if FHTMLContent = Value then Exit;

  FHTMLContent := Value;

  if Assigned(FOnHTMLContentChange) then FOnHTMLContentChange(Self);
end;

procedure TBaseInfoWindow.SetMaxWidth(const Value: Integer);
begin
  if FMaxWidth = Value then Exit;

  FMaxWidth := Value;

  if Assigned(FOnMaxWidthChange) then FOnMaxWidthChange(Self);
end;

procedure TBaseInfoWindow.SetPixelOffset(const Value: TIWSize);
begin
  if FPixelOffset = Value then Exit;

  FPixelOffset := Value;

  if Assigned(FOnPixelOffsetChange) then FOnPixelOffsetChange(Self);
end;

{ TIWSize }

constructor TIWSize.Create(InfoWin: TBaseInfoWindow);
begin
  inherited Create;

  FInfoWin := InfoWin;
end;

procedure TIWSize.SetHeight(const Value: Integer);
begin
  if Height = Value then Exit;

  inherited;

  if Assigned(FInfoWin.OnPixelOffsetChange) then FInfoWin.OnPixelOffsetChange(FInfoWin);
end;

procedure TIWSize.SetWidth(const Value: Integer);
begin
  if Width = Value then Exit;

  inherited;

  if Assigned(FInfoWin.OnPixelOffsetChange) then FInfoWin.OnPixelOffsetChange(FInfoWin);
end;

end.
