{
TGMCircle component

  ES: componente para crear crculos en un mapa de Google Maps mediante el
      componente TGMMap
  EN: component to create circles on Google Map map using the component TGMMap

=========================================================================
MODO DE USO/HOW TO USE

  ES: poner el componente en el formulario, linkarlo a un TGMMap y poner los
      crculos a mostrar
  EN: put the component into a form, link to a TGMMap and put the circled to
      show
=========================================================================
History:

ver 0.1.7
  ES:
    cambio: modificados mtodos Set y ShowElement para que usen el nuevo mtodo
      ChangeProperties heredado de TLinkedComponent
  EN:
    change: modified all Set and ShowElements methods to use the new method
      ChangeProperties inherited from TLinkedComponent

ver 0.1.6
  ES:
    nuevo: TSizeable -> aadido mtodo Assign
    error: TSizeable -> cuando no era circular, el radio creca ms que el mximo
  EN:
    new: TSizeable -> added Assign method
    bug: TSizeable -> when no circular, radius was growing more than the maximum

ver 0.1.5
  ES:
    cambio: la propiedad Radius pasa a ser un entero
    nuevo: aadida propiedad AutoResize
  EN:
    change: Radius property becomes an integer
    new: added AutoResize property

ver 0.1.4
  ES: primera versin
  EN: first version
=========================================================================
IMPORTANTE PROGRAMADORES: Por favor, si tienes comentarios, mejoras,
  ampliaciones, errores y/o cualquier otro tipo de sugerencia, envame un correo a:
  gmlib@cadetill.com

IMPORTANT PROGRAMMERS: please, if you have comments, improvements, enlargements,
  errors and/or any another type of suggestion, please send me a mail to:
  gmlib@cadetill.com
=========================================================================

Copyright () 2012, by Xavier Martinez (cadetill)

@author Xavier Martinez (cadetill)
@web  http://www.cadetill.com
}
unit GMCircle;

interface

uses
  Classes, Graphics, ExtCtrls,
  GMMap, GMLatLng, GMLatLngBounds;

type
  TCircle = class;

  { ****************************************************************************
    internal class
  **************************************************************************** }
  TSizeable = class(TPersistent)
  private
    FMax: Integer;
    FIncrement: Integer;
    FMin: Integer;
    FCircular: Boolean;
    FActive: Boolean;
    FOwner: TCircle;
    FSpeed: Integer;
    FTimer: TTimer;
    procedure SetActive(const Value: Boolean);
    procedure SetCircular(const Value: Boolean);
    procedure SetIncrement(const Value: Integer);
    procedure SetMax(const Value: Integer);
    procedure SetMin(const Value: Integer);
    procedure SetSpeed(const Value: Integer);

    procedure OnTimer(Sender: TObject);
  public
    constructor Create(aOwner: TCircle); virtual;
    destructor Destroy; override;

    procedure Assign(Source: TPersistent); override;
  published
    property Active: Boolean read FActive write SetActive;
    property Min: Integer read FMin write SetMin;
    property Max: Integer read FMax write SetMax;
    property Increment: Integer read FIncrement write SetIncrement;
    property Circular: Boolean read FCircular write SetCircular;
    property Speed: Integer read FSpeed write SetSpeed;
  end;

  { ****************************************************************************
    internal class
  **************************************************************************** }
  TCircle = class(TLinkedComponent)
  private
    FFillColor: TColor;
    FRadius: Integer;
    FStrokeColor: TColor;
    FVisible: Boolean;
    FStrokeWeight: Integer;
    FFillOpacity: Real;
    FClickable: Boolean;
    FCenter: TLatLng;
    FStrokeOpacity: Real;
    FEditable: Boolean;
    FIsUpdating: Boolean;
    FAutoResize: TSizeable;
    procedure SetClickable(const Value: Boolean);
    procedure SetEditable(const Value: Boolean);
    procedure SetFillColor(const Value: TColor);
    procedure SetFillOpacity(const Value: Real);
    procedure SetRadius(const Value: Integer);
    procedure SetStrokeColor(const Value: TColor);
    procedure SetStrokeOpacity(const Value: Real);
    procedure SetStrokeWeight(const Value: Integer);
    procedure SetVisible(const Value: Boolean);
    procedure OnChangeLatLng(Sender: TObject);
  protected
    procedure SetIdxList(const Value: Cardinal); override;
    function ChangeProperties: Boolean; override;
  public
    constructor Create(Collection: TCollection); override;
    destructor Destroy; override;

    procedure Assign(Source: TPersistent); override;

    procedure CenterMapTo;
    function GetBounds: TLatLngBounds;
  published
    property Center: TLatLng read FCenter write FCenter;
    property Clickable: Boolean read FClickable write SetClickable;
    property Editable: Boolean read FEditable write SetEditable;
    property FillColor: TColor read FFillColor write SetFillColor;
    property FillOpacity: Real read FFillOpacity write SetFillOpacity; // 0 to 1
    property Radius: Integer read FRadius write SetRadius;
    property StrokeColor: TColor read FStrokeColor write SetStrokeColor;
    property StrokeOpacity: Real read FStrokeOpacity write SetStrokeOpacity; // 0 to 1
    property StrokeWeight: Integer read FStrokeWeight write SetStrokeWeight; // 1 to 10
    property Visible: Boolean read FVisible write SetVisible;
    property AutoResize: TSizeable read FAutoResize write FAutoResize;
    property InfoWindow;
    property Text;
  end;

  { ****************************************************************************
    internal class
  **************************************************************************** }
  TCircles = class(TLinkedComponents)
  private
    procedure SetItems(I: Integer; const Value: TCircle);
    function GetItems(I: Integer): TCircle;
  protected
    function GetOwner: TPersistent; override;
  public
    function Add: TCircle;
    function Insert(Index: Integer): TCircle;

    property Items[I: Integer]: TCircle read GetItems write SetItems; default;
  end;

  { ****************************************************************************
    https://developers.google.com/maps/documentation/javascript/reference?hl=en#Circle
  **************************************************************************** }
  TGMCircle = class(TGMLinkedComponent)
  private
    FOnStrokeColorChange: TLinkedComponentChange;
    FOnRightClick: TLatLngIdxEvent;
    FOnMouseDown: TLatLngIdxEvent;
    FOnVisibleChange: TLinkedComponentChange;
    FOnMouseMove: TLatLngIdxEvent;
    FOnStrokeWeightChange: TLinkedComponentChange;
    FOnMouseUp: TLatLngIdxEvent;
    FOnFillOpacityChange: TLinkedComponentChange;
    FOnClickableChange: TLinkedComponentChange;
    FOnCenterChange: TLatLngIdxEvent;
    FOnMouseOut: TLatLngIdxEvent;
    FOnStrokeOpacityChange: TLinkedComponentChange;
    FOnEditableChange: TLinkedComponentChange;
    FOnDblClick: TLatLngIdxEvent;
    FOnMouseOver: TLatLngIdxEvent;
    FOnFillColorChange: TLinkedComponentChange;
    FOnRadiusChange: TLinkedComponentChange;
    FOnClick: TLatLngIdxEvent;
  protected
    function GetAPIUrl: string; override;

    function GetItems(I: Integer): TCircle;

    procedure ShowElements; override;
    procedure EventFired(EventType: TEventType; Params: array of const); override;
    function GetCollectionItemClass: TLinkedComponentClass; override;
    function GetCollectionClass: TLinkedComponentsClass; override;
  public
    function Add(Lat: Real = 0; Lng: Real = 0; Radius: Integer = 0): TCircle;

    property Items[I: Integer]: TCircle read GetItems; default;
  published
    property VisualObjects;
    // eventos
    // events
        // both (map and properties)
    property OnCenterChange: TLatLngIdxEvent read FOnCenterChange write FOnCenterChange;
    property OnRadiusChange: TLinkedComponentChange read FOnRadiusChange write FOnRadiusChange;
        // from map
    property OnClick: TLatLngIdxEvent read FOnClick write FOnClick;
    property OnDblClick: TLatLngIdxEvent read FOnDblClick write FOnDblClick;
    property OnMouseDown: TLatLngIdxEvent read FOnMouseDown write FOnMouseDown;
    property OnMouseMove: TLatLngIdxEvent read FOnMouseMove write FOnMouseMove;
    property OnMouseOut: TLatLngIdxEvent read FOnMouseOut write FOnMouseOut;
    property OnMouseOver: TLatLngIdxEvent read FOnMouseOver write FOnMouseOver;
    property OnMouseUp: TLatLngIdxEvent read FOnMouseUp write FOnMouseUp;
    property OnRightClick: TLatLngIdxEvent read FOnRightClick write FOnRightClick;
        // from properties
    property OnClickableChange: TLinkedComponentChange read FOnClickableChange write FOnClickableChange;
    property OnEditableChange: TLinkedComponentChange read FOnEditableChange write FOnEditableChange;
    property OnFillColorChange: TLinkedComponentChange read FOnFillColorChange write FOnFillColorChange;
    property OnFillOpacityChange: TLinkedComponentChange read FOnFillOpacityChange write FOnFillOpacityChange;
    property OnStrokeColorChange: TLinkedComponentChange read FOnStrokeColorChange write FOnStrokeColorChange;
    property OnStrokeOpacityChange: TLinkedComponentChange read FOnStrokeOpacityChange write FOnStrokeOpacityChange;
    property OnStrokeWeightChange: TLinkedComponentChange read FOnStrokeWeightChange write FOnStrokeWeightChange;
    property OnVisibleChange: TLinkedComponentChange read FOnVisibleChange write FOnVisibleChange;
  end;

implementation

uses
  Lang, WebControl, GMConstants,
  SysUtils;

{ TGMCircle }

function TGMCircle.Add(Lat, Lng: Real; Radius: Integer): TCircle;
begin
  Result := TCircle(inherited Add);
  Result.Center.Lat := Lat;
  Result.Center.Lng := Lng;
  Result.Radius := Radius;
end;

procedure TGMCircle.EventFired(EventType: TEventType; Params: array of const);
var
  LL: TLatLng;
  OldEvent: TNotifyEvent;
begin
  inherited;

  if EventType = etInfoWinCloseClick then Exit;

  if EventType = etCircleRadiusChange then
  begin
    if High(Params) <> 2 then
      raise Exception.Create(GetTranslateText('Nmero de parmetros incorrecto', Map.Language));

    if Params[0].VType <> vtExtended then
      raise Exception.Create(GetTranslateText('Tipo de parmetro incorrecto', Map.Language));

    if Params[1].VType <> vtInteger then
      raise Exception.Create(GetTranslateText('Tipo de parmetro incorrecto', Map.Language));

    if Params[1].VInteger > VisualObjects.Count - 1 then
      raise Exception.Create(GetTranslateText('Valor de parmetro incorrecto', Map.Language));

    if Params[2].VType <> vtBoolean then
      raise Exception.Create(GetTranslateText('Tipo de parmetro incorrecto', Map.Language));

    if Items[Params[1].VInteger].FIsUpdating then
    begin
      if not Params[0].VBoolean then
        Items[Params[1].VInteger].FIsUpdating := False;
      Exit;
    end;

    Items[Params[1].VInteger].FIsUpdating := True;
    Items[Params[1].VInteger].Radius := Params[0].VInteger;
    if Assigned(FOnRadiusChange) then FOnRadiusChange(Self, Params[1].VInteger, Items[Params[1].VInteger]);
    Items[Params[1].VInteger].FIsUpdating := False;

    Exit;
  end;

  if High(Params) <> 2 then
    raise Exception.Create(GetTranslateText('Nmero de parmetros incorrecto', Map.Language));

  if (Params[0].VType <> vtExtended) or (Params[1].VType <> vtExtended) then
    raise Exception.Create(GetTranslateText('Tipo de parmetro incorrecto', Map.Language));

  if Params[2].VType <> vtInteger then
    raise Exception.Create(GetTranslateText('Tipo de parmetro incorrecto', Map.Language));

  if Params[2].VInteger > VisualObjects.Count - 1 then
    raise Exception.Create(GetTranslateText('Valor de parmetro incorrecto', Map.Language));

  LL := TLatLng.Create(ControlPrecision(Params[0].VExtended^, Map.Precision),
                       ControlPrecision(Params[1].VExtended^, Map.Precision));
  try
    case EventType of
      etCircleClick: if Assigned(FOnClick) then FOnClick(Self, LL, Params[2].VInteger, Items[Params[2].VInteger]);
      etCircleDblClick: if Assigned(FOnDblClick) then FOnDblClick(Self, LL, Params[2].VInteger, Items[Params[2].VInteger]);
      etCircleMouseDown: if Assigned(FOnMouseDown) then FOnMouseDown(Self, LL, Params[2].VInteger, Items[Params[2].VInteger]);
      etCircleMouseMove: if Assigned(FOnMouseMove) then FOnMouseMove(Self, LL, Params[2].VInteger, Items[Params[2].VInteger]);
      etCircleMouseOut: if Assigned(FOnMouseOut) then FOnMouseOut(Self, LL, Params[2].VInteger, Items[Params[2].VInteger]);
      etCircleMouseOver: if Assigned(FOnMouseOver) then FOnMouseOver(Self, LL, Params[2].VInteger, Items[Params[2].VInteger]);
      etCircleMouseUp: if Assigned(FOnMouseUp) then FOnMouseUp(Self, LL, Params[2].VInteger, Items[Params[2].VInteger]);
      etCircleRightClick: if Assigned(FOnRightClick) then FOnRightClick(Self, LL, Params[2].VInteger, Items[Params[2].VInteger]);
      etCircleCenterChange:
      begin
        if Items[Params[2].VInteger].FIsUpdating then
        begin
          Items[Params[2].VInteger].FIsUpdating := False;
          Exit;
        end;

        OldEvent := Items[Params[2].VInteger].Center.OnChange;
        Items[Params[2].VInteger].Center.OnChange := nil;

        Items[Params[2].VInteger].Center.Assign(LL);
        if Assigned(FOnCenterChange) then FOnCenterChange(Self, LL, Params[2].VInteger, Items[Params[2].VInteger]);

        Items[Params[2].VInteger].Center.OnChange := OldEvent;
      end;
    end;
  finally
    FreeAndNil(LL);
  end;
end;

function TGMCircle.GetAPIUrl: string;
begin
  Result := 'https://developers.google.com/maps/documentation/javascript/reference?hl=en#Circle';
end;

function TGMCircle.GetCollectionClass: TLinkedComponentsClass;
begin
  Result := TCircles;
end;

function TGMCircle.GetCollectionItemClass: TLinkedComponentClass;
begin
  Result := TCircle;
end;

function TGMCircle.GetItems(I: Integer): TCircle;
begin
  Result := TCircle(inherited Items[i]);
end;

procedure TGMCircle.ShowElements;
var
  i: Integer;
begin
  if not ExecuteScript('DeleteObjects', IntToStr(IdxList)) then Exit;

  for i:= 0 to VisualObjects.Count - 1 do
    Items[i].ChangeProperties;
end;

{ TCircles }

function TCircles.Add: TCircle;
begin
  Result := TCircle(inherited Add);
end;

function TCircles.GetItems(I: Integer): TCircle;
begin
  Result := TCircle(inherited Items[I]);
end;

function TCircles.GetOwner: TPersistent;
begin
  Result := TGMCircle(inherited GetOwner);
end;

function TCircles.Insert(Index: Integer): TCircle;
begin
  Result := TCircle(inherited Insert(Index));
end;

procedure TCircles.SetItems(I: Integer; const Value: TCircle);
begin
  inherited SetItem(I, Value);
end;

{ TCircle }

procedure TCircle.Assign(Source: TPersistent);
begin
  inherited;

  if Source is TCircle then
  begin
    Center.Assign(TCircle(Source).Center);
    Clickable := TCircle(Source).Clickable;
    Editable  := TCircle(Source).Editable;
    FillColor := TCircle(Source).FillColor;
    FillOpacity := TCircle(Source).FillOpacity;
    Radius := TCircle(Source).Radius;
    StrokeColor := TCircle(Source).StrokeColor;
    StrokeOpacity := TCircle(Source).StrokeOpacity;
    StrokeWeight := TCircle(Source).StrokeWeight;
    Visible := TCircle(Source).Visible;
  end;
end;

procedure TCircle.CenterMapTo;
begin
  if Assigned(Collection) and (Collection is TCircles) and
     Assigned(TCircles(Collection).FGMLinkedComponent) and
     Assigned(TCircles(Collection).FGMLinkedComponent.Map) then
    TCircles(Collection).FGMLinkedComponent.Map.SetCenter(FCenter.Lat, FCenter.Lng);
end;

function TCircle.ChangeProperties: Boolean;
const
  StrParams = '%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s';
var
  Params: string;
begin
  inherited;

  Result := False;

  if not Assigned(Collection) or not(Collection is TCircles) or
     not Assigned(TCircles(Collection).FGMLinkedComponent) or
     not TGMCircle(TCircles(Collection).FGMLinkedComponent).AutoUpdate or
     not Assigned(TGMCircle(TCircles(Collection).FGMLinkedComponent).Map) or
     (csDesigning in TGMCircle(TCircles(Collection).FGMLinkedComponent).ComponentState) then
    Exit;

  Params := Format(StrParams, [
                    IntToStr(IdxList),
                    LowerCase(BoolToStr(Clickable, True)),
                    LowerCase(BoolToStr(Editable, True)),
                    QuotedStr(TGMCircle(TCircles(Collection).FGMLinkedComponent).Map.TColorToStr(FillColor)),
                    StringReplace(FloatToStr(FillOpacity), ',', '.', [rfReplaceAll]),
                    QuotedStr(TGMCircle(TCircles(Collection).FGMLinkedComponent).Map.TColorToStr(StrokeColor)),
                    StringReplace(FloatToStr(StrokeOpacity), ',', '.', [rfReplaceAll]),
                    IntToStr(StrokeWeight),
                    LowerCase(BoolToStr(Visible, True)),
                    Center.LatToStr(TGMCircle(TCircles(Collection).FGMLinkedComponent).Map.Precision),
                    Center.LngToStr(TGMCircle(TCircles(Collection).FGMLinkedComponent).Map.Precision),
                    StringReplace(FloatToStr(Radius), ',', '.', [rfReplaceAll]),
                    IntToStr(Index),
                    QuotedStr(InfoWindow.GetConvertedString),
                    LowerCase(BoolToStr(InfoWindow.DisableAutoPan, True)),
                    IntToStr(InfoWindow.MaxWidth),
                    IntToStr(InfoWindow.PixelOffset.Height),
                    IntToStr(InfoWindow.PixelOffset.Width),
                    LowerCase(BoolToStr(InfoWindow.CloseOtherBeforeOpen, True))
                  ]);

  Result := TGMCircle(TCircles(Collection).FGMLinkedComponent).ExecuteScript('MakeCircle', Params);
  TGMCircle(TCircles(Collection).FGMLinkedComponent).ErrorControl;
end;

constructor TCircle.Create(Collection: TCollection);
begin
  inherited;

  FAutoResize := TSizeable.Create(Self);
  FCenter := TLatLng.Create;
  FCenter.OnChange := OnChangeLatLng;
  FClickable := True;
  FEditable := False;
  FFillOpacity := 0.5;
  FFillColor := clRed;
  FRadius := 0;
  FStrokeColor := clBlack;
  FStrokeOpacity := 1;
  FStrokeWeight := 2;
  FVisible := True;
  FIsUpdating := False;
end;

destructor TCircle.Destroy;
begin
  if Assigned(FCenter) then FreeAndNil(FCenter);
  if Assigned(FAutoResize) then FreeAndNil(FAutoResize);

  inherited;
end;

function TCircle.GetBounds: TLatLngBounds;
const
  StrParams = '%s,%s';
var
  WC: TWebControl;
  Params: string;
begin
  Result := nil;

  if not Assigned(Collection) or not (Collection is TCircles) or
     not Assigned(TCircles(Collection).FGMLinkedComponent) or
     not Assigned(TCircles(Collection).FGMLinkedComponent.Map) then
    Exit;

  Params := Format(StrParams, [IntToStr(IdxList), IntToStr(Index)]);
  if not TGMCircle(TCircles(Collection).FGMLinkedComponent).ExecuteScript('CircleGetBounds', Params) then
    Exit;

  WC := TWebControl.Create(TCircles(Collection).FGMLinkedComponent.Map.WebBrowser);
  Result := TLatLngBounds.Create;
  try
    Result.NE.Lat := WC.GetFloatField(CircleForm, CircleFormNELat);
    Result.NE.Lng := WC.GetFloatField(CircleForm, CircleFormNELng);
    Result.SW.Lat := WC.GetFloatField(CircleForm, CircleFormSWLat);
    Result.SW.Lng := WC.GetFloatField(CircleForm, CircleFormSWLng);
  finally
    FreeAndNil(WC);
  end;
end;

procedure TCircle.OnChangeLatLng(Sender: TObject);
{var
  Tmp: string;}
begin
  FIsUpdating := True;
  ChangeProperties;
  if Assigned(TGMCircle(TCircles(Collection).FGMLinkedComponent).FOnCenterChange) then
    TGMCircle(TCircles(Collection).FGMLinkedComponent).FOnCenterChange(
                  TGMCircle(TCircles(Collection).FGMLinkedComponent),
                  Center,
                  Index,
                  Self);

{  if Assigned(Collection) and (Collection is TCircles) and
     Assigned(TCircles(Collection).FGMLinkedComponent) and
     Assigned(TCircles(Collection).FGMLinkedComponent.Map) then
  begin
    Tmp := Center.ToUrlValue(TCircles(Collection).FGMLinkedComponent.Map.Precision);
    FIsUpdating := True;
    SetProperty('CircleSetCenter', 'OnCenterChange', Tmp);
  end;}
end;

procedure TCircle.SetClickable(const Value: Boolean);
begin
  if FClickable = Value then Exit;

  FClickable := Value;

  ChangeProperties;
  if Assigned(TGMCircle(TCircles(Collection).FGMLinkedComponent).FOnClickableChange) then
    TGMCircle(TCircles(Collection).FGMLinkedComponent).FOnClickableChange(
                  TGMCircle(TCircles(Collection).FGMLinkedComponent),
                  Index,
                  Self);
end;

procedure TCircle.SetEditable(const Value: Boolean);
begin
  if FEditable = Value then Exit;

  FEditable := Value;

  ChangeProperties;
  if Assigned(TGMCircle(TCircles(Collection).FGMLinkedComponent).FOnEditableChange) then
    TGMCircle(TCircles(Collection).FGMLinkedComponent).FOnEditableChange(
                  TGMCircle(TCircles(Collection).FGMLinkedComponent),
                  Index,
                  Self);
end;

procedure TCircle.SetFillColor(const Value: TColor);
begin
  if FFillColor = Value then Exit;

  FFillColor := Value;

  ChangeProperties;
  if Assigned(TGMCircle(TCircles(Collection).FGMLinkedComponent).FOnFillColorChange) then
    TGMCircle(TCircles(Collection).FGMLinkedComponent).FOnFillColorChange(
                  TGMCircle(TCircles(Collection).FGMLinkedComponent),
                  Index,
                  Self);
end;

procedure TCircle.SetFillOpacity(const Value: Real);
begin
  if FFillOpacity = Value then Exit;

  FFillOpacity := Value;

  ChangeProperties;
  if Assigned(TGMCircle(TCircles(Collection).FGMLinkedComponent).FOnFillOpacityChange) then
    TGMCircle(TCircles(Collection).FGMLinkedComponent).FOnFillOpacityChange(
                  TGMCircle(TCircles(Collection).FGMLinkedComponent),
                  Index,
                  Self);
end;

procedure TCircle.SetIdxList(const Value: Cardinal);
begin
  inherited;

end;

procedure TCircle.SetRadius(const Value: Integer);
begin
  if FRadius = Value then Exit;

  FRadius := Value;

  if FRadius < 0 then FRadius := 0;

  if Assigned(Collection) and (Collection is TCircles) and
     Assigned(TCircles(Collection).FGMLinkedComponent) and
     Assigned(TCircles(Collection).FGMLinkedComponent.Map) then
  begin
    FIsUpdating := True;
    if FAutoResize.Active then
    begin
      if not TGMCircle(TCircles(Collection).FGMLinkedComponent).ExecuteScript('CircleSetRadius', Format('%s,%s,%s', [
                    IntToStr(IdxList),
                    StringReplace(FloatToStr(Radius), ',', '.', [rfReplaceAll]),
                    IntToStr(Index)
                    ])) then
      FIsUpdating := False;
    end
    else
      if not ChangeProperties then
        FIsUpdating := False;
    if Assigned(TGMCircle(TCircles(Collection).FGMLinkedComponent).FOnRadiusChange) then
      TGMCircle(TCircles(Collection).FGMLinkedComponent).FOnRadiusChange(
                    TGMCircle(TCircles(Collection).FGMLinkedComponent),
                    Index,
                    Self);
  end;
end;

procedure TCircle.SetStrokeColor(const Value: TColor);
begin
  if FStrokeColor = Value then Exit;

  FStrokeColor := Value;

  ChangeProperties;
  if Assigned(TGMCircle(TCircles(Collection).FGMLinkedComponent).FOnStrokeColorChange) then
    TGMCircle(TCircles(Collection).FGMLinkedComponent).FOnStrokeColorChange(
                  TGMCircle(TCircles(Collection).FGMLinkedComponent),
                  Index,
                  Self);
end;

procedure TCircle.SetStrokeOpacity(const Value: Real);
begin
  if FStrokeOpacity = Value then Exit;

  FStrokeOpacity := Value;

  ChangeProperties;
  if Assigned(TGMCircle(TCircles(Collection).FGMLinkedComponent).FOnStrokeOpacityChange) then
    TGMCircle(TCircles(Collection).FGMLinkedComponent).FOnStrokeOpacityChange(
                  TGMCircle(TCircles(Collection).FGMLinkedComponent),
                  Index,
                  Self);
end;

procedure TCircle.SetStrokeWeight(const Value: Integer);
begin
  if FStrokeWeight = Value then Exit;

  FStrokeWeight := Value;

  ChangeProperties;
  if Assigned(TGMCircle(TCircles(Collection).FGMLinkedComponent).FOnStrokeWeightChange) then
    TGMCircle(TCircles(Collection).FGMLinkedComponent).FOnStrokeWeightChange(
                  TGMCircle(TCircles(Collection).FGMLinkedComponent),
                  Index,
                  Self);
end;

procedure TCircle.SetVisible(const Value: Boolean);
begin
  if FVisible = Value then Exit;

  FVisible := Value;

  ChangeProperties;
  if Assigned(TGMCircle(TCircles(Collection).FGMLinkedComponent).FOnVisibleChange) then
    TGMCircle(TCircles(Collection).FGMLinkedComponent).FOnVisibleChange(
                  TGMCircle(TCircles(Collection).FGMLinkedComponent),
                  Index,
                  Self);
end;

{ TSizeable }

procedure TSizeable.Assign(Source: TPersistent);
begin
  if Source is TSizeable then
  begin
    Active := TSizeable(Source).Active;
    Min := TSizeable(Source).Min;
    Max := TSizeable(Source).Max;
    Increment := TSizeable(Source).Increment;
    Circular := TSizeable(Source).Circular;
    Speed := TSizeable(Source).Speed;
  end
  else
    inherited Assign(Source);
end;

constructor TSizeable.Create(aOwner: TCircle);
begin
  FOwner := aOwner;
  FActive := False;
  FCircular := True;
  FMin := 0;
  FMax := 0;
  FIncrement := 100;
  FTimer := TTimer.Create(nil);
  FTimer.Enabled := False;
  FTimer.OnTimer := OnTimer;
end;

destructor TSizeable.Destroy;
begin
  if Assigned(FTimer) then FreeAndNil(FTimer);

  inherited;
end;

procedure TSizeable.OnTimer(Sender: TObject);
begin
  if (FOwner.Radius > FMax) then
    if FCircular then FOwner.Radius := FMin
    else Active := False
  else FOwner.Radius := FOwner.Radius + FIncrement;
end;

procedure TSizeable.SetActive(const Value: Boolean);
begin
  if FActive = Value then Exit;

  FActive := Value;

  if FActive then
  begin
    FOwner.Radius := FMin;
    FTimer.Interval := FIncrement;
    FTimer.Enabled := True;
  end
  else
    FTimer.Enabled := False;
end;

procedure TSizeable.SetCircular(const Value: Boolean);
begin
  FCircular := Value;
end;

procedure TSizeable.SetIncrement(const Value: Integer);
begin
  FIncrement := Value;
end;

procedure TSizeable.SetMax(const Value: Integer);
begin
  FMax := Value;
end;

procedure TSizeable.SetMin(const Value: Integer);
begin
  FMin := Value;
end;

procedure TSizeable.SetSpeed(const Value: Integer);
begin
  if FSpeed = Value then Exit;

  FSpeed := Value;

  if FActive then FTimer.Interval := FSpeed;
end;

end.
