{
TGMDirection component

  ES: componente para mostrar la ruta a seguir entre dos puntos en un mapa de Google
      Maps mediante el componente TGMMap
  EN: component to show the route to follow between two points on Google Map map using
      the component TGMMap

=========================================================================
MODO DE USO/HOW TO USE

  ES: poner el componente en el formulario, unirlo a un TGMMap, establecer las
    propiedades de bsqueda y reproduccin y hacer una llamada al mtodo Execute
  EN: put the component on the form, link it to a TGMMap, set the request and
    render properties and call the Execute method
=========================================================================
History:

ver 0.1.6
  ES: primera versin
  EN: first version
=========================================================================
IMPORTANTE PROGRAMADORES: Por favor, si tienes comentarios, mejoras,
  ampliaciones, errores y/o cualquier otro tipo de sugerencia, envame un correo a:
  gmlib@cadetill.com

IMPORTANT PROGRAMMERS: please, if you have comments, improvements, enlargements,
  errors and/or any another type of suggestion, please send me a mail to:
  gmlib@cadetill.com
=========================================================================

Copyright () 2012, by Xavier Martinez (cadetill)

@author Xavier Martinez (cadetill)
@web  http://www.cadetill.com
}
unit GMDirection;

interface

uses
  Classes, Contnrs, Graphics, {$IF CompilerVersion < 19}Controls,{$IFEND}
  GMLatLng, GMConstants, GMMap, GMLatLngBounds;

type
  TDirectionsRequest = class;
  TGMDirection = class;

  { ****************************************************************************
    https://developers.google.com/maps/documentation/javascript/reference?hl=en#Distance
  **************************************************************************** }
  TDistance = class
  private
    FValue: Integer;
    FText: string;
  public
    constructor Create; virtual;

    procedure Assign(Source: TObject); virtual;

    property Text: string read FText;
    property Value: Integer read FValue;
  end;

  { ****************************************************************************
    https://developers.google.com/maps/documentation/javascript/reference?hl=en#Duration
  **************************************************************************** }
  TDuration = class
    FValue: Integer;
    FText: string;
  public
    constructor Create; virtual;

    procedure Assign(Source: TObject); virtual;

    property Text: string read FText;
    property Value: Integer read FValue;
  end;

  { ****************************************************************************
    https://developers.google.com/maps/documentation/javascript/reference?hl=en#TransitAgency
  **************************************************************************** }
  TTransitAgency = class
  private
    FName: string;
    FPhone: string;
    FUrl: string;
  public
    constructor Create; virtual;

    procedure Assign(Source: TObject); virtual;

    property Name: string read FName;
    property Phone: string read FPhone;
    property Url: string read FUrl;
  end;

  { ****************************************************************************
    https://developers.google.com/maps/documentation/javascript/reference?hl=en#TransitStop
  **************************************************************************** }
  TTransitStop = class
  private
    FName: string;
    FLocation: TLatLng;
  public
    constructor Create; virtual;
    destructor Destroy; override;

    procedure Assign(Source: TObject); virtual;

    property Location: TLatLng read FLocation;
    property Name: string read FName;
  end;

  { ****************************************************************************
    https://developers.google.com/maps/documentation/javascript/reference?hl=en#Time
  **************************************************************************** }
  TTimeClass = class
  private
    FValue: TTime;
    FTimeZone: string;
    FText: string;
  public
    constructor Create; virtual;

    procedure Assign(Source: TObject); virtual;

    property Text: string read FText;
    property TimeZone: string read FTimeZone;
    property Value: TTime read FValue;
  end;

  { ****************************************************************************
    https://developers.google.com/maps/documentation/javascript/reference?hl=en#TransitVehicle
  **************************************************************************** }
  TTransitVehicle = class
  private
    FName: string;
    FLocalIcon: string;
    FIcon: string;
    FVehicleType: TVehicleType;
  public
    constructor Create; virtual;

    procedure Assign(Source: TObject); virtual;

    property Icon: string read FIcon;
    property LocalIcon: string read FLocalIcon;
    property Name: string read FName;
    property VehicleType: TVehicleType read FVehicleType;
  end;

  { ****************************************************************************
    https://developers.google.com/maps/documentation/javascript/reference?hl=en#TransitLine
  **************************************************************************** }
  TTransitLine = class
  private
    FAgencie: TObjectList;
    FName: string;
    FTextColor: string;
    FVehicle: TTransitVehicle;
    FColor: string;
    FIcon: string;
    FShortName: string;
    FUrl: string;
    function GetAgencie(Index: Integer): TTransitAgency;
    function GetCountAgencie: Integer;
  public
    constructor Create; virtual;
    destructor Destroy; override;

    procedure Assign(Source: TObject); virtual;

    property Agencie[Index: Integer]: TTransitAgency read GetAgencie;
    property CountAgencie: Integer read GetCountAgencie;
    property Color: string read FColor;
    property Icon: string read FIcon;
    property Name: string read FName;
    property ShortName: string read FShortName;
    property TextColor: string read FTextColor;
    property Url: string read FUrl;
    property Vehicle: TTransitVehicle read FVehicle;
  end;

  { ****************************************************************************
    https://developers.google.com/maps/documentation/javascript/reference?hl=en#TransitDetails
  **************************************************************************** }
  TTransitDetails = class
  private
    FArribalStop: TTransitStop;
    FArribalTime: TTimeClass;
    FDepartureStop: TTransitStop;
    FDepartureTime: TTimeClass;
    FHeadsign: string;
    FNumStops: Integer;
    FLine: TTransitLine;
    FHeadway: Integer;
  public
    constructor Create; virtual;
    destructor Destroy; override;

    procedure Assign(Source: TObject); virtual;

    property ArribalStop: TTransitStop read FArribalStop;
    property ArribalTime: TTimeClass read FArribalTime;
    property DepartureStop: TTransitStop read FDepartureStop;
    property DepartureTime: TTimeClass read FDepartureTime;
    property Headsign: string read FHeadsign;
    property Headway: Integer read FHeadway;
    property Line: TTransitLine read FLine;
    property NumStops: Integer read FNumStops;
  end;

  { ****************************************************************************
    https://developers.google.com/maps/documentation/javascript/reference?hl=en#DirectionsStep
  **************************************************************************** }
  TDirectionsStep = class
  private
    FPath: TObjectList;
    FDistance: TDistance;
    FTravelMode: TTravelMode;
    FDuration: TDuration;
    FStartLocation: TLatLng;
    FInstructions: string;
    FTransit: TTransitDetails;
    FEndLocation: TLatLng;
    FSteps: TDirectionsStep;
    function GetPath(Index: integer): TLatLng;
    function GetCountPath: Integer;
  public
    constructor Create; virtual;
    destructor Destroy; override;

    procedure Assign(Source: TObject); virtual;

    property Distance: TDistance read FDistance;
    property Duration: TDuration read FDuration;
    property EndLocation: TLatLng read FEndLocation;
    property Instructions: string read FInstructions;
    property Path[Index: integer]: TLatLng read GetPath;
    property CountPath: Integer read GetCountPath;
    property StartLocation: TLatLng read FStartLocation;
    property Steps: TDirectionsStep read FSteps;
    property Transit: TTransitDetails read FTransit;
    property TravelMode: TTravelMode read FTravelMode;
  end;

  { ****************************************************************************
    https://developers.google.com/maps/documentation/javascript/reference?hl=en#DirectionsLeg
  **************************************************************************** }
  TDirectionsLeg = class
  private
    FDepartureTime: TDuration;
    FArrivalTime: TDuration;
    FDistance: TDistance;
    FDuration: TDuration;
    FEndAddress: string;
    FStartAddress: string;
    FEndLocation: TLatLng;
    FStartLocation: TLatLng;
    FViaWaypoints: TObjectList;
    FStep: TObjectList;
    function GetViaWaypoints(Index: Integer): TLatLng;
    function GetStep(Index: Integer): TDirectionsStep;
    function GetCountStep: Integer;
    function GetCountViaWaypoints: Integer;
  public
    constructor Create; virtual;
    destructor Destroy; override;

    procedure Assign(Source: TObject); virtual;

    property ArrivalTime: TDuration read FArrivalTime;
    property DepartureTime: TDuration read FDepartureTime;
    property Distance: TDistance read FDistance;
    property Duration: TDuration read FDuration;
    property EndAddress: string read FEndAddress;
    property StartAddress: string read FStartAddress;
    property EndLocation: TLatLng read FEndLocation;
    property StartLocation: TLatLng read FStartLocation;
    property ViaWaypoints[Index: Integer]: TLatLng read GetViaWaypoints;
    property Step[Index: Integer]: TDirectionsStep read GetStep;
    property CountViaWaypoints: Integer read GetCountViaWaypoints;
    property CountStep: Integer read GetCountStep;
  end;

  { ****************************************************************************
    https://developers.google.com/maps/documentation/javascript/reference?hl=en#DirectionsRoute
  **************************************************************************** }
  TDirectionsRoute = class
  private
    FSumary: string;
    FBounds: TLatLngBounds;
    FCopyrights: string;
    FWarning: TStringList;
    FWaypointOrder: TStringList;
    FOverviewPath: TObjectList;
    FLeg: TObjectList;
    function GetLeg(Index: Integer): TDirectionsLeg;
    function GetOverviewPath(Index: Integer): TLatLng;
    function GetWarning(Index: Integer): string;
    function GetWaypointOrder(Index: Integer): Integer;
    function GetCountLeg: Integer;
    function GetCountOverviewPath: Integer;
    function GetCountWarning: Integer;
    function GetCountWaypointOrder: Integer;
  public
    constructor Create; virtual;
    destructor Destroy; override;

    procedure Assign(Source: TObject); virtual;

    property Sumary: string read FSumary;
    property Bounds: TLatLngBounds read FBounds;
    property Copyrights: string read FCopyrights;
    property Warning[Index: Integer]: string read GetWarning;
    property WaypointOrder[Index: Integer]: Integer read GetWaypointOrder;
    property OverviewPath[Index: Integer]: TLatLng read GetOverviewPath;
    property Leg[Index: Integer]: TDirectionsLeg read GetLeg;
    property CountWarning: Integer read GetCountWarning;
    property CountWaypointOrder: Integer read GetCountWaypointOrder;
    property CountOverviewPath: Integer read GetCountOverviewPath;
    property CountLeg: Integer read GetCountLeg;
  end;

  TDirectionsRenderer = class;
  { ****************************************************************************
    https://developers.google.com/maps/documentation/javascript/reference?hl=en#DirectionsResult
  **************************************************************************** }
  TDirectionsResult = class
  private
    FOwner: TGMDirection;
    FRoutes: TObjectList;
    FXMLData: TStringList;
    FIndex: Integer;
    FStatus: TDirectionsStatus;
    FFromTo: string;
    FDirectionsRender: TDirectionsRenderer;
    function GetCount: Integer;
    function GetRoutes(Index: Integer): TDirectionsRoute;
    procedure OnXMLChange(Sender: TObject);
    procedure OnDirectionsRenderChange(Sender: TObject);
  protected
  public
    constructor Create(aOwner: TGMDirection; Index: Integer); virtual;
    destructor Destroy; override;

    procedure Assign(Source: TObject); virtual;

    procedure ShowRoute(RouteIndex: Integer = 0; HiddeOthers: Boolean = True);

    property Index: Integer read FIndex;
    property Routes[Index: Integer]: TDirectionsRoute read GetRoutes;
    property Count: Integer read GetCount;
    property XMLData: TStringList read FXMLData;
    property Status: TDirectionsStatus read FStatus;
    property FromTo: string read FFromTo;
    property DirectionsRender: TDirectionsRenderer read FDirectionsRender write FDirectionsRender;
  end;

  { ****************************************************************************
    internal class
  **************************************************************************** }
  TLatLngStr = class(TPersistent)
  private
    FLatLng: TLatLng;
    FAddress: string;
  public
    constructor Create; virtual;
    destructor Destroy; override;

    procedure Assign(Source: TPersistent); override;
    function ToString: string; {$IF CompilerVersion < 19}virtual;{$ELSE}override;{$IFEND}
  published
    property LatLng: TLatLng read FLatLng write FLatLng;
    property Address: string read FAddress write FAddress;
  end;

  { ****************************************************************************
    https://developers.google.com/maps/documentation/javascript/reference?hl=en#TransitOptions
  **************************************************************************** }
  TTransitOptions = class(TPersistent)
  private
    FArrivalTime: TTime;
    FDepartureTime: TTime;
    FArrivalDate: TDate;
    FDepartureDate: TDate;
    procedure SetArrivalDate(const Value: TDate);
    procedure SetArrivalTime(const Value: TTime);
    procedure SetDepartureDate(const Value: TDate);
    procedure SetDepartureTime(const Value: TTime);
  public
    constructor Create; virtual;

    procedure Assign(Source: TPersistent); override;
  published
    property ArrivalTime: TTime read FArrivalTime write SetArrivalTime;
    property ArrivalDate: TDate read FArrivalDate write SetArrivalDate;
    property DepartureTime: TTime read FDepartureTime write SetDepartureTime;
    property DepartureDate: TDate read FDepartureDate write SetDepartureDate;
  end;

  { ****************************************************************************
    internal class
  **************************************************************************** }
  TWaypoint = class(TCollectionItem)
  private
    FStopOver: Boolean;
    FLocation: TLatLngStr;
    FTitle: string;
    procedure SetTitle(const Value: string);
  protected
    function GetDisplayName: string; override;
  public
    constructor Create(Collection: TCollection); override;
    destructor Destroy; override;

    procedure Assign(Source: TPersistent); override;
  published
    property Location: TLatLngStr read FLocation write FLocation;
    property StopOver: Boolean read FStopOver write FStopOver;
    property Title: string read FTitle write SetTitle;
  end;

  { ****************************************************************************
    internal class
  **************************************************************************** }
  TWaypointsList = class(TCollection)
  private
    FOwner: TPersistent;

    function GetItems(I: Integer): TWaypoint;
    procedure SetItems(I: Integer; const Value: TWaypoint);
  protected
    function GetOwner: TPersistent; override;
  public
    constructor Create(AOwner: TPersistent); virtual;

    function Add: TWaypoint;
    function Insert(Index: Integer): TWaypoint;
    procedure Delete(Index: Integer);
    procedure Move(CurIndex, NewIndex: Integer);
    procedure Clear;

    property Items[I: Integer]: TWaypoint read GetItems write SetItems; default;
  end;

  { ****************************************************************************
    https://developers.google.com/maps/documentation/javascript/reference?hl=en#DirectionsRenderer
  **************************************************************************** }
  TDirectionsRequest = class(TPersistent)
  private
    FDestination: TLatLngStr;
    FWaypointsList: TWaypointsList;
    FOrigin: TLatLngStr;
    FOptimizeWaypoints: Boolean;
    FTravelMode: TTravelMode;
    FTransitOpt: TTransitOptions;
    FUnitSystem: TUnitSystem;
    FAvoidHighways: Boolean;
    FRegion: TRegion;
    FAvoidTolls: Boolean;
    FProvideRouteAlt: Boolean;

    FOwner: TPersistent;

    function GetCount: Integer;
    function GetWaypoint(I: Integer): TWaypoint;
    procedure SetWaypointsList(const Value: TWaypointsList);
  protected
    function GetOwner: TPersistent; override;
  public
    constructor Create(aOwner: TPersistent); virtual;
    destructor Destroy; override;

    procedure Assign(Source: TPersistent); override;

    property Count: Integer read GetCount;
    property Waypoints[I: Integer]: TWaypoint read GetWaypoint; default;
  published
    property AvoidHighways: Boolean read FAvoidHighways write FAvoidHighways;
    property AvoidTolls: Boolean read FAvoidTolls write FAvoidTolls;
    property Destination: TLatLngStr read FDestination write FDestination;
    property OptimizeWaypoints: Boolean read FOptimizeWaypoints write FOptimizeWaypoints;
    property Origin: TLatLngStr read FOrigin write FOrigin;
    property ProvideRouteAlt: Boolean read FProvideRouteAlt write FProvideRouteAlt;
    property Region: TRegion read FRegion write FRegion;
    property TransitOpt: TTransitOptions read FTransitOpt write FTransitOpt;
    property TravelMode: TTravelMode read FTravelMode write FTravelMode;
    property UnitSystem: TUnitSystem read FUnitSystem write FUnitSystem;
    property WaypointsList: TWaypointsList read FWaypointsList write SetWaypointsList;
  end;

  { ****************************************************************************
    https://developers.google.com/maps/documentation/javascript/reference?hl=en#MarkerOptions
  **************************************************************************** }
  TMarkerOptions = class(TPersistent)
  private
    FDraggable: Boolean;
    FIcon: string;
    FFlat: Boolean;
    FClickable: Boolean;
    FOnChange: TNotifyEvent;
    procedure SetClickable(const Value: Boolean);
    procedure SetDraggable(const Value: Boolean);
    procedure SetFlat(const Value: Boolean);
    procedure SetIcon(const Value: string);
  protected
    property OnChange: TNotifyEvent read FOnChange write FOnChange;
  public
    constructor Create(aOwner: TPersistent); virtual;

    procedure Assign(Source: TPersistent); override;
  published
    property Clickable: Boolean read FClickable write SetClickable;
    property Draggable: Boolean read FDraggable write SetDraggable;
    property Flat: Boolean read FFlat write SetFlat;
    property Icon: string read FIcon write SetIcon;
  end;

  { ****************************************************************************
    https://developers.google.com/maps/documentation/javascript/reference?hl=en#PolylineOptions
  **************************************************************************** }
  TPolylineOptions = class(TPersistent)
  private
    FStrokeColor: TColor;
    FStrokeWeight: Integer;
    FGeodesic: Boolean;
    FClickable: Boolean;
    FStrokeOpacity: Real;
    FOnChange: TNotifyEvent;
    procedure SetStrokeOpacity(const Value: Real);
    procedure SetStrokeWeight(const Value: Integer);
    procedure SetClickable(const Value: Boolean);
    procedure SetGeodesic(const Value: Boolean);
    procedure SetStrokeColor(const Value: TColor);
  protected
    property OnChange: TNotifyEvent read FOnChange write FOnChange;
  public
    constructor Create(aOwner: TPersistent); virtual;

    procedure Assign(Source: TPersistent); override;
  published
    property Clickable: Boolean read FClickable write SetClickable;
    property Geodesic: Boolean read FGeodesic write SetGeodesic;
    property StrokeColor: TColor read FStrokeColor write SetStrokeColor;
    property StrokeOpacity: Real read FStrokeOpacity write SetStrokeOpacity; // 0 to 1
    property StrokeWeight: Integer read FStrokeWeight write SetStrokeWeight; // 1 to 10
  end;

  { ****************************************************************************
    https://developers.google.com/maps/documentation/javascript/reference?hl=en#DirectionsRendererOptions
  **************************************************************************** }
  TDirectionsRenderer = class(TPersistent)
  private
    FSuppressPolylines: Boolean;
    FSuppressMarkers: Boolean;
    FSuppressBicyclingLayer: Boolean;
    FPolylineOptions: TPolylineOptions;
    FMarkerOptions: TMarkerOptions;
    FDraggable: Boolean;
    FSuppressInfoWindows: Boolean;
    FPreserveViewport: Boolean;
    FOnChange: TNotifyEvent;
    procedure SetDraggable(const Value: Boolean);
    procedure SetPreserveViewport(const Value: Boolean);
    procedure SetSuppressBicyclingLayer(const Value: Boolean);
    procedure SetSuppressInfoWindows(const Value: Boolean);
    procedure SetSuppressMarkers(const Value: Boolean);
    procedure SetSuppressPolylines(const Value: Boolean);
    procedure OnMarkerOptionsChange(Sender: TObject);
    procedure OnPolylineOptionsChange(Sender: TObject);
  protected
    property OnChange: TNotifyEvent read FOnChange write FOnChange;
  public
    constructor Create(aOwner: TObject); virtual;
    destructor Destroy; override;

    procedure Assign(Source: TPersistent); override;
  published
    property Draggable: Boolean read FDraggable write SetDraggable;
    property MarkerOptions: TMarkerOptions read FMarkerOptions write FMarkerOptions;
    property PolylineOptions: TPolylineOptions read FPolylineOptions write FPolylineOptions;
    property PreserveViewport: Boolean read FPreserveViewport write SetPreserveViewport;
    property SuppressBicyclingLayer: Boolean read FSuppressBicyclingLayer write SetSuppressBicyclingLayer;
    property SuppressInfoWindows: Boolean read FSuppressInfoWindows write SetSuppressInfoWindows;
    property SuppressMarkers: Boolean read FSuppressMarkers write SetSuppressMarkers;
    property SuppressPolylines: Boolean read FSuppressPolylines write SetSuppressPolylines;
  end;

  { ****************************************************************************
    https://developers.google.com/maps/documentation/javascript/reference#DirectionsService
    https://developers.google.com/maps/documentation/javascript/directions
    https://developers.google.com/maps/documentation/directions
  **************************************************************************** }
  TGMDirection = class(TGMObjects)
  private
    FDirectionsRender: TDirectionsRenderer;
    FDirectionsRequest: TDirectionsRequest;
    FAutoShow: Boolean;
    FHiddeOthers: Boolean;
    FCountDirResult: Integer;

    FDirectionsResult: TObjectList;
    FOnDirectionsChanged: TNotifyEvent;

    function GetCount: Integer;
    function GetDirectionResult(I: Integer): TDirectionsResult;
  protected
    function GetAPIUrl: string; override;

    procedure DeleteMapObjects; override;
    procedure ShowElements; override;
    procedure EventFired(EventType: TEventType; Params: array of const); override;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;

    procedure Assign(Source: TPersistent); override;

    function AddWaypoint: TWaypoint;
    procedure DeleteWaypoint(Index: Integer);
    procedure MoveWaypoint(CurIndex, NewIndex: Integer);
    procedure ClearWaypoint;

    procedure Execute;
    procedure ShowRoute(ResultIndex, RouteIndex: Integer; HiddeOthers: Boolean = True);
    procedure Delete(ResultIndex: Integer);

    property Count: Integer read GetCount;
    property DirectionsResult[I: Integer]: TDirectionsResult read GetDirectionResult; default;
  published
    property AutoShow: Boolean read FAutoShow write FAutoShow;
    property HiddeOthers: Boolean read FHiddeOthers write FHiddeOthers;
    property DirectionsRender: TDirectionsRenderer read FDirectionsRender write FDirectionsRender;
    property DirectionsRequest: TDirectionsRequest read FDirectionsRequest write FDirectionsRequest;
    property OnDirectionsChanged: TNotifyEvent read FOnDirectionsChanged write FOnDirectionsChanged;
  end;

implementation

uses
  Lang, GMFunctions, WebControl,
  DateUtils, SysUtils, XMLIntf, XMLDoc;

{ TGMDirection }

function TGMDirection.AddWaypoint: TWaypoint;
begin
  Result := FDirectionsRequest.FWaypointsList.Add;
end;

procedure TGMDirection.Assign(Source: TPersistent);
begin
  if Source is TGMDirection then
  begin
    Map := TGMDirection(Source).Map;
    AutoShow := TGMDirection(Source).AutoShow;
    HiddeOthers := TGMDirection(Source).HiddeOthers;
    DirectionsRender.Assign(TGMDirection(Source).DirectionsRender);
    DirectionsRequest.Assign(TGMDirection(Source).DirectionsRequest);
    FDirectionsResult.Assign(TGMDirection(Source).FDirectionsResult);
  end
  else
    inherited Assign(Source);
end;

procedure TGMDirection.ClearWaypoint;
begin
  FDirectionsRequest.FWaypointsList.Clear;
end;

constructor TGMDirection.Create(AOwner: TComponent);
begin
  inherited;

  FCountDirResult := 0;
  FAutoShow := True;
  FHiddeOthers := True;
  FDirectionsRender := TDirectionsRenderer.Create(Self);
  FDirectionsRequest := TDirectionsRequest.Create(Self);
  FDirectionsResult := TObjectList.Create;
end;

procedure TGMDirection.Delete(ResultIndex: Integer);
begin
  if ResultIndex >= Count then
    raise Exception.Create(GetTranslateText('ndice fuera de rango', Language));

  ExecuteScript('DeleteObject', Format('%d,%d', [IdxList, DirectionsResult[ResultIndex].Index]));
  FDirectionsResult.Delete(ResultIndex);
end;

procedure TGMDirection.DeleteMapObjects;
begin
  ExecuteScript('DeleteObjects', IntToStr(IdxList));
end;

procedure TGMDirection.DeleteWaypoint(Index: Integer);
begin
  FDirectionsRequest.FWaypointsList.Delete(Index);
end;

destructor TGMDirection.Destroy;
begin
  if Assigned(FDirectionsRender) then FreeAndNil(FDirectionsRender);
  if Assigned(FDirectionsRequest) then FreeAndNil(FDirectionsRequest);
  if Assigned(FDirectionsResult) then FreeAndNil(FDirectionsResult);

  inherited;
end;

function TGMDirection.GetAPIUrl: string;
begin
  Result := 'https://developers.google.com/maps/documentation/javascript/reference#DirectionsService';
end;

function TGMDirection.GetCount: Integer;
begin
  Result := FDirectionsResult.Count;
end;

function TGMDirection.GetDirectionResult(I: Integer): TDirectionsResult;
begin
  Result := TDirectionsResult(FDirectionsResult[i]);
end;

procedure TGMDirection.MoveWaypoint(CurIndex, NewIndex: Integer);
begin
  FDirectionsRequest.FWaypointsList.Move(CurIndex, NewIndex);
end;

procedure TGMDirection.ShowElements;
begin
end;

procedure TGMDirection.ShowRoute(ResultIndex, RouteIndex: Integer;
  HiddeOthers: Boolean);
begin
  if ResultIndex >= Count then
    raise Exception.Create(GetTranslateText('ndice fuera de rango', Language));

  TDirectionsResult(FDirectionsResult[ResultIndex]).ShowRoute(RouteIndex, HiddeOthers);
end;

procedure TGMDirection.EventFired(EventType: TEventType;
  Params: array of const);
var
  Idx: Integer;
  i: Integer;
begin
  inherited;

//  if EventType = etInfoWinCloseClick then Exit;

  if EventType = etDirectionsChanged then
  begin
    if High(Params) <> 1 then
      raise Exception.Create(GetTranslateText('Nmero de parmetros incorrecto', Map.Language));

    {$IF CompilerVersion < 21}  // ES: si la versin es inferior a la 2010 - EN: if lower than 2010 version
    if Params[0].VType <> vtAnsiString then
    {$ELSE}
    if Params[0].VType <> vtUnicodeString then
    {$IFEND}
      raise Exception.Create(GetTranslateText('Tipo de parmetro incorrecto', Map.Language));

    if Params[1].VType <> vtInteger then
      raise Exception.Create(GetTranslateText('Tipo de parmetro incorrecto', Map.Language));

    // ES: buscamos indice en TDirectionsResult // EN: search index in TDirectionsResult
    Idx := -1;
    for i := 0 to Count - 1 do
      if TDirectionsResult(FDirectionsResult[i]).FIndex = Params[1].VInteger then
      begin
        Idx := i;
        Break;
      end;

    if Idx = -1 then
      raise Exception.Create(GetTranslateText('Valor de parmetro incorrecto', Map.Language));

    {$IF CompilerVersion < 21}  // ES: si la versin es inferior a la 2010 - EN: if lower than 2010 version
    TDirectionsResult(FDirectionsResult[Idx]).FXMLData.Text := String(PChar(Params[0].VAnsiString));
    {$ELSE}
    TDirectionsResult(FDirectionsResult[Idx]).FXMLData.Text := String(PChar(Params[0].VUnicodeString));
    {$IFEND}
  end;
end;

procedure TGMDirection.Execute;
  function WaypointsToStr: string;
  const
    WayP = '%s~|~%s~|~%s~|~%s';
  var
    i: Integer;
  begin
    Result := '';
    for i := 0 to DirectionsRequest.Count - 1 do
    begin
      if i <> 0 then Result := Result + '$|$';

      Result := Result + Format(WayP, [GetConvertedString(DirectionsRequest[i].Location.Address),
                                       DirectionsRequest[i].Location.LatLng.LatToStr,
                                       DirectionsRequest[i].Location.LatLng.LngToStr,
                                       LowerCase(BoolToStr(DirectionsRequest[i].StopOver, True))])
    end;
  end;
const
  StrParams = '%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s';
var
  Params: string;
  WC: TWebControl;
  DirResult: TDirectionsResult;
begin
  if not Assigned(Map) then
    raise Exception.Create(GetTranslateText('Mapa no asignado', Language));

  Inc(FCountDirResult);
  Params := Format(StrParams, [
                  LowerCase(BoolToStr(DirectionsRequest.AvoidHighways, True)),
                  LowerCase(BoolToStr(DirectionsRequest.AvoidTolls, True)),
                  QuotedStr(GetConvertedString(DirectionsRequest.Destination.Address)),
                  DirectionsRequest.Destination.LatLng.LatToStr(Map.Precision),
                  DirectionsRequest.Destination.LatLng.LngToStr(Map.Precision),
                  LowerCase(BoolToStr(DirectionsRequest.OptimizeWaypoints, True)),
                  QuotedStr(GetConvertedString(DirectionsRequest.Origin.Address)),
                  DirectionsRequest.Origin.LatLng.LatToStr(Map.Precision),
                  DirectionsRequest.Origin.LatLng.LngToStr(Map.Precision),
                  LowerCase(BoolToStr(DirectionsRequest.ProvideRouteAlt, True)),
                  QuotedStr(TTransform.RegionToStr(DirectionsRequest.Region)),
                  IntToStr(MilliSecondsBetween(DirectionsRequest.TransitOpt.ArrivalDate + DirectionsRequest.TransitOpt.ArrivalTime, EncodeDateTime(1970, 1, 1, 0, 0, 0, 1))),
                  IntToStr(MilliSecondsBetween(DirectionsRequest.TransitOpt.DepartureDate + DirectionsRequest.TransitOpt.DepartureTime, EncodeDateTime(1970, 1, 1, 0, 0, 0, 1))),
                  QuotedStr(TTransform.TravelModeToStr(DirectionsRequest.TravelMode)),
                  QuotedStr(TTransform.UnitSystemToStr(DirectionsRequest.UnitSystem)),
                  QuotedStr(WaypointsToStr),
                  LowerCase(BoolToStr(HiddeOthers, True)),
                  LowerCase(BoolToStr(AutoShow, True)),
                  LowerCase(BoolToStr(DirectionsRender.Draggable, True)),
                  LowerCase(BoolToStr(DirectionsRender.PreserveViewport, True)),
                  LowerCase(BoolToStr(DirectionsRender.SuppressBicyclingLayer, True)),
                  LowerCase(BoolToStr(DirectionsRender.SuppressInfoWindows, True)),
                  LowerCase(BoolToStr(DirectionsRender.SuppressMarkers, True)),
                  LowerCase(BoolToStr(DirectionsRender.SuppressPolylines, True)),
                  LowerCase(BoolToStr(DirectionsRender.MarkerOptions.Clickable, True)),
                  LowerCase(BoolToStr(DirectionsRender.MarkerOptions.Draggable, True)),
                  LowerCase(BoolToStr(DirectionsRender.MarkerOptions.Flat, True)),
                  QuotedStr(DirectionsRender.MarkerOptions.Icon),
                  LowerCase(BoolToStr(DirectionsRender.PolylineOptions.Clickable, True)),
                  LowerCase(BoolToStr(DirectionsRender.PolylineOptions.Geodesic, True)),
                  QuotedStr(Map.TColorToStr(DirectionsRender.PolylineOptions.StrokeColor)),
                  StringReplace(FloatToStr(DirectionsRender.PolylineOptions.StrokeOpacity), ',', '.', [rfReplaceAll]),
                  IntToStr(DirectionsRender.PolylineOptions.StrokeWeight),
                  IntToStr(IdxList),
                  IntToStr(FCountDirResult)
                  ]);

  ExecuteScript('GetDirections', Params);

  DirResult := TDirectionsResult.Create(Self, FCountDirResult);
  DirResult.FDirectionsRender.Assign(FDirectionsRender);
  DirResult.FFromTo := DirectionsRequest.Origin.ToString + ' to ' + DirectionsRequest.Destination.ToString;
  WC := TWebControl.Create(Map.WebBrowser);
  try
    DirResult.XMLData.Text := WC.GetStringField(DirectionsForm, DirectionsFormXML);
  finally
    FreeAndNil(WC);
  end;
  FDirectionsResult.Add(DirResult);
end;

{ TLatLngStr }

procedure TLatLngStr.Assign(Source: TPersistent);
begin
  if Source is TLatLngStr then
  begin
    LatLng.Assign(TLatLngStr(Source).LatLng);
    Address := TLatLngStr(Source).Address;
  end
  else
    inherited Assign(Source);
end;

constructor TLatLngStr.Create;
begin
  FLatLng := TLatLng.Create;
  FAddress := '';
end;

destructor TLatLngStr.Destroy;
begin
  if Assigned(FLatLng) then FreeAndNil(FLatLng);

  inherited;
end;

function TLatLngStr.ToString: string;
begin
  if FAddress <> '' then Result := FAddress
  else Result := LatLng.ToStr(0);
end;

{ TTransitOptions }

procedure TTransitOptions.Assign(Source: TPersistent);
begin
  if Source is TTransitOptions then
  begin
    ArrivalTime := TTransitOptions(Source).ArrivalTime;
    ArrivalDate := TTransitOptions(Source).ArrivalDate;
    DepartureTime := TTransitOptions(Source).DepartureTime;
    DepartureDate := TTransitOptions(Source).DepartureDate;
  end
  else
    inherited Assign(Source);
end;

constructor TTransitOptions.Create;
begin
  FArrivalTime := EncodeTime(0, 0, 0, 1);
  FDepartureTime := EncodeTime(0, 0, 0, 1);
  FArrivalDate := EncodeDate(1970, 1, 1);
  FDepartureDate := EncodeDate(1970, 1, 1);
end;

procedure TTransitOptions.SetArrivalDate(const Value: TDate);
begin
  if FArrivalDate = Value then Exit;

  FArrivalDate := Value;
  if FArrivalDate < EncodeDate(1970, 1, 1) then
    FArrivalDate := EncodeDate(1970, 1, 1);
end;

procedure TTransitOptions.SetArrivalTime(const Value: TTime);
begin
  if FArrivalTime = Value then Exit;

  FArrivalTime := Value;
  if FArrivalTime < EncodeTime(0, 0, 0, 1) then
    FArrivalTime := EncodeTime(0, 0, 0, 1);
end;

procedure TTransitOptions.SetDepartureDate(const Value: TDate);
begin
  if FDepartureDate = Value then Exit;

  FDepartureDate := Value;
  if FDepartureDate < EncodeDate(1970, 1, 1) then
    FDepartureDate := EncodeDate(1970, 1, 1);
end;

procedure TTransitOptions.SetDepartureTime(const Value: TTime);
begin
  if FDepartureTime = Value then Exit;

  FDepartureTime := Value;
  if FDepartureTime < EncodeTime(0, 0, 0, 1) then
    FDepartureTime := EncodeTime(0, 0, 0, 1);
end;

{ TWaypoint }

procedure TWaypoint.Assign(Source: TPersistent);
begin
  if Source is TWaypoint then
  begin
    Location.Assign(TWaypoint(Source).Location);
    StopOver := TWaypoint(Source).StopOver;
  end
  else
    inherited Assign(Source);
end;

constructor TWaypoint.Create(Collection: TCollection);
begin
  inherited;

  FLocation := TLatLngStr.Create;
  FStopOver := False;
end;

destructor TWaypoint.Destroy;
begin
  if Assigned(FLocation) then FreeAndNil(FLocation);

  inherited;
end;

function TWaypoint.GetDisplayName: string;
begin
  if Length(FTitle) > 0 then
  begin
    if Length(FTitle) > 20 then
      Result := Copy(FTitle, 0, 20) + '...'
    else
      Result := FTitle;
  end
  else
  begin
    if Length(FLocation.Address) > 0 then
    begin
      if Length(FLocation.Address) > 20 then
        Result := Copy(FLocation.Address, 0, 20) + '...'
      else
        Result := FLocation.Address;
      FTitle := Result;
    end
    else
    begin
      Result := inherited GetDisplayName;
      FTitle := Result;
    end;
  end;
end;

procedure TWaypoint.SetTitle(const Value: string);
begin
  FTitle := Value;

  if FTitle = '' then GetDisplayName;
end;

{ TWaypointsList }

function TWaypointsList.Add: TWaypoint;
begin
  Result := TWaypoint(inherited Add);
end;

procedure TWaypointsList.Clear;
begin
  inherited Clear;
end;

constructor TWaypointsList.Create(AOwner: TPersistent);
begin
  inherited Create(TWaypoint);

  FOwner := AOwner;
end;

procedure TWaypointsList.Delete(Index: Integer);
begin
  inherited Delete(Index);
end;

function TWaypointsList.GetItems(I: Integer): TWaypoint;
begin
  Result := TWaypoint(inherited Items[I]);
end;

function TWaypointsList.GetOwner: TPersistent;
begin
  Result := FOwner;
end;

function TWaypointsList.Insert(Index: Integer): TWaypoint;
begin
  Result := TWaypoint(inherited Insert(Index));
end;

procedure TWaypointsList.Move(CurIndex, NewIndex: Integer);
begin
  Items[CurIndex].Index := NewIndex;
end;

procedure TWaypointsList.SetItems(I: Integer; const Value: TWaypoint);
begin
  inherited SetItem(I, Value);
end;

{ TDirectionsRoute }

procedure TDirectionsRoute.Assign(Source: TObject);
begin
  if Source is TDirectionsRoute then
  begin
    FSumary := TDirectionsRoute(Source).Sumary;
    FCopyrights := TDirectionsRoute(Source).Copyrights;
    FBounds.Assign(TDirectionsRoute(Source).Bounds);
    FWarning.Assign(TDirectionsRoute(Source).FWarning);
    FWaypointOrder.Assign(TDirectionsRoute(Source).FWaypointOrder);
    FOverviewPath.Assign(TDirectionsRoute(Source).FOverviewPath);
    FLeg.Assign(TDirectionsRoute(Source).FLeg);
  end;
end;

constructor TDirectionsRoute.Create;
begin
  FSumary := '';
  FBounds := TLatLngBounds.Create;
  FCopyrights := '';
  FWarning := TStringList.Create;
  FWaypointOrder := TStringList.Create;
  FOverviewPath := TObjectList.Create;
  FLeg := TObjectList.Create;
end;

destructor TDirectionsRoute.Destroy;
begin
  if Assigned(FBounds) then FreeAndNil(FBounds);
  if Assigned(FWarning) then FreeAndNil(FWarning);
  if Assigned(FWaypointOrder) then FreeAndNil(FWaypointOrder);
  if Assigned(FOverviewPath) then FreeAndNil(FOverviewPath);
  if Assigned(FLeg) then FreeAndNil(FLeg);

  inherited;
end;

function TDirectionsRoute.GetCountLeg: Integer;
begin
  Result := FLeg.Count;
end;

function TDirectionsRoute.GetCountOverviewPath: Integer;
begin
  Result := FOverviewPath.Count;
end;

function TDirectionsRoute.GetCountWarning: Integer;
begin
  Result := FWarning.Count;
end;

function TDirectionsRoute.GetCountWaypointOrder: Integer;
begin
  Result := FWaypointOrder.Count;
end;

function TDirectionsRoute.GetLeg(Index: Integer): TDirectionsLeg;
begin
  Result := nil;
  if Assigned(FLeg[Index]) then
    Result := TDirectionsLeg(FLeg[Index]);
end;

function TDirectionsRoute.GetOverviewPath(Index: Integer): TLatLng;
begin
  Result := nil;
  if Assigned(FOverviewPath[Index]) then
    Result := TLatLng(FOverviewPath[Index]);
end;

function TDirectionsRoute.GetWarning(Index: Integer): string;
begin
  Result := '';
  if FWarning.Count >= Index then
    Result := FWarning[Index];
end;

function TDirectionsRoute.GetWaypointOrder(Index: Integer): Integer;
begin
  Result := -1;
  if FWarning.Count >= Index then
    Result := StrToInt(FWarning[Index]);
end;

{ TDirectionsRequest }

procedure TDirectionsRequest.Assign(Source: TPersistent);
begin
  if Source is TDirectionsRequest then
  begin
    AvoidHighways := TDirectionsRequest(Source).AvoidHighways;
    AvoidTolls := TDirectionsRequest(Source).AvoidTolls;
    Destination.Assign(TDirectionsRequest(Source).Destination);
    OptimizeWaypoints := TDirectionsRequest(Source).OptimizeWaypoints;
    Origin.Assign(TDirectionsRequest(Source).Origin);
    ProvideRouteAlt := TDirectionsRequest(Source).ProvideRouteAlt;
    Region := TDirectionsRequest(Source).Region;
    TransitOpt.Assign(TDirectionsRequest(Source).TransitOpt);
    TravelMode := TDirectionsRequest(Source).TravelMode;
    UnitSystem := TDirectionsRequest(Source).UnitSystem;
    WaypointsList.Assign(TDirectionsRequest(Source).WaypointsList);
  end
  else
    inherited Assign(Source);
end;

constructor TDirectionsRequest.Create(aOwner: TPersistent);
begin
  inherited Create;

  FAvoidHighways := False;
  FAvoidTolls := False;
  FDestination := TLatLngStr.Create;
  FOptimizeWaypoints := False;
  FOrigin := TLatLngStr.Create;
  FProvideRouteAlt := True;
  FRegion := r_NO_REGION;
  FTransitOpt := TTransitOptions.Create;
  FTravelMode := tmDRIVING;
  FUnitSystem := usMETRIC;
  FOwner := aOwner;
  FWaypointsList := TWaypointsList.Create(Self);
end;

destructor TDirectionsRequest.Destroy;
begin
  if Assigned(FDestination) then FreeAndNil(FDestination);
  if Assigned(FOrigin) then FreeAndNil(FOrigin);
  if Assigned(FTransitOpt) then FreeAndNil(FTransitOpt);
  if Assigned(FWaypointsList) then FreeAndNil(FWaypointsList);

  inherited;
end;

function TDirectionsRequest.GetCount: Integer;
begin
  Result := FWaypointsList.Count;
end;

function TDirectionsRequest.GetOwner: TPersistent;
begin
  Result := FOwner;
end;

function TDirectionsRequest.GetWaypoint(I: Integer): TWaypoint;
begin
  Result := TWaypoint(FWaypointsList[i]);
end;

procedure TDirectionsRequest.SetWaypointsList(const Value: TWaypointsList);
begin
  FWaypointsList.Assign(Value);
end;

{ TDistance }

procedure TDistance.Assign(Source: TObject);
begin
  if Source is TDistance then
  begin
    FText := TDistance(Source).Text;
    FValue := TDistance(Source).Value;
  end;
end;

constructor TDistance.Create;
begin
  FText := '';
  FValue := 0;
end;

{ TDuration }

procedure TDuration.Assign(Source: TObject);
begin
  if Source is TDistance then
  begin
    FText := TDistance(Source).Text;
    FValue := TDistance(Source).Value;
  end;
end;

constructor TDuration.Create;
begin
  FText := '';
  FValue := 0;
end;

{ TDirectionsStep }

procedure TDirectionsStep.Assign(Source: TObject);
begin
  if Source is TDirectionsStep then
  begin
    FDistance.Assign(TDirectionsStep(Source).Distance);
    FDuration.Assign(TDirectionsStep(Source).Duration);
    FEndLocation.Assign(TDirectionsStep(Source).EndLocation);
    FInstructions := TDirectionsStep(Source).Instructions;
    FPath.Assign(TDirectionsStep(Source).FPath);
    FStartLocation.Assign(TDirectionsStep(Source).StartLocation);
    FTransit.Assign(TDirectionsStep(Source).Transit);
    FTravelMode := TDirectionsStep(Source).TravelMode;
    FSteps.Assign(TDirectionsStep(Source).FSteps);
  end;
end;

constructor TDirectionsStep.Create;
begin
  FDistance := TDistance.Create;
  FDuration := TDuration.Create;
  FEndLocation := TLatLng.Create;
  FInstructions := '';
  FPath := TObjectList.Create;
  FStartLocation := TLatLng.Create;
  FTransit := TTransitDetails.Create;
  FTravelMode := tmDRIVING;
end;

destructor TDirectionsStep.Destroy;
begin
  if Assigned(FDistance) then FreeAndNil(FDistance);
  if Assigned(FDuration) then FreeAndNil(FDuration);
  if Assigned(FEndLocation) then FreeAndNil(FEndLocation);
  if Assigned(FPath) then FreeAndNil(FPath);
  if Assigned(FStartLocation) then FreeAndNil(FStartLocation);
  if Assigned(FSteps) then FreeAndNil(FSteps);
  if Assigned(FTransit) then FreeAndNil(FTransit);

  inherited;
end;

function TDirectionsStep.GetCountPath: Integer;
begin
  Result := FPath.Count;
end;

function TDirectionsStep.GetPath(Index: integer): TLatLng;
begin
  Result := nil;
  if Assigned(FPath[Index]) then
    Result := TLatLng(FPath[Index]);
end;

{ TTransitDetails }

procedure TTransitDetails.Assign(Source: TObject);
begin
  if Source is TTransitDetails then
  begin
    FArribalStop.Assign(TTransitDetails(Source).ArribalStop);
    FArribalTime.Assign(TTransitDetails(Source).ArribalTime);
    FDepartureStop.Assign(TTransitDetails(Source).DepartureStop);
    FDepartureTime.Assign(TTransitDetails(Source).DepartureTime);
    FHeadsign := TTransitDetails(Source).Headsign;
    FHeadway := TTransitDetails(Source).Headway;
    FLine.Assign(TTransitDetails(Source).Line);
    FNumStops := TTransitDetails(Source).NumStops;
  end;
end;

constructor TTransitDetails.Create;
begin
  FArribalStop := TTransitStop.Create;
  FArribalTime := TTimeClass.Create;
  FDepartureStop := TTransitStop.Create;
  FDepartureTime := TTimeClass.Create;
  FHeadsign := '';
  FHeadway := 0;
  FLine := TTransitLine.Create;
  FNumStops := 0;
end;

destructor TTransitDetails.Destroy;
begin
  if Assigned(FArribalStop) then FreeAndNil(FArribalStop);
  if Assigned(FArribalTime) then FreeAndNil(FArribalTime);
  if Assigned(FDepartureStop) then FreeAndNil(FDepartureStop);
  if Assigned(FDepartureTime) then FreeAndNil(FDepartureTime);
  if Assigned(FLine) then FreeAndNil(FLine);

  inherited;
end;

{ TTransitStop }

procedure TTransitStop.Assign(Source: TObject);
begin
  if Source is TTransitStop then
  begin
    FLocation.Assign(TTransitStop(Source).Location);
    FName := TTransitStop(Source).Name;
  end;
end;

constructor TTransitStop.Create;
begin
  FLocation := TLatLng.Create;
  FName := '';
end;

destructor TTransitStop.Destroy;
begin
  if Assigned(FLocation) then FreeAndNil(FLocation);

  inherited;
end;

{ TTimeClass }

procedure TTimeClass.Assign(Source: TObject);
begin
  if Source is TTimeClass then
  begin
    FText := TTimeClass(Source).Text;
    FTimeZone := TTimeClass(Source).TimeZone;
    FValue := TTimeClass(Source).Value;
  end;
end;

constructor TTimeClass.Create;
begin
  FText := '';
  FTimeZone := '';
  FValue := EncodeTime(0, 0, 0, 1);
end;

{ TTransitLine }

procedure TTransitLine.Assign(Source: TObject);
begin
  if Source is TTransitLine then
  begin
    FAgencie.Assign(TTransitLine(Source).FAgencie);
    FColor := TTransitLine(Source).Color;
    FIcon := TTransitLine(Source).Icon;
    FName := TTransitLine(Source).Name;
    FShortName := TTransitLine(Source).ShortName;
    FTextColor := TTransitLine(Source).TextColor;
    FUrl := TTransitLine(Source).Url;
    FVehicle.Assign(TTransitLine(Source).Vehicle);
  end;
end;

constructor TTransitLine.Create;
begin
  FAgencie := TObjectList.Create;
  FColor := '';
  FIcon := '';
  FName := '';
  FShortName := '';
  FTextColor := '';
  FUrl := '';
  FVehicle := TTransitVehicle.Create;
end;

destructor TTransitLine.Destroy;
begin
  if Assigned(FAgencie) then FreeAndNil(FAgencie);
  if Assigned(FVehicle) then FreeAndNil(FVehicle);

  inherited;
end;

function TTransitLine.GetAgencie(Index: Integer): TTransitAgency;
begin
  Result := nil;
  if Assigned(FAgencie[Index]) then
    Result := TTransitAgency(FAgencie[Index]);
end;

function TTransitLine.GetCountAgencie: Integer;
begin
  Result := FAgencie.Count;
end;

{ TTransitAgency }

procedure TTransitAgency.Assign(Source: TObject);
begin
  if Source is TTransitAgency then
  begin
    FName := TTransitAgency(Source).Name;
    FPhone := TTransitAgency(Source).Phone;
    FUrl := TTransitAgency(Source).Url;
  end;
end;

constructor TTransitAgency.Create;
begin
  FName := '';
  FPhone := '';
  FUrl := '';
end;

{ TTransitVehicle }

procedure TTransitVehicle.Assign(Source: TObject);
begin
  if Source is TTransitVehicle then
  begin
    FIcon := TTransitVehicle(Source).Icon;
    FLocalIcon := TTransitVehicle(Source).LocalIcon;
    FName := TTransitVehicle(Source).Name;
    FVehicleType := TTransitVehicle(Source).VehicleType;
  end;
end;

constructor TTransitVehicle.Create;
begin
  FIcon := '';
  FLocalIcon := '';
  FName := '';
  FVehicleType := vtOTHER;
end;

{ TMarkerOptions }

procedure TMarkerOptions.Assign(Source: TPersistent);
begin
  if Source is TMarkerOptions then
  begin
    Clickable := TMarkerOptions(Source).Clickable;
    Draggable := TMarkerOptions(Source).Draggable;
    Flat := TMarkerOptions(Source).Flat;
    Icon := TMarkerOptions(Source).Icon;
  end
  else inherited;
end;

constructor TMarkerOptions.Create(aOwner: TPersistent);
begin
  FClickable := True;
  FDraggable := False;
  FFlat := True;
  FIcon := '';
end;

procedure TMarkerOptions.SetClickable(const Value: Boolean);
begin
  if FClickable = Value then Exit;

  FClickable := Value;
  if Assigned(FOnChange) then FOnChange(Self);
end;

procedure TMarkerOptions.SetDraggable(const Value: Boolean);
begin
  if FDraggable = Value then Exit;

  FDraggable := Value;
  if Assigned(FOnChange) then FOnChange(Self);
end;

procedure TMarkerOptions.SetFlat(const Value: Boolean);
begin
  if FFlat = Value then Exit;

  FFlat := Value;
  if Assigned(FOnChange) then FOnChange(Self);
end;

procedure TMarkerOptions.SetIcon(const Value: string);
begin
  if FIcon = Value then Exit;

  FIcon := Value;
  if Assigned(FOnChange) then FOnChange(Self);
end;

{ TPolylineOptions }

procedure TPolylineOptions.Assign(Source: TPersistent);
begin
  if Source is TPolylineOptions then
  begin
    Clickable := TPolylineOptions(Source).Clickable;
    Geodesic := TPolylineOptions(Source).Geodesic;
    StrokeColor := TPolylineOptions(Source).StrokeColor;
    StrokeOpacity := TPolylineOptions(Source).StrokeOpacity;
    StrokeWeight := TPolylineOptions(Source).StrokeWeight;
  end
  else inherited;
end;

constructor TPolylineOptions.Create(aOwner: TPersistent);
begin
  FClickable := True;
  FGeodesic := False;
  FStrokeColor := clBlue;
  FStrokeOpacity := 0.5;
  FStrokeWeight := 2;
end;

procedure TPolylineOptions.SetClickable(const Value: Boolean);
begin
  if FClickable = Value then Exit;

  FClickable := Value;
  if Assigned(FOnChange) then FOnChange(Self);
end;

procedure TPolylineOptions.SetGeodesic(const Value: Boolean);
begin
  if FGeodesic = Value then Exit;

  FGeodesic := Value;
  if Assigned(FOnChange) then FOnChange(Self);
end;

procedure TPolylineOptions.SetStrokeColor(const Value: TColor);
begin
  if FStrokeColor = Value then Exit;

  FStrokeColor := Value;
  if Assigned(FOnChange) then FOnChange(Self);
end;

procedure TPolylineOptions.SetStrokeOpacity(const Value: Real);
begin
  if FStrokeOpacity = Value then Exit;

  FStrokeOpacity := Value;
  if FStrokeOpacity < 0 then FStrokeOpacity := 0;
  if FStrokeOpacity > 1 then FStrokeOpacity := 1;
  FStrokeOpacity := Trunc(FStrokeOpacity * 100) / 100;
  if Assigned(FOnChange) then FOnChange(Self);
end;

procedure TPolylineOptions.SetStrokeWeight(const Value: Integer);
begin
  if FStrokeWeight = Value then Exit;

  FStrokeWeight := Value;
  if FStrokeWeight < 1 then FStrokeWeight := 1;
  if FStrokeWeight > 10 then FStrokeWeight := 10;
  if Assigned(FOnChange) then FOnChange(Self);
end;

{ TDirectionsRenderer }

procedure TDirectionsRenderer.Assign(Source: TPersistent);
begin
  if Source is TDirectionsRenderer then
  begin
    Draggable := TDirectionsRenderer(Source).Draggable;
    MarkerOptions.Assign(TDirectionsRenderer(Source).MarkerOptions);
    PolylineOptions.Assign(TDirectionsRenderer(Source).PolylineOptions);
    PreserveViewport := TDirectionsRenderer(Source).PreserveViewport;
    SuppressBicyclingLayer := TDirectionsRenderer(Source).SuppressBicyclingLayer;
    SuppressInfoWindows := TDirectionsRenderer(Source).SuppressInfoWindows;
    SuppressMarkers := TDirectionsRenderer(Source).SuppressMarkers;
    SuppressPolylines := TDirectionsRenderer(Source).SuppressPolylines;
  end
  else inherited;
end;

constructor TDirectionsRenderer.Create(aOwner: TObject);
begin
  FDraggable := False;
  FMarkerOptions := TMarkerOptions.Create(Self);
  FMarkerOptions.OnChange := OnMarkerOptionsChange;
  FPolylineOptions := TPolylineOptions.Create(Self);
  FPolylineOptions.OnChange := OnPolylineOptionsChange;
  FPreserveViewport := False;
  FSuppressBicyclingLayer := False;
  FSuppressInfoWindows := False;
  FSuppressMarkers := False;
  FSuppressPolylines := False;
end;

destructor TDirectionsRenderer.Destroy;
begin
  if Assigned(FMarkerOptions) then FreeAndNil(FMarkerOptions);
  if Assigned(FPolylineOptions) then FreeAndNil(FPolylineOptions);

  inherited;
end;

procedure TDirectionsRenderer.OnMarkerOptionsChange(Sender: TObject);
begin
  if Assigned(FOnChange) then FOnChange(Self);
end;

procedure TDirectionsRenderer.OnPolylineOptionsChange(Sender: TObject);
begin
  if Assigned(FOnChange) then FOnChange(Self);
end;

procedure TDirectionsRenderer.SetDraggable(const Value: Boolean);
begin
  if FDraggable = Value then Exit;

  FDraggable := Value;
  if Assigned(FOnChange) then FOnChange(Self);
end;

procedure TDirectionsRenderer.SetPreserveViewport(const Value: Boolean);
begin
  if FPreserveViewport = Value then Exit;

  FPreserveViewport := Value;
  if Assigned(FOnChange) then FOnChange(Self);
end;

procedure TDirectionsRenderer.SetSuppressBicyclingLayer(const Value: Boolean);
begin
  if FSuppressBicyclingLayer = Value then Exit;

  FSuppressBicyclingLayer := Value;
  if Assigned(FOnChange) then FOnChange(Self);
end;

procedure TDirectionsRenderer.SetSuppressInfoWindows(const Value: Boolean);
begin
  if FSuppressInfoWindows = Value then Exit;

  FSuppressInfoWindows := Value;
  if Assigned(FOnChange) then FOnChange(Self);
end;

procedure TDirectionsRenderer.SetSuppressMarkers(const Value: Boolean);
begin
  if FSuppressMarkers = Value then Exit;

  FSuppressMarkers := Value;
  if Assigned(FOnChange) then FOnChange(Self);
end;

procedure TDirectionsRenderer.SetSuppressPolylines(const Value: Boolean);
begin
  if FSuppressPolylines = Value then Exit;

  FSuppressPolylines := Value;
  if Assigned(FOnChange) then FOnChange(Self);
end;

{ TDirectionsResult }

procedure TDirectionsResult.Assign(Source: TObject);
begin
  if Source is TDirectionsResult then
    FRoutes.Assign(TDirectionsResult(Source).FRoutes);
end;

constructor TDirectionsResult.Create(aOwner: TGMDirection; Index: Integer);
begin
  FRoutes := TObjectList.Create;
  FXMLData := TStringList.Create;
  FXMLData.OnChange := OnXMLChange;
  FOwner := aOwner;
  FIndex := Index;
  FDirectionsRender := TDirectionsRenderer.Create(Self);
  FDirectionsRender.OnChange := OnDirectionsRenderChange;
end;

destructor TDirectionsResult.Destroy;
begin
  if Assigned(FRoutes) then FreeAndNil(FRoutes);
  if Assigned(FXMLData) then FreeAndNil(FXMLData);
  if Assigned(FDirectionsRender) then FreeAndNil(FDirectionsRender);

  inherited;
end;

function TDirectionsResult.GetCount: Integer;
begin
  Result := FRoutes.Count;
end;

function TDirectionsResult.GetRoutes(Index: Integer): TDirectionsRoute;
begin
  Result := nil;
  if Assigned(FRoutes[Index]) then
    Result := TDirectionsRoute(FRoutes[Index]);
end;

procedure TDirectionsResult.OnDirectionsRenderChange(Sender: TObject);
const
  StrParams = '%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s';
var
  Params: string;
begin
  if not Assigned(FOwner) then Exit;

  Params := Format(StrParams, [
                  LowerCase(BoolToStr(DirectionsRender.Draggable, True)),
                  LowerCase(BoolToStr(DirectionsRender.PreserveViewport, True)),
                  LowerCase(BoolToStr(DirectionsRender.SuppressBicyclingLayer, True)),
                  LowerCase(BoolToStr(DirectionsRender.SuppressInfoWindows, True)),
                  LowerCase(BoolToStr(DirectionsRender.SuppressMarkers, True)),
                  LowerCase(BoolToStr(DirectionsRender.SuppressPolylines, True)),
                  LowerCase(BoolToStr(DirectionsRender.MarkerOptions.Clickable, True)),
                  LowerCase(BoolToStr(DirectionsRender.MarkerOptions.Draggable, True)),
                  LowerCase(BoolToStr(DirectionsRender.MarkerOptions.Flat, True)),
                  QuotedStr(DirectionsRender.MarkerOptions.Icon),
                  LowerCase(BoolToStr(DirectionsRender.PolylineOptions.Clickable, True)),
                  LowerCase(BoolToStr(DirectionsRender.PolylineOptions.Geodesic, True)),
                  QuotedStr(FOwner.Map.TColorToStr(DirectionsRender.PolylineOptions.StrokeColor)),
                  StringReplace(FloatToStr(DirectionsRender.PolylineOptions.StrokeOpacity), ',', '.', [rfReplaceAll]),
                  IntToStr(DirectionsRender.PolylineOptions.StrokeWeight),
                  IntToStr(FOwner.IdxList),
                  IntToStr(FIndex)
                  ]);

  FOwner.ExecuteScript('DirectionsSetDirRenderOptions', Params);
end;

procedure TDirectionsResult.OnXMLChange(Sender: TObject);
var
  XML: IXMLDocument;

  procedure ParseNodes(Node: IXMLNode);
    procedure GetLatLng(LatLng: TLatLng; Node: IXMLNode);
    begin
      while Assigned(Node) do
      begin
        if Node.NodeName = LBL_D_LAT then
          LatLng.Lat := LatLng.StringToReal(Node.NodeValue);

        if Node.NodeName = LBL_D_LNG then
          LatLng.Lng := LatLng.StringToReal(Node.NodeValue);

        Node := Node.NextSibling;
      end;
    end;

    procedure GetLatLngBounds(LatLngBounds: TLatLngBounds; Node: IXMLNode);
    begin
      while Assigned(Node) do
      begin
        if (Node.NodeName = LBL_D_SOUTHWEST) and (Node.ChildNodes.Count = 2) then
          GetLatLng(LatLngBounds.SW, Node.ChildNodes.First);

        if (Node.NodeName = LBL_D_NORTHEAST) and (Node.ChildNodes.Count = 2) then
          GetLatLng(LatLngBounds.NE, Node.ChildNodes.First);

        Node := Node.NextSibling;
      end;
    end;

    function ParseRouteNode(Node: IXMLNode): TDirectionsRoute;
      function ParseLegNode(Node: IXMLNode): TDirectionsLeg;
        procedure GetDuration(Duration: TDuration; Node: IXMLNode);
        begin
          while Assigned(Node) do
          begin
            if Node.NodeName = LBL_D_TEXT then Duration.FText := Node.NodeValue;
            if Node.NodeName = LBL_D_VALUE then Duration.FValue := Node.NodeValue;

            Node := Node.NextSibling;
          end;
        end;
        procedure GetDistance(Distance: TDistance; Node: IXMLNode);
        begin
          while Assigned(Node) do
          begin
            if Node.NodeName = LBL_D_TEXT then Distance.FText := Node.NodeValue;
            if Node.NodeName = LBL_D_VALUE then Distance.FValue := Node.NodeValue;

            Node := Node.NextSibling;
          end;
        end;

        function ParseStepNode(Node: IXMLNode): TDirectionsStep;
          procedure ParseTransit(Transit: TTransitDetails; Node: IXMLNode);
            procedure GetTransitStop(TStop: TTransitStop; Node: IXMLNode);
            begin
              while Assigned(Node) do
              begin
                // ES: etiqueta "location" (slo una)   // EN: "location" tag (only one)
                if (Node.NodeName = LBL_D_LOCATION) and (Node.ChildNodes.Count = 2) then
                  GetLatLng(TStop.Location, Node.ChildNodes.First);

                // ES: etiqueta "name" (slo una)   // EN: "name" tag (only one)
                if Node.NodeName = LBL_D_NAME then TStop.FName := Node.NodeValue;

                Node := Node.NextSibling;
              end;
            end;
            procedure GetTTime(Time: TTimeClass; Node: IXMLNode);
            begin
              while Assigned(Node) do
              begin
                // ES: etiqueta "text" (slo una)   // EN: "text" tag (only one)
                if Node.NodeName = LBL_D_TEXT then Time.FText := Node.NodeValue;

                // ES: etiqueta "time_zone" (slo una)   // EN: "time_zone" tag (only one)
                if Node.NodeName = LBL_D_TIMEZONE then Time.FTimeZone := Node.NodeValue;

                // ES: etiqueta "value" (slo una)   // EN: "value" tag (only one)
                if Node.NodeName = LBL_D_VALUE then Time.FValue := Node.NodeValue;

                Node := Node.NextSibling;
              end;
            end;
            procedure GetTransitLine(TLine: TTransitLine; Node: IXMLNode);
              procedure GetTransitVehicle(TVehicle: TTransitVehicle; Node: IXMLNode);
              begin
                while Assigned(Node) do
                begin
                  // ES: etiqueta "icon" (slo una)   // EN: "icon" tag (only one)
                  if Node.NodeName = LBL_D_ICON then TVehicle.FIcon := Node.NodeValue;

                  // ES: etiqueta "local_icon" (slo una)   // EN: "local_icon" tag (only one)
                  if Node.NodeName = LBL_D_LOCALICON then TVehicle.FLocalIcon := Node.NodeValue;

                  // ES: etiqueta "name" (slo una)   // EN: "name" tag (only one)
                  if Node.NodeName = LBL_D_NAME then TVehicle.FName := Node.NodeValue;

                  // ES: etiqueta "type" (slo una)   // EN: "type" tag (only one)
                  if Node.NodeName = LBL_D_TYPE then TVehicle.FVehicleType := TTransform.StrToVehicleType(Node.NodeValue);

                  Node := Node.NextSibling;
                end;
              end;
              function GetTransitAgency(Node: IXMLNode): TTransitAgency;
              begin
                Result := TTransitAgency.Create;

                while Assigned(Node) do
                begin
                  // ES: etiqueta "name" (slo una)   // EN: "name" tag (only one)
                  if Node.NodeName = LBL_D_NAME then Result.FName := Node.NodeValue;

                  // ES: etiqueta "phone" (slo una)   // EN: "phone" tag (only one)
                  if Node.NodeName = LBL_D_PHONE then Result.FPhone := Node.NodeValue;

                  // ES: etiqueta "url" (slo una)   // EN: "url" tag (only one)
                  if Node.NodeName = LBL_D_URL then Result.FUrl := Node.NodeValue;

                  Node := Node.NextSibling;
                end;
              end;
            begin
              while Assigned(Node) do
              begin
                // ES: etiqueta "color" (slo una)   // EN: "color" tag (only one)
                if Node.NodeName = LBL_D_COLOR then TLine.FColor := Node.NodeValue;

                // ES: etiqueta "icon" (slo una)   // EN: "icon" tag (only one)
                if Node.NodeName = LBL_D_ICON then TLine.FIcon := Node.NodeValue;

                // ES: etiqueta "name" (slo una)   // EN: "name" tag (only one)
                if Node.NodeName = LBL_D_NAME then TLine.FName := Node.NodeValue;

                // ES: etiqueta "short_name" (slo una)   // EN: "short_name" tag (only one)
                if Node.NodeName = LBL_D_SHORTNAME then TLine.FShortName := Node.NodeValue;

                // ES: etiqueta "text_color" (slo una)   // EN: "text_color" tag (only one)
                if Node.NodeName = LBL_D_TEXTCOLOR then TLine.FTextColor := Node.NodeValue;

                // ES: etiqueta "url" (slo una)   // EN: "url" tag (only one)
                if Node.NodeName = LBL_D_URL then TLine.FUrl := Node.NodeValue;

                // ES: etiqueta "vehicle" (slo una)   // EN: "vehicle" tag (only one)
                if (Node.NodeName = LBL_D_VEHICLE) and (Node.ChildNodes.Count > 0) then
                  GetTransitVehicle(TLine.FVehicle, Node.ChildNodes.First);

                // ES: etiqueta "agencies" (ninguna, una o varias)  // EN: "agencies" (none, one or more)
                if (Node.NodeName = LBL_D_AGENCIES) and (Node.ChildNodes.Count > 0) then
                  TLine.FAgencie.Add(GetTransitAgency(Node.ChildNodes.First));

                Node := Node.NextSibling;
              end;
            end;
          begin
            while Assigned(Node) do
            begin
              // ES: etiqueta "arrival_stop" (slo una)   // EN: "arrival_stop" tag (only one)
              if (Node.NodeName = LBL_D_ARRIVALSTOP) and (Node.ChildNodes.Count > 0) then
                GetTransitStop(Transit.FArribalStop, Node.ChildNodes.First);

              // ES: etiqueta "arrival_time" (slo una)   // EN: "arrival_time" tag (only one)
              if (Node.NodeName = LBL_D_ARRIVALTIME) and (Node.ChildNodes.Count > 0) then
                GetTTime(Transit.FArribalTime, Node.ChildNodes.First);

              // ES: etiqueta "departure_stop" (slo una)   // EN: "departure_stop" tag (only one)
              if (Node.NodeName = LBL_D_DEPARTURESTOP) and (Node.ChildNodes.Count > 0) then
                GetTransitStop(Transit.FDepartureStop, Node.ChildNodes.First);

              // ES: etiqueta "departure_time" (slo una)   // EN: "departure_time" tag (only one)
              if (Node.NodeName = LBL_D_DEPARTURETIME) and (Node.ChildNodes.Count > 0) then
                GetTTime(Transit.FDepartureTime, Node.ChildNodes.First);

              // ES: etiqueta "headsign" (slo una)   // EN: "headsign" tag (only one)
              if Node.NodeName = LBL_D_HEADSIGN then Transit.FHeadsign := Node.NodeValue;

              // ES: etiqueta "headway" (slo una)   // EN: "headway" tag (only one)
              if Node.NodeName = LBL_D_HEADWAY then Transit.FHeadway := Node.NodeValue;

              // ES: etiqueta "num_stops" (slo una)   // EN: "num_stops" tag (only one)
              if Node.NodeName = LBL_D_NUMSTOPS then Transit.FNumStops := Node.NodeValue;

              // ES: etiqueta "line" (slo una)   // EN: "line" tag (only one)
              if (Node.NodeName = LBL_D_LINE) and (Node.ChildNodes.Count > 0) then
                GetTransitLine(Transit.FLine, Node.ChildNodes.First);

              Node := Node.NextSibling;
            end;
          end;
        var
          LL: TLatLng;
        begin
          Result := TDirectionsStep.Create;
          while Assigned(Node) do
          begin
            // ES: etiqueta "distance" (slo una)   // EN: "distance" tag (only one)
            if (Node.NodeName = LBL_D_DISTANCE) and (Node.ChildNodes.Count = 2) then
              GetDistance(Result.FDistance, Node.ChildNodes.First);

            // ES: etiqueta "duration" (slo una)   // EN: "duration" tag (only one)
            if (Node.NodeName = LBL_D_DURATION) and (Node.ChildNodes.Count = 2) then
              GetDuration(Result.FDuration, Node.ChildNodes.First);

            // ES: etiqueta "end_location" (slo una)   // EN: "end_location" tag (only one)
            if (Node.NodeName = LBL_D_ENDLOCATION) and (Node.ChildNodes.Count = 2) then
              GetLatLng(Result.FEndLocation, Node.ChildNodes.First);

            // ES: etiqueta "start_location" (slo una)   // EN: "start_location" tag (only one)
            if (Node.NodeName = LBL_D_STARTLOCATION) and (Node.ChildNodes.Count = 2) then
              GetLatLng(Result.FStartLocation, Node.ChildNodes.First);

            // ES: etiqueta "instructions" (slo una)   // EN: "instructions" tag (only one)
            if Node.NodeName = LBL_D_INSTRUCTIONS then Result.FInstructions := Node.NodeValue;

            // ES: etiqueta "path" (ninguna, una o varias)   // EN: "path" tag (ninguna, una o varias)
            if (Node.NodeName = LBL_D_PATH) and (Node.ChildNodes.Count = 2) then
            begin
              LL := TLatLng.Create;
              GetLatLng(LL, Node.ChildNodes.First);
              Result.FPath.Add(LL);
            end;

            // ES: etiqueta "steps" (slo una)  // EN: "steps" (only one)
            if (Node.NodeName = LBL_D_STEPS) and (Node.ChildNodes.Count > 0) then
            begin
              Result.FSteps := TDirectionsStep.Create;
              Result.FSteps.Assign(ParseStepNode(Node.ChildNodes.First));
            end;

            // ES: etiqueta "travel_mode" (slo una)   // EN: "travel_mode" tag (only one)
            if Node.NodeName = LBL_D_TRAVELMODE then Result.FTravelMode := TTransform.StrToTravelMode(Node.NodeValue);

            // ES: etiqueta "transit" (ninguna, una o varias)  // EN: "transit" (none, one or more)
            if (Node.NodeName = LBL_D_TRANSIT) and (Node.ChildNodes.Count > 0) then
              ParseTransit(Result.FTransit, Node.ChildNodes.First);

            Node := Node.NextSibling;
          end;
        end;
      var
        LL: TLatLng;
      begin
        Result := TDirectionsLeg.Create;

        while Assigned(Node) do
        begin
          // ES: etiqueta "arrival_time" (slo una)   // EN: "arrival_time" tag (only one)
          if (Node.NodeName = LBL_D_ARRIVALTIME) and (Node.ChildNodes.Count = 2) then
            GetDuration(Result.FArrivalTime, Node.ChildNodes.First);

          // ES: etiqueta "departure_time" (slo una)   // EN: "departure_time" tag (only one)
          if (Node.NodeName = LBL_D_DEPARTURETIME) and (Node.ChildNodes.Count = 2) then
            GetDuration(Result.FDepartureTime, Node.ChildNodes.First);

          // ES: etiqueta "distance" (slo una)   // EN: "distance" tag (only one)
          if (Node.NodeName = LBL_D_DISTANCE) and (Node.ChildNodes.Count = 2) then
            GetDistance(Result.FDistance, Node.ChildNodes.First);

          // ES: etiqueta "duration" (slo una)   // EN: "duration" tag (only one)
          if (Node.NodeName = LBL_D_DURATION) and (Node.ChildNodes.Count = 2) then
            GetDuration(Result.FDuration, Node.ChildNodes.First);

          // ES: etiqueta "end_address" (slo una)   // EN: "end_address" tag (only one)
          if Node.NodeName = LBL_D_ENDADDRESS then Result.FEndAddress := Node.NodeValue;

          // ES: etiqueta "start_address" (slo una)   // EN: "start_address" tag (only one)
          if Node.NodeName = LBL_D_STARTADDRESS then Result.FStartAddress := Node.NodeValue;

          // ES: etiqueta "end_location" (slo una)   // EN: "end_location" tag (only one)
          if (Node.NodeName = LBL_D_ENDLOCATION) and (Node.ChildNodes.Count = 2) then
            GetLatLng(Result.FEndLocation, Node.ChildNodes.First);

          // ES: etiqueta "start_location" (slo una)   // EN: "start_location" tag (only one)
          if (Node.NodeName = LBL_D_STARTLOCATION) and (Node.ChildNodes.Count = 2) then
            GetLatLng(Result.FStartLocation, Node.ChildNodes.First);

          // ES: etiqueta "via_waypoints" (ninguna, una o varias)   // EN: "via_waypoints" tag (ninguna, una o varias)
          if (Node.NodeName = LBL_D_VIAWAYPOINTS) and (Node.ChildNodes.Count = 2) then
          begin
            LL := TLatLng.Create;
            GetLatLng(LL, Node.ChildNodes.First);
            Result.FViaWaypoints.Add(LL);
          end;

          // ES: etiqueta "step" (ninguna, una o varias)  // EN: "step" (none, one or more)
          if (Node.NodeName = LBL_D_STEP) and (Node.ChildNodes.Count > 0) then
            Result.FStep.Add(ParseStepNode(Node.ChildNodes.First));

          Node := Node.NextSibling;
        end;
      end;
    var
      LL: TLatLng;
    begin
      Result := TDirectionsRoute.Create;

      while Assigned(Node) do
      begin
        // ES: etiqueta "summary" (slo una)   // EN: "summary" tag (only one)
        if Node.NodeName = LBL_D_SUMMARY then Result.FSumary := Node.NodeValue;

        // ES: etiqueta "copyrights" (slo una)   // EN: "copyrights" tag (only one)
        if Node.NodeName = LBL_D_COPYRIGHTS then Result.FCopyrights := Node.NodeValue;

        // ES: etiqueta "bounds" (slo una)   // EN: "bounds" tag (only one)
        if (Node.NodeName = LBL_D_BOUNDS) and (Node.ChildNodes.Count = 2) then
          GetLatLngBounds(Result.Bounds, Node.ChildNodes.First);

        // ES: etiqueta "warning" (ninguna, una o varias)   // EN: "warning" tag (ninguna, una o varias)
        if Node.NodeName = LBL_D_WARNING then Result.FWarning.Add(Node.NodeValue);

        // ES: etiqueta "waypoint_order" (ninguna, una o varias)   // EN: "waypoint_order" tag (ninguna, una o varias)
        if Node.NodeName = LBL_D_WAYPOINTORDER then Result.FWaypointOrder.Add(Node.NodeValue);

        // ES: etiqueta "overview_path" (ninguna, una o varias)   // EN: "overview_path" tag (ninguna, una o varias)
        if (Node.NodeName = LBL_D_OVERVIEWPATH) and (Node.ChildNodes.Count = 2) then
        begin
          LL := TLatLng.Create;
          GetLatLng(LL, Node.ChildNodes.First);
          Result.FOverviewPath.Add(LL);
        end;

        // ES: etiqueta "leg" (ninguna, una o varias)  // EN: "leg" (none, one or more)
        if (Node.NodeName = LBL_D_LEG) and (Node.ChildNodes.Count > 0) then
          Result.FLeg.Add(ParseLegNode(Node.ChildNodes.First));

        Node := Node.NextSibling;
      end;
    end;
  begin
    // ES: nos posicionamos en "DirectionsResponse" // EN: go to "DirectionsResponse" tag
    while Assigned(Node) and (Node.NodeName <> LBL_D_DIRECTIONSRESPONSE) do
      Node := Node.NextSibling;

    if not Assigned(Node) or (Node.ChildNodes.Count = 0) then Exit;
    Node := Node.ChildNodes.First;

    while Assigned(Node) do
    begin
      // ES: etiqueta "status" (slo una)   // EN: "status" tag (only one)
      if Node.NodeName = LBL_D_STATUS then
        FStatus := TTransform.StrToDirectionsStatus(Node.NodeValue);

      // ES: etiqueta "route" (ninguna, una o varias)  // EN: "route" (none, one or more)
      if (Node.NodeName = LBL_D_ROUTE) and (Node.ChildNodes.Count > 0) then
        FRoutes.Add(ParseRouteNode(Node.ChildNodes.First));

      Node := Node.NextSibling;
    end;
  end;
begin
  if FXMLData.Text = '' then Exit;

  FRoutes.Clear;
  XML := TXMLDocument.Create(nil);
  try
    {$IF CompilerVersion < 21}  // ES: si la versin es inferior a la 2010 - EN: if lower than 2010 version
    XML.XML.Text := Utf8ToAnsi(FXMLData.Text);
    {$ELSE}
    XML.XML.Text := FXMLData.Text;
    {$IFEND}
    XML.Active := True;
    XML.Version := '1.0';
    XML.Encoding := 'utf-8';

    ParseNodes(XML.ChildNodes.First);
  finally
    XML := nil;
  end;
end;

procedure TDirectionsResult.ShowRoute(RouteIndex: Integer; HiddeOthers: Boolean);
const
  StrParams = '%d,%d,%d,%s';
begin
  if RouteIndex >= Count then
  begin
    if Assigned(FOwner) then
      raise Exception.Create(GetTranslateText('ndice fuera de rango', FOwner.Language))
    else
      raise Exception.Create(GetTranslateText('ndice fuera de rango', English));
  end;

  if Assigned(FOwner) then
    FOwner.ExecuteScript('DirectionsShowRoute', Format(StrParams, [FOwner.IdxList, FIndex, RouteIndex, LowerCase(BoolToStr(HiddeOthers, True))]));
end;

{ TDirectionsLeg }

procedure TDirectionsLeg.Assign(Source: TObject);
begin
  if Source is TDirectionsLeg then
  begin
    FArrivalTime.Assign(TDirectionsLeg(Source).ArrivalTime);
    FDepartureTime.Assign(TDirectionsLeg(Source).DepartureTime);
    FDistance.Assign(TDirectionsLeg(Source).Distance);
    FDuration.Assign(TDirectionsLeg(Source).Duration);
    FEndLocation.Assign(TDirectionsLeg(Source).EndLocation);
    FStartLocation.Assign(TDirectionsLeg(Source).StartLocation);
    FViaWaypoints.Assign(TDirectionsLeg(Source).FViaWaypoints);
    FStep.Assign(TDirectionsLeg(Source).FStep);
    FEndAddress := TDirectionsLeg(Source).EndAddress;
    FStartAddress := TDirectionsLeg(Source).StartAddress;
  end;
end;

constructor TDirectionsLeg.Create;
begin
  FArrivalTime := TDuration.Create;
  FDepartureTime := TDuration.Create;
  FDistance := TDistance.Create;
  FDuration := TDuration.Create;
  FEndAddress := '';
  FStartAddress := '';
  FEndLocation := TLatLng.Create;
  FStartLocation := TLatLng.Create;
  FViaWaypoints := TObjectList.Create;
  FStep := TObjectList.Create;
end;

destructor TDirectionsLeg.Destroy;
begin
  if Assigned(FArrivalTime) then FreeAndNil(FArrivalTime);
  if Assigned(FDepartureTime) then FreeAndNil(FDepartureTime);
  if Assigned(FDistance) then FreeAndNil(FDistance);
  if Assigned(FDuration) then FreeAndNil(FDuration);
  if Assigned(FEndLocation) then FreeAndNil(FEndLocation);
  if Assigned(FStartLocation) then FreeAndNil(FStartLocation);
  if Assigned(FViaWaypoints) then FreeAndNil(FViaWaypoints);
  if Assigned(FStep) then FreeAndNil(FStep);

  inherited;
end;

function TDirectionsLeg.GetCountStep: Integer;
begin
  Result := FStep.Count;
end;

function TDirectionsLeg.GetCountViaWaypoints: Integer;
begin
  Result := FViaWaypoints.Count;
end;

function TDirectionsLeg.GetStep(Index: Integer): TDirectionsStep;
begin
  Result := nil;
  if Assigned(FStep[Index]) then
    Result := TDirectionsStep(FStep[Index]);
end;

function TDirectionsLeg.GetViaWaypoints(Index: Integer): TLatLng;
begin
  Result := nil;
  if Assigned(FViaWaypoints[Index]) then
    Result := TLatLng(FViaWaypoints[Index]);
end;

end.
