{
TLatLngBounds class

  ES: Una instancia de LatLngBounds representa un rectngulo en coordenadas
      geogrficas, incluido uno que cruza el meridiano de 180 grados.
  EN: A LatLngBounds instance represents a rectangle in geographical coordinates,
      including one that crosses the 180 degrees longitudinal meridian.

=========================================================================
History:

ver 0.1.5
  ES:
    - cambio: se suprime la funcin GetCenter
  EN:
    - change: GetCenter function is deleted

ver 0.1.3
  ES:
    - nuevo: aadida funcin GetCenter
    - cambio: se publican las propiedades NE y SW
  EN:
    - new: added function GetCenter
    - change: NE and SW are published now

ver 0.1:
  ES: primera versin
  EN: first version
=========================================================================
IMPORTANTE PROGRAMADORES: Por favor, si tienes comentarios, mejoras,
  ampliaciones, errores y/o cualquier otro tipo de sugerencia, envame un correo a:
  gmlib@cadetill.com

IMPORTANT PROGRAMMERS: please, if you have comments, improvements, enlargements,
  errors and/or any another type of suggestion, please send me a mail to:
  gmlib@cadetill.com
=========================================================================

Copyright () 2011, by Xavier Martinez (cadetill)

@author Xavier Martinez (cadetill)
@web  http://www.cadetill.com
}
unit GMLatLngBounds;

interface

uses
  Classes,
  GMLatLng;

type
  { ****************************************************************************
    https://developers.google.com/maps/documentation/javascript/reference?hl=en#LatLngBounds
  **************************************************************************** }
  TLatLngBounds = class(TPersistent)
  private
    FNE: TLatLng;
    FSW: TLatLng;
  public
    constructor Create(SWLat: Real = 0; SWLng: Real = 0; NELat: Real = 0; NELng: Real = 0); overload; virtual;
    constructor Create(SW, NE: TLatLng); overload; virtual;
    destructor Destroy; override;

    procedure Assign(Source: TPersistent); override;

    function IsEqual(Other: TLatLngBounds): Boolean;
    function ToStr(Precision: Integer = 6): string;
    function ToUrlValue(Precision: Integer = 6): string;
  published
    property SW: TLatLng read FSW write FSW;
    property NE: TLatLng read FNE write FNE;
  end;

implementation

uses
  SysUtils;

{ TLatLngBounds }

procedure TLatLngBounds.Assign(Source: TPersistent);
begin
  if Source is TLatLngBounds then
  begin
    SW.Assign(TLatLngBounds(Source).SW);
    NE.Assign(TLatLngBounds(Source).NE);
  end
  else
    inherited Assign(Source);
end;

constructor TLatLngBounds.Create(SWLat, SWLng, NELat, NELng: Real);
begin
  FSW := TLatLng.Create(SWLat, SWLng);
  FNE := TLatLng.Create(NELat, NELng);
end;

constructor TLatLngBounds.Create(SW, NE: TLatLng);
begin
  FSW := TLatLng.Create;
  FNE := TLatLng.Create;

  FSW.Assign(SW);
  FNE.Assign(NE);
end;

destructor TLatLngBounds.Destroy;
begin
  if Assigned(FSW) then FreeAndNil(FSW);
  if Assigned(FNE) then FreeAndNil(FNE);

  inherited;
end;

function TLatLngBounds.IsEqual(Other: TLatLngBounds): Boolean;
begin
  Result := FNE.IsEqual(Other.NE) and FSW.IsEqual(Other.SW);
end;

function TLatLngBounds.ToStr(Precision: Integer): string;
const
  Str = '(%s, %s)';
begin
  Result := Format(Str, [SW.ToStr(Precision), NE.ToStr(Precision)]);
end;

function TLatLngBounds.ToUrlValue(Precision: Integer): string;
const
  Str = '%s,%s';
begin
  Result := Format(Str, [SW.ToUrlValue(Precision), NE.ToUrlValue(Precision)]);
end;

end.
