{
TGMPolygon component

  ES: componente para mostrar polgonos en un mapa de Google Maps mediante el
      componente TGMMap
  EN: component to show polygons on Google Map map using the component TGMMap

=========================================================================
MODO DE USO/HOW TO USE

  ES: poner el componente en el formulario, linkarlo a un TGMMap y poner los
      polgonos a mostrar
  EN: put the component into a form, link to a TGMMap and put the polygons to show
=========================================================================
History:

ver 0.1.7
  ES:
    cambio: modificados mtodos Set y ShowElement para que usen el nuevo mtodo
      ChangeProperties heredado de TLinkedComponent
    cambio: creada clase intermedia (TBasePolyline) entre TLinkedComponent y TPolyline.
      TPolygon desciende ahora de TBasePolyline.
    cambio: creada clase intermedia (TGMBasePolyline) entre TGMLinkedComponent y
      TGMPolyline. TGMPolygon desciende ahora de TGMBasePolyline.
  EN:
    change: modified all Set and ShowElements methods to use the new method
      ChangeProperties inherited from TLinkedComponent
    change: TGMPolygon -> created new base class (TBasePolyline) between
      TLinkedComponent and TPolyline. TPolygon inherited now from TBasePolyline.
    change: TGMPolygon -> created new base class (TGMBasePolyline) between
      TGMLinkedComponent and TGMPolyline. TGMPolygon inherited now from TGMBasePolyline.

ver 0.1.4
  ES:
    cambio: cambio en el JavaScript de algunos mtodos Set
  EN:
    change: JavaScript changed from some Set methods

ver 0.1.3
  ES:
    cambio: mtodos Set cambiados para evitar duplicidad de cdigo
  EN:
    change: changed Set methods to avoid duplicate code

ver 0.1.2:
  ES- primera versin
  EN- first version
=========================================================================
IMPORTANTE PROGRAMADORES: Por favor, si tienes comentarios, mejoras,
  ampliaciones, errores y/o cualquier otro tipo de sugerencia, envame un correo a:
  gmlib@cadetill.com

IMPORTANT PROGRAMMERS: please, if you have comments, improvements, enlargements,
  errors and/or any another type of suggestion, please send me a mail to:
  gmlib@cadetill.com
=========================================================================

Copyright () 2012, by Xavier Martinez (cadetill)

@author Xavier Martinez (cadetill)
@web  http://www.cadetill.com
}
unit GMPolygon;

interface

uses
  Classes, Graphics,
  GMPolyline, GMMap;

type
  { ****************************************************************************
    internal class
  **************************************************************************** }
  TPolygon = class(TBasePolyline)
  private
    FFillColor: TColor;
    FFillOpacity: Real;
    procedure SetFillColor(const Value: TColor);
    procedure SetFillOpacity(const Value: Real);
  protected
    function ChangeProperties: Boolean; override;
  public
    constructor Create(Collection: TCollection); override;

    procedure Assign(Source: TPersistent); override;
  published
    property FillColor: TColor read FFillColor write SetFillColor;
    property FillOpacity: Real read FFillOpacity write SetFillOpacity; // 0 to 1
  end;

  { ****************************************************************************
    internal class
  **************************************************************************** }
  TPolygons = class(TPolylines)
  private
    procedure SetItems(I: Integer; const Value: TPolygon);
    function GetItems(I: Integer): TPolygon;
  protected
    function GetOwner: TPersistent; override;
  public
    function Add: TPolygon;
    function Insert(Index: Integer): TPolygon;

    property Items[I: Integer]: TPolygon read GetItems write SetItems; default;
  end;

  { ****************************************************************************
    https://developers.google.com/maps/documentation/javascript/reference?hl=en#Polygon
  **************************************************************************** }
  TGMPolygon = class(TGMBasePolyline)
  private
    FOnFillOpacityChange: TLinkedComponentChange;
    FOnFillColorChange: TLinkedComponentChange;
  protected
    function GetAPIUrl: string; override;

    function GetItems(I: Integer): TPolygon;

    procedure ShowElements; override;
    function GetCollectionItemClass: TLinkedComponentClass; override;
    function GetCollectionClass: TLinkedComponentsClass; override;
  public
    function Add: TPolygon;

    property Items[I: Integer]: TPolygon read GetItems; default;
  published
    // eventos
    // events
        // from properties
    property OnFillColorChange: TLinkedComponentChange read FOnFillColorChange write FOnFillColorChange;
    property OnFillOpacityChange: TLinkedComponentChange read FOnFillOpacityChange write FOnFillOpacityChange;
  end;

implementation

uses
  SysUtils,
  GMFunctions, GMConstants;

{ TGMPolygon }

function TGMPolygon.Add: TPolygon;
begin
  Result := TPolygon(inherited Add);
end;

function TGMPolygon.GetAPIUrl: string;
begin
  Result := 'https://developers.google.com/maps/documentation/javascript/reference?hl=en#Polygon';
end;

function TGMPolygon.GetCollectionClass: TLinkedComponentsClass;
begin
  Result := TPolygons;
end;

function TGMPolygon.GetCollectionItemClass: TLinkedComponentClass;
begin
  Result := TPolygon;
end;

function TGMPolygon.GetItems(I: Integer): TPolygon;
begin
  Result := TPolygon(inherited Items[i]);
end;

procedure TGMPolygon.ShowElements;
var
  i: Integer;
begin
  if not ExecuteScript('DeleteObjects', IntToStr(IdxList)) then Exit;

  for i:= 0 to VisualObjects.Count - 1 do
    Items[i].ChangeProperties;
end;

{ TPolygons }

function TPolygons.Add: TPolygon;
begin
  Result := TPolygon(inherited Add);
end;

function TPolygons.GetItems(I: Integer): TPolygon;
begin
  Result := TPolygon(inherited Items[I]);
end;

function TPolygons.GetOwner: TPersistent;
begin
  Result := TGMPolygon(inherited GetOwner);
end;

function TPolygons.Insert(Index: Integer): TPolygon;
begin
  Result := TPolygon(inherited Insert(Index));
end;

procedure TPolygons.SetItems(I: Integer; const Value: TPolygon);
begin
  inherited SetItem(I, Value);
end;

{ TPolygon }

procedure TPolygon.Assign(Source: TPersistent);
begin
  inherited;

  if Source is TPolyline then
  begin
    FillColor := TPolygon(Source).FillColor;
    FillOpacity := TPolygon(Source).FillOpacity;
  end;
end;

function TPolygon.ChangeProperties: Boolean;
const
  StrParams = '%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s';
var
  Params: string;
begin
  inherited;

  Result := False;

  if not Assigned(Collection) or not(Collection is TPolygons) or
     not Assigned(TPolygons(Collection).FGMLinkedComponent) or
     not TGMPolygon(TPolygons(Collection).FGMLinkedComponent).AutoUpdate or
     not Assigned(TGMPolygon(TPolygons(Collection).FGMLinkedComponent).Map) or
     (csDesigning in TGMPolygon(TPolygons(Collection).FGMLinkedComponent).ComponentState) then
    Exit;

  Params := Format(StrParams, [
                  IntToStr(IdxList),
                  LowerCase(BoolToStr(Clickable, True)),
                  LowerCase(BoolToStr(Editable, True)),
                  QuotedStr(TGMPolygon(TPolygons(Collection).FGMLinkedComponent).Map.TColorToStr(FillColor)),
                  StringReplace(FloatToStr(FillOpacity), ',', '.', [rfReplaceAll]),
                  LowerCase(BoolToStr(Geodesic, True)),
                  QuotedStr(TGMPolygon(TPolygons(Collection).FGMLinkedComponent).Map.TColorToStr(StrokeColor)),
                  StringReplace(FloatToStr(StrokeOpacity), ',', '.', [rfReplaceAll]),
                  IntToStr(StrokeWeight),
                  LowerCase(BoolToStr(Visible, True)),
                  QuotedStr(PolylineToStr),
                  IntToStr(Index),
                  QuotedStr(InfoWindow.GetConvertedString),
                  LowerCase(BoolToStr(InfoWindow.DisableAutoPan, True)),
                  IntToStr(InfoWindow.MaxWidth),
                  IntToStr(InfoWindow.PixelOffset.Height),
                  IntToStr(InfoWindow.PixelOffset.Width),
                  LowerCase(BoolToStr(InfoWindow.CloseOtherBeforeOpen, True))
                  ]);

  Result := TGMPolygon(TPolygons(Collection).FGMLinkedComponent).ExecuteScript('MakePolygon', Params);
  TGMPolygon(TPolygons(Collection).FGMLinkedComponent).ErrorControl;
end;

constructor TPolygon.Create(Collection: TCollection);
begin
  inherited;

  FFillOpacity := 0.5;
  FFillColor := clRed;
end;

procedure TPolygon.SetFillColor(const Value: TColor);
begin
  if FFillColor = Value then Exit;

  FFillColor := Value;

  ChangeProperties;
  if Assigned(TGMPolygon(TPolygons(Collection).FGMLinkedComponent).FOnFillColorChange) then
    TGMPolygon(TPolygons(Collection).FGMLinkedComponent).FOnFillColorChange(
                  TGMPolygon(TPolygons(Collection).FGMLinkedComponent),
                  Index,
                  Self);
end;

procedure TPolygon.SetFillOpacity(const Value: Real);
begin
  if FFillOpacity = Value then Exit;

  FFillOpacity := Value;
  if FFillOpacity < 0 then FFillOpacity := 0;
  if FFillOpacity > 1 then FFillOpacity := 1;
  FFillOpacity := Trunc(FFillOpacity * 100) / 100;

  ChangeProperties;
  if Assigned(TGMPolygon(TPolygons(Collection).FGMLinkedComponent).FOnFillOpacityChange) then
    TGMPolygon(TPolygons(Collection).FGMLinkedComponent).FOnFillOpacityChange(
                  TGMPolygon(TPolygons(Collection).FGMLinkedComponent),
                  Index,
                  Self);
end;

end.
