{
TGMPolyline component

  ES: componente para mostrar polilineas en un mapa de Google Maps mediante el
      componente TGMMap
  EN: component to show polylines on Google Map map using the component TGMMap

=========================================================================
MODO DE USO/HOW TO USE

  ES: poner el componente en el formulario, linkarlo a un TGMMap y poner las
      polilineas a mostrar
  EN: put the component into a form, link to a TGMMap and put the polylines to show
=========================================================================
History:

ver 0.1.7
  ES:
    cambio: modificados mtodos Set y ShowElement para que usen el nuevo mtodo
      ChangeProperties heredado de TLinkedComponent
    cambio: creada clase intermedia (TBasePolyline) entre TLinkedComponent y TPolyline
    cambio: creada clase intermedia (TGMBasePolyline) entre TGMLinkedComponent y
      TGMPolyline.
    cambio: TPolyline -> nueva propiedad Icon para definir el tipo de linea de la
      polilinea.
  EN:
    change: modified all Set and ShowElements methods to use the new method
      ChangeProperties inherited from TLinkedComponent
    change: created new class (TBasePolyline) between TLinkedComponent and TPolyline
    change: created new class (TGMBasePolyline) between TGMLinkedComponent and
      TGMPolyline.
    cambio: TPolyline -> added Icon property to define the line of polyline.

ver 0.1.6
  ES:
    error: TGMPolyline -> corregido error cuando se intentaba mostrar una polilinea
      sin TLinePoints
  EN:
    bug: TGMPolyline -> bug fixed when we try to show a polyline without TLinePoints

ver 0.1.5
  ES:
    nuevo: TPolyline -> aadida propiedad CountLinePoints
    nuevo: TLinePoint -> aadido mtodo ToStr
    nuevo: TLinePoint -> aadida mtodo StringToReal
  EN:
    new: TPolyline -> CountLinePoints property added
    new: TLinePoint -> ToStr method added
    new: TLinePoint -> StringToReal method added

ver 0.1.4
  ES:
    cambio: cambio en el JavaScript de algunos mtodos Set
  EN:
    change: JavaScript changed from some Set methods

ver 0.1.3
  ES:
    cambio: mtodos Set cambiados para evitar duplicidad de cdigo
  EN:
    change: changed Set methods to avoid duplicate code

ver 0.1.2
  ES:
    cambio: cuando cambia la lat/lng de un TLinePoint, se actualiza en el mapa
    error: implementado SetPath en JavaScript
  EN:
    change: when change the lat/lng of a TLinePoint, the map is updated
    bug: implemented SetPath into JavaScript

ver 0.1.1:
  ES- primera versin
  EN- first version
=========================================================================
IMPORTANTE PROGRAMADORES: Por favor, si tienes comentarios, mejoras,
  ampliaciones, errores y/o cualquier otro tipo de sugerencia, envame un correo a:
  gmlib@cadetill.com

IMPORTANT PROGRAMMERS: please, if you have comments, improvements, enlargements,
  errors and/or any another type of suggestion, please send me a mail to:
  gmlib@cadetill.com
=========================================================================

Copyright () 2012, by Xavier Martinez (cadetill)

@author Xavier Martinez (cadetill)
@web  http://www.cadetill.com
}
unit GMPolyline;

interface

uses
  Classes, Graphics,
  GMMap, GMLatLng, GMConstants;

type
  TBasePolyline = class;

  { ****************************************************************************
    internal class
  **************************************************************************** }
  TLinePoint = class(TCollectionItem)
  private
    FLatLng: TLatLng;
    function GetLat: Real;
    function GetLng: Real;
  protected
  public
    constructor Create(Collection: TCollection); override;
    destructor Destroy; override;

    procedure Assign(Source: TPersistent); override;

    procedure SetLat(const Value: Real);
    procedure SetLng(const Value: Real);

    function LatToStr(Precision: Integer = 6): string;
    function LngToStr(Precision: Integer = 6): string;
    function ToStr(Precision: Integer = 6): string;
    function StringToReal(Value: string): Real;
  published
    property Lat: Real read GetLat write SetLat;
    property Lng: Real read GetLng write SetLng;
  end;

  { ****************************************************************************
    internal class
  **************************************************************************** }
  TLinePoints = class(TCollection)
  private
    function GetItems(I: Integer): TLinePoint;
    procedure SetItems(I: Integer; const Value: TLinePoint);
  protected
    FPolyline: TBasePolyline;

    function GetOwner: TPersistent; override;

    function Add: TLinePoint;
    function Insert(Index: Integer): TLinePoint;
    procedure Delete(Index: Integer);
    procedure Move(CurIndex, NewIndex: Integer);
    procedure Clear;

    property Items[I: Integer]: TLinePoint read GetItems write SetItems; default;
  public
    constructor Create(Polyline: TBasePolyline; ItemClass: TCollectionItemClass);

    procedure Assign(Source: TPersistent); override;
  end;

  { ****************************************************************************
    internal class
  **************************************************************************** }
  TSymbol = class(TPersistent)
  private
    FPath: TSymbolPath;
    FFillColor: TColor;
    FStrokeColor: TColor;
    FStrokeWeight: Integer;
    FFillOpacity: Real;
    FStrokeOpacity: Real;
    FOnChange: TNotifyEvent;
    procedure SetFillColor(const Value: TColor);
    procedure SetFillOpacity(const Value: Real);
    procedure SetPath(const Value: TSymbolPath);
    procedure SetStrokeColor(const Value: TColor);
    procedure SetStrokeOpacity(const Value: Real);
    procedure SetStrokeWeight(const Value: Integer);
  protected
    property OnChange: TNotifyEvent read FOnChange write FOnChange;
  public
    constructor Create; virtual;

    procedure Assign(Source: TPersistent); override;
  published
    property FillColor: TColor read FFillColor write SetFillColor;
    property FillOpacity: Real read FFillOpacity write SetFillOpacity; // 0 to 1
    property Path: TSymbolPath read FPath write SetPath;
    property StrokeColor: TColor read FStrokeColor write SetStrokeColor;
    property StrokeOpacity: Real read FStrokeOpacity write SetStrokeOpacity; // 0 to 1
    property StrokeWeight: Integer read FStrokeWeight write SetStrokeWeight; // 1 to 10
  end;

  { ****************************************************************************
    internal class
  **************************************************************************** }
  TValue = class(TPersistent)
  private
    FValue: Integer;
    FMeasure: TMeasure;
    FOnChange: TNotifyEvent;
    procedure SetMeasure(const Value: TMeasure);
    procedure SetValue(const Value: Integer);
  protected
    property OnChange: TNotifyEvent read FOnChange write FOnChange;
  public
    constructor Create; virtual;

    procedure Assign(Source: TPersistent); override;
  published
    property Value: Integer read FValue write SetValue;
    property Measure: TMeasure read FMeasure write SetMeasure;
  end;

  { ****************************************************************************
    internal class
  **************************************************************************** }
  TIconSequence = class(TPersistent)
  private
    FOwner: TBasePolyline;
    FIcon: TSymbol;
    FDistRepeat: TValue;
    FOffSet: TValue;
    FOnChange: TNotifyEvent;
  protected
    procedure OnOffSetChange(Sender: TObject);
    procedure OnDistRepeatChange(Sender: TObject);
    procedure OnIconChange(Sender: TObject);

    property OnChange: TNotifyEvent read FOnChange write FOnChange;
  public
    constructor Create(aOwner: TBasePolyline); virtual;
    destructor Destroy; override;

    procedure Assign(Source: TPersistent); override;
  published
    property Icon: TSymbol read FIcon write FIcon;
    property OffSet: TValue read FOffSet write FOffSet;
    property DistRepeat: TValue read FDistRepeat write FDistRepeat;
  end;

  { ****************************************************************************
    internal class
  **************************************************************************** }
  TBasePolyline = class(TLinkedComponent)
  private
    FStrokeColor: TColor;
    FVisible: Boolean;
    FStrokeWeight: Integer;
    FGeodesic: Boolean;
    FClickable: Boolean;
    FStrokeOpacity: Real;
    FEditable: Boolean;
    FLinePoints: TLinePoints;
    FAutoUpdatePath: Boolean;
    procedure SetClickable(const Value: Boolean);
    procedure SetEditable(const Value: Boolean);
    procedure SetGeodesic(const Value: Boolean);
    procedure SetStrokeColor(const Value: TColor);
    procedure SetStrokeOpacity(const Value: Real);
    procedure SetStrokeWeight(const Value: Integer);
    procedure SetVisible(const Value: Boolean);
    function GetItems(I: Integer): TLinePoint;
  protected
    procedure SetIdxList(const Value: Cardinal); override;

    function PolylineToStr: string;
    function GetCountLinePoints: Integer;
  public
    constructor Create(Collection: TCollection); override;
    destructor Destroy; override;

    procedure Assign(Source: TPersistent); override;

    procedure CenterMapTo;
    procedure GetPath;
    procedure SetPath;

    function AddLinePoint(Lat: Real = 0; Lng: Real = 0): TLinePoint; overload;
    function AddLinePoint(Lat, Lng: string): TLinePoint; overload;
    function InsertLinePoint(Index: Integer; Lat, Lng: Real): TLinePoint;
    procedure DeleteLinePoint(Index: Integer);
    procedure MoveLinePoint(CurIndex, NewIndex: Integer);
    procedure ClearLinePoints;

    property CountLinePoints: Integer read GetCountLinePoints;
    property Items[I: Integer]: TLinePoint read GetItems; default;
  published
    property Clickable: Boolean read FClickable write SetClickable;
    property Editable: Boolean read FEditable write SetEditable;
    property Geodesic: Boolean read FGeodesic write SetGeodesic;
    property StrokeColor: TColor read FStrokeColor write SetStrokeColor;
    property StrokeOpacity: Real read FStrokeOpacity write SetStrokeOpacity; // 0 to 1
    property StrokeWeight: Integer read FStrokeWeight write SetStrokeWeight; // 1 to 10
    property Visible: Boolean read FVisible write SetVisible;
    property AutoUpdatePath: Boolean read FAutoUpdatePath write FAutoUpdatePath;
    // ES/EN: TCollection
    property LinePoints: TLinePoints read FLinePoints write FLinePoints;
    property InfoWindow;
    property Text;
  end;

  { ****************************************************************************
    https://developers.google.com/maps/documentation/javascript/reference?hl=en#Polyline
  **************************************************************************** }
  TGMBasePolyline = class(TGMLinkedComponent)
  private
    FOnRightClick: TLatLngIdxEvent;
    FOnMouseDown: TLatLngIdxEvent;
    FOnMouseMove: TLatLngIdxEvent;
    FOnMouseUp: TLatLngIdxEvent;
    FOnMouseOut: TLatLngIdxEvent;
    FOnDblClick: TLatLngIdxEvent;
    FOnMouseOver: TLatLngIdxEvent;
    FOnClick: TLatLngIdxEvent;
    FOnStrokeColorChange: TLinkedComponentChange;
    FOnVisibleChange: TLinkedComponentChange;
    FOnStrokeWeightChange: TLinkedComponentChange;
    FOnGeodesicChange: TLinkedComponentChange;
    FOnClickableChange: TLinkedComponentChange;
    FOnStrokeOpacityChange: TLinkedComponentChange;
    FOnEditableChange: TLinkedComponentChange;
    FOnPathChange: TLinkedComponentChange;
  protected
    function GetItems(I: Integer): TBasePolyline;

    procedure EventFired(EventType: TEventType; Params: array of const); override;
  public
    procedure GetPath;
    procedure SetPath;

    property Items[I: Integer]: TBasePolyline read GetItems; default;
  published
    property VisualObjects;
    // eventos
    // events
        // from map
    property OnClick: TLatLngIdxEvent read FOnClick write FOnClick;
    property OnDblClick: TLatLngIdxEvent read FOnDblClick write FOnDblClick;
    property OnMouseDown: TLatLngIdxEvent read FOnMouseDown write FOnMouseDown;
    property OnMouseMove: TLatLngIdxEvent read FOnMouseMove write FOnMouseMove;
    property OnMouseOut: TLatLngIdxEvent read FOnMouseOut write FOnMouseOut;
    property OnMouseOver: TLatLngIdxEvent read FOnMouseOver write FOnMouseOver;
    property OnMouseUp: TLatLngIdxEvent read FOnMouseUp write FOnMouseUp;
    property OnRightClick: TLatLngIdxEvent read FOnRightClick write FOnRightClick;
        // from properties
    property OnClickableChange: TLinkedComponentChange read FOnClickableChange write FOnClickableChange;
    property OnEditableChange: TLinkedComponentChange read FOnEditableChange write FOnEditableChange;
    property OnGeodesicChange: TLinkedComponentChange read FOnGeodesicChange write FOnGeodesicChange;
    property OnStrokeColorChange: TLinkedComponentChange read FOnStrokeColorChange write FOnStrokeColorChange;
    property OnStrokeOpacityChange: TLinkedComponentChange read FOnStrokeOpacityChange write FOnStrokeOpacityChange;
    property OnStrokeWeightChange: TLinkedComponentChange read FOnStrokeWeightChange write FOnStrokeWeightChange;
    property OnVisibleChange: TLinkedComponentChange read FOnVisibleChange write FOnVisibleChange;
    property OnPathChange: TLinkedComponentChange read FOnPathChange write FOnPathChange;
  end;

  { ****************************************************************************
    internal class
  **************************************************************************** }
  TPolyline = class(TBasePolyline)
  private
    FIcon: TIconSequence;
    procedure OnIconChange(Sender: TObject);
  protected
    function ChangeProperties: Boolean; override;
  public
    constructor Create(Collection: TCollection); override;
    destructor Destroy; override;

    procedure Assign(Source: TPersistent); override;
  published
    property Icon: TIconSequence read FIcon write FIcon;
  end;

  { ****************************************************************************
    internal class
  **************************************************************************** }
  TPolylines = class(TLinkedComponents)
  private
    procedure SetItems(I: Integer; const Value: TPolyline);
    function GetItems(I: Integer): TPolyline;
  protected
    function GetOwner: TPersistent; override;
  public
    function Add: TPolyline;
    function Insert(Index: Integer): TPolyline;

    property Items[I: Integer]: TPolyline read GetItems write SetItems; default;
  end;

  { ****************************************************************************
    internal class
  **************************************************************************** }
  TGMPolyline = class(TGMBasePolyline)
  private
    FOnIconChange: TLinkedComponentChange;
  protected
    function GetAPIUrl: string; override;

    function GetItems(I: Integer): TPolyline;

    procedure ShowElements; override;
    function GetCollectionItemClass: TLinkedComponentClass; override;
    function GetCollectionClass: TLinkedComponentsClass; override;
  public
    function Add: TPolyline;

    property Items[I: Integer]: TPolyline read GetItems; default;
  published
    property OnIconChange: TLinkedComponentChange read FOnIconChange write FOnIconChange;
  end;

implementation

uses
  SysUtils,
  Lang, WebControl, GMFunctions;

{ TGMPolyline }

function TGMPolyline.Add: TPolyline;
begin
  Result := TPolyline(inherited Add);
end;

function TGMPolyline.GetAPIUrl: string;
begin
  Result := 'https://developers.google.com/maps/documentation/javascript/reference?hl=en#Polyline';
end;

function TGMPolyline.GetCollectionClass: TLinkedComponentsClass;
begin
  Result := TPolylines;
end;

function TGMPolyline.GetCollectionItemClass: TLinkedComponentClass;
begin
  Result := TPolyline;
end;

function TGMPolyline.GetItems(I: Integer): TPolyline;
begin
  Result := TPolyline(inherited Items[i]);
end;

procedure TGMPolyline.ShowElements;
var
  i: Integer;
begin
  if not ExecuteScript('DeleteObjects', IntToStr(IdxList)) then Exit;

  for i:= 0 to VisualObjects.Count - 1 do
    Items[i].ChangeProperties;
end;

{ TPolylines }

function TPolylines.Add: TPolyline;
begin
  Result := TPolyline(inherited Add);
end;

function TPolylines.GetItems(I: Integer): TPolyline;
begin
  Result := TPolyline(inherited Items[I]);
end;

function TPolylines.GetOwner: TPersistent;
begin
  Result := TGMPolyline(inherited GetOwner);
end;

function TPolylines.Insert(Index: Integer): TPolyline;
begin
  Result := TPolyline(inherited Insert(Index));
end;

procedure TPolylines.SetItems(I: Integer; const Value: TPolyline);
begin
  inherited SetItem(I, Value);
end;

{ TPolyline }

procedure TPolyline.Assign(Source: TPersistent);
begin
  inherited;

  if Source is TPolyline then
  begin
    Icon.Assign(TPolyline(Source).Icon);
  end;
end;

function TPolyline.ChangeProperties: Boolean;
const
  StrParams = '%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s';
var
  Params: string;
  DistRepeat: string;
  Offset: string;
begin
  inherited;

  Result := False;

  if not Assigned(Collection) or not(Collection is TPolylines) or
     not Assigned(TPolylines(Collection).FGMLinkedComponent) or
     not TGMPolyline(TPolylines(Collection).FGMLinkedComponent).AutoUpdate or
     not Assigned(TGMPolyline(TPolylines(Collection).FGMLinkedComponent).Map) or
     (csDesigning in TGMPolyline(TPolylines(Collection).FGMLinkedComponent).ComponentState) then
    Exit;

  case Icon.DistRepeat.Measure of
    mPixels: DistRepeat := IntToStr(Icon.DistRepeat.Value) + 'px';
    else DistRepeat := IntToStr(Icon.DistRepeat.Value) + '%';
  end;

  case Icon.OffSet.Measure of
    mPixels: Offset := IntToStr(Icon.OffSet.Value) + 'px';
    else Offset := IntToStr(Icon.OffSet.Value) + '%';
  end;

  Params := Format(StrParams, [
                  IntToStr(IdxList),
                  IntToStr(Index),
                  LowerCase(BoolToStr(Clickable, True)),
                  LowerCase(BoolToStr(Editable, True)),
                  LowerCase(BoolToStr(Geodesic, True)),
                  QuotedStr(TGMPolyline(TPolylines(Collection).FGMLinkedComponent).Map.TColorToStr(StrokeColor)),
                  StringReplace(FloatToStr(StrokeOpacity), ',', '.', [rfReplaceAll]),
                  IntToStr(StrokeWeight),
                  LowerCase(BoolToStr(Visible, True)),
                  QuotedStr(PolylineToStr),
                  QuotedStr(InfoWindow.GetConvertedString),
                  LowerCase(BoolToStr(InfoWindow.DisableAutoPan, True)),
                  IntToStr(InfoWindow.MaxWidth),
                  IntToStr(InfoWindow.PixelOffset.Height),
                  IntToStr(InfoWindow.PixelOffset.Width),
                  LowerCase(BoolToStr(InfoWindow.CloseOtherBeforeOpen, True)),
                  QuotedStr(DistRepeat),
                  QuotedStr(TGMPolyline(TPolylines(Collection).FGMLinkedComponent).Map.TColorToStr(Icon.Icon.FillColor)),
                  StringReplace(FloatToStr(Icon.Icon.FillOpacity), ',', '.', [rfReplaceAll]),
                  QuotedStr(TTransform.SymbolPathToStr(Icon.Icon.Path)),
                  QuotedStr(TGMPolyline(TPolylines(Collection).FGMLinkedComponent).Map.TColorToStr(Icon.Icon.StrokeColor)),
                  StringReplace(FloatToStr(Icon.Icon.StrokeOpacity), ',', '.', [rfReplaceAll]),
                  IntToStr(Icon.Icon.StrokeWeight),
                  QuotedStr(Offset)
                  ]);

  Result := TGMPolyline(TPolylines(Collection).FGMLinkedComponent).ExecuteScript('MakePolyline', Params);
  TGMPolyline(TPolylines(Collection).FGMLinkedComponent).ErrorControl;
end;

constructor TPolyline.Create(Collection: TCollection);
begin
  inherited Create(Collection);

  FIcon := TIconSequence.Create(Self);
  FIcon.OnChange := OnIconChange;
end;

destructor TPolyline.Destroy;
begin
  if Assigned(FIcon) then FreeAndNil(FIcon);

  inherited;
end;

procedure TPolyline.OnIconChange(Sender: TObject);
begin
  if ChangeProperties and Assigned(TGMPolyline(TPolylines(Collection).FGMLinkedComponent).FOnIconChange) then
    TGMPolyline(TPolylines(Collection).FGMLinkedComponent).FOnIconChange(
                  TGMPolyline(TPolylines(Collection).FGMLinkedComponent),
                  Index,
                  Self);
end;

{ TLinePoints }

function TLinePoints.Add: TLinePoint;
begin
  Result := TLinePoint(inherited Add);

  if Assigned(FPolyline) then FPolyline.ChangeProperties;
end;

procedure TLinePoints.Assign(Source: TPersistent);
begin
  inherited;

  if Source is TLinePoints then
    FPolyline := TLinePoints(Source).FPolyline;
end;

procedure TLinePoints.Clear;
begin
  inherited Clear;

  if Assigned(FPolyline) then FPolyline.ChangeProperties;
end;

constructor TLinePoints.Create(Polyline: TBasePolyline; ItemClass: TCollectionItemClass);
begin
  inherited Create(ItemClass);

  FPolyline := Polyline;
end;

procedure TLinePoints.Delete(Index: Integer);
begin
  inherited Delete(Index);

  if Assigned(FPolyline) then FPolyline.ChangeProperties;
end;

function TLinePoints.GetItems(I: Integer): TLinePoint;
begin
  Result := TLinePoint(inherited Items[I]);
end;

function TLinePoints.GetOwner: TPersistent;
begin
  Result := FPolyline;
end;

function TLinePoints.Insert(Index: Integer): TLinePoint;
begin
  Result := TLinePoint(inherited Insert(Index));

  if Assigned(FPolyline) then FPolyline.ChangeProperties;
end;

procedure TLinePoints.Move(CurIndex, NewIndex: Integer);
begin
  Items[CurIndex].Index := NewIndex;

  if Assigned(FPolyline) then FPolyline.ChangeProperties;
end;

procedure TLinePoints.SetItems(I: Integer; const Value: TLinePoint);
begin
  inherited SetItem(I, Value);
end;

{ TLinePoint }

procedure TLinePoint.Assign(Source: TPersistent);
begin
  if Source is TLinePoint then
  begin
    Lat := TLinePoint(Source).Lat;
    Lng := TLinePoint(Source).Lng;
  end
  else
    inherited Assign(Source);
end;

constructor TLinePoint.Create(Collection: TCollection);
begin
  inherited;

  FLatLng := TLatLng.Create;
end;

destructor TLinePoint.Destroy;
begin
  if Assigned(FLatLng) then FreeAndNil(FLatLng);

  inherited;
end;

function TLinePoint.GetLat: Real;
begin
  Result := FLatLng.Lat;
end;

function TLinePoint.GetLng: Real;
begin
  Result := FLatLng.Lng;
end;

function TLinePoint.LatToStr(Precision: Integer): string;
begin
  Result := FLatLng.LatToStr(Precision);
end;

function TLinePoint.LngToStr(Precision: Integer): string;
begin
  Result := FLatLng.LngToStr(Precision);
end;

procedure TLinePoint.SetLat(const Value: Real);
begin
  if FLatLng.Lat = Value then Exit;

  FLatLng.Lat := Value;

  if Assigned(Collection) and
     Assigned(TLinePoints(Collection).FPolyline) and
     Assigned(TLinePoints(Collection).FPolyline.Collection) and
     Assigned(TPolylines(TLinePoints(Collection).FPolyline.Collection).FGMLinkedComponent) and
     TPolylines(TLinePoints(Collection).FPolyline.Collection).FGMLinkedComponent.AutoUpdate then
    TLinePoints(Collection).FPolyline.SetPath;
end;

procedure TLinePoint.SetLng(const Value: Real);
begin
  if FLatLng.Lng = Value then Exit;

  FLatLng.Lng := Value;

  if Assigned(Collection) and
     Assigned(TLinePoints(Collection).FPolyline) and
     Assigned(TLinePoints(Collection).FPolyline.Collection) and
     Assigned(TPolylines(TLinePoints(Collection).FPolyline.Collection).FGMLinkedComponent) and
     TPolylines(TLinePoints(Collection).FPolyline.Collection).FGMLinkedComponent.AutoUpdate then
    TLinePoints(Collection).FPolyline.SetPath;
end;

function TLinePoint.StringToReal(Value: string): Real;
begin
  Result := FLatLng.StringToReal(Value);
end;

function TLinePoint.ToStr(Precision: Integer): string;
begin
  Result := FLatLng.ToStr(Precision);
end;

{ TIconSequence }

procedure TIconSequence.Assign(Source: TPersistent);
begin
  inherited;

  if Source is TIconSequence then
  begin
    Icon.Assign(TIconSequence(Source).Icon);
    DistRepeat.Assign(TIconSequence(Source).DistRepeat);
    OffSet.Assign(TIconSequence(Source).OffSet);
  end;
end;

constructor TIconSequence.Create(aOwner: TBasePolyline);
begin
  FOwner := aOwner;
  FIcon := TSymbol.Create;
  FIcon.OnChange := OnIconChange;
  FOffSet := TValue.Create;
  FOffSet.FValue := 100;
  FOffSet.FMeasure := mPercentage;
  FOffSet.OnChange := OnOffSetChange;
  FDistRepeat := TValue.Create;
  FDistRepeat.FValue := 0;
  FDistRepeat.FMeasure := mPixels;
  FDistRepeat.OnChange := OnDistRepeatChange;
end;

destructor TIconSequence.Destroy;
begin
  if Assigned(FIcon) then FreeAndNil(FIcon);
  if Assigned(FOffSet) then FreeAndNil(FOffSet);
  if Assigned(FDistRepeat) then FreeAndNil(FDistRepeat);

  inherited;
end;

procedure TIconSequence.OnDistRepeatChange(Sender: TObject);
begin
  if Assigned(FOnChange) then FOnChange(Self);
end;

procedure TIconSequence.OnIconChange(Sender: TObject);
begin
  if Assigned(FOnChange) then FOnChange(Self);
end;

procedure TIconSequence.OnOffSetChange(Sender: TObject);
begin
  if Assigned(FOnChange) then FOnChange(Self);
end;

{ TSymbol }

procedure TSymbol.Assign(Source: TPersistent);
begin
  inherited;

  if Source is TSymbol then
  begin
    FillColor := TSymbol(Source).FillColor;
    FillOpacity := TSymbol(Source).FillOpacity;
    Path := TSymbol(Source).Path;
    StrokeColor := TSymbol(Source).StrokeColor;
    StrokeOpacity := TSymbol(Source).StrokeOpacity;
    StrokeWeight := TSymbol(Source).StrokeWeight;
  end;
end;

constructor TSymbol.Create;
begin
  FPath := spNONE;
end;

procedure TSymbol.SetFillColor(const Value: TColor);
begin
  if FFillColor = Value then Exit;

  FFillColor := Value;
  if Assigned(FOnChange) then FOnChange(Self);
end;

procedure TSymbol.SetFillOpacity(const Value: Real);
begin
  if FFillOpacity = Value then Exit;

  FFillOpacity := Value;
  if FFillOpacity < 0 then FFillOpacity := 0;
  if FFillOpacity > 1 then FFillOpacity := 1;
  FFillOpacity := Trunc(FFillOpacity * 100) / 100;
  if Assigned(FOnChange) then FOnChange(Self);
end;

procedure TSymbol.SetPath(const Value: TSymbolPath);
begin
  if FPath = Value then Exit;

  FPath := Value;
  if Assigned(FOnChange) then FOnChange(Self);
end;

procedure TSymbol.SetStrokeColor(const Value: TColor);
begin
  if FStrokeColor = Value then Exit;

  FStrokeColor := Value;
  if Assigned(FOnChange) then FOnChange(Self);
end;

procedure TSymbol.SetStrokeOpacity(const Value: Real);
begin
  if FStrokeOpacity = Value then Exit;

  FStrokeOpacity := Value;
  if FStrokeOpacity < 0 then FStrokeOpacity := 0;
  if FStrokeOpacity > 1 then FStrokeOpacity := 1;
  FStrokeOpacity := Trunc(FStrokeOpacity * 100) / 100;
  if Assigned(FOnChange) then FOnChange(Self);
end;

procedure TSymbol.SetStrokeWeight(const Value: Integer);
begin
  if FStrokeWeight = Value then Exit;

  FStrokeWeight := Value;
  if FStrokeWeight < 1 then FStrokeWeight := 1;
  if FStrokeWeight > 10 then FStrokeWeight := 10;
  if Assigned(FOnChange) then FOnChange(Self);
end;

{ TValue }

procedure TValue.Assign(Source: TPersistent);
begin
  inherited;

  if Source is TValue then
  begin
    Value := TValue(Source).Value;
    Measure := TValue(Source).Measure;
  end;
end;

constructor TValue.Create;
begin
  FValue := 0;
  FMeasure := mPixels;
end;

procedure TValue.SetMeasure(const Value: TMeasure);
begin
  if FMeasure = Value then Exit;

  FMeasure := Value;
  if Assigned(FOnChange) then FOnChange(Self);
end;

procedure TValue.SetValue(const Value: Integer);
begin
  if FValue = Value then Exit;

  FValue := Value;
  if Assigned(FOnChange) then FOnChange(Self);
end;

{ TBasePolyline }

function TBasePolyline.AddLinePoint(Lat, Lng: string): TLinePoint;
begin
  Result := LinePoints.Add;
  Result.Lat := Result.FLatLng.StringToReal(Lat);
  Result.Lng := Result.FLatLng.StringToReal(Lng);
end;

function TBasePolyline.AddLinePoint(Lat, Lng: Real): TLinePoint;
begin
  Result := LinePoints.Add;
  Result.Lat := Lat;
  Result.Lng := Lng;
end;

procedure TBasePolyline.Assign(Source: TPersistent);
begin
  inherited;

  if Source is TBasePolyline then
  begin
    Clickable := TBasePolyline(Source).Clickable;
    Editable  := TBasePolyline(Source).Editable;
    Geodesic := TBasePolyline(Source).Geodesic;
    StrokeColor := TBasePolyline(Source).StrokeColor;
    StrokeOpacity := TBasePolyline(Source).StrokeOpacity;
    StrokeWeight := TBasePolyline(Source).StrokeWeight;
    Visible := TBasePolyline(Source).Visible;
    LinePoints.Assign(TBasePolyline(Source).LinePoints);
    AutoUpdatePath := TBasePolyline(Source).AutoUpdatePath;
  end;
end;

procedure TBasePolyline.CenterMapTo;
begin
  if Assigned(Collection) and (Collection is TPolylines) and
     Assigned(TPolylines(Collection).FGMLinkedComponent) and
     Assigned(TPolylines(Collection).FGMLinkedComponent.Map) and
     (LinePoints.Count > 0) then
    TPolylines(Collection).FGMLinkedComponent.Map.SetCenter(Items[0].Lat, Items[0].Lng);
end;

procedure TBasePolyline.ClearLinePoints;
begin
  LinePoints.Clear;
end;

constructor TBasePolyline.Create(Collection: TCollection);
begin
  inherited;

  FClickable := True;
  FEditable := False;
  FGeodesic := False;
  FStrokeColor := clBlack;
  FStrokeOpacity := 1;
  FStrokeWeight := 2;
  FVisible := True;
  FLinePoints := TLinePoints.Create(Self, TLinePoint);
  FAutoUpdatePath := True;
end;

procedure TBasePolyline.DeleteLinePoint(Index: Integer);
begin
  LinePoints.Delete(Index);
end;

destructor TBasePolyline.Destroy;
begin
  if Assigned(FLinePoints) then FreeAndNil(FLinePoints);

  inherited;
end;

function TBasePolyline.GetCountLinePoints: Integer;
begin
  Result := LinePoints.Count;
end;

function TBasePolyline.GetItems(I: Integer): TLinePoint;
begin
  Result := TLinePoint(FLinePoints[i]);
end;

procedure TBasePolyline.GetPath;
const
  StrParams = '%s,%s';
var
  WC: TWebControl;
  Params: string;
  Path: string;
  i: Integer;
  L1, L2: TStringList;
  IsEqual: Boolean;
  LL: TLatLng;
  AutoUpdate: Boolean;
begin
  if not Assigned(Collection) or not (Collection is TPolylines) or
     not Assigned(TPolylines(Collection).FGMLinkedComponent) or
     not Assigned(TPolylines(Collection).FGMLinkedComponent.Map) then
    Exit;

  Params := Format(StrParams, [IntToStr(IdxList), IntToStr(Index)]);
  if not TGMBasePolyline(TPolylines(Collection).FGMLinkedComponent).ExecuteScript('PolylineGetPath', Params) then
    Exit;

  WC := TWebControl.Create(TPolylines(Collection).FGMLinkedComponent.Map.WebBrowser);
  try
    Path := WC.GetStringField(PolylineForm, PolylineFormPath);
  finally
    FreeAndNil(WC);
  end;

  if Path = '' then Exit;

  L1 := TStringList.Create;
  L2 := TStringList.Create;
  LL := TLatLng.Create;
  AutoUpdate := TPolylines(Collection).FGMLinkedComponent.AutoUpdate;
  try
    L1.Delimiter := ';';
    L2.Delimiter := '|';
    L1.DelimitedText := Path;

    // ES: miramos si no ha cambiado // EN: check if not have changed
    if LinePoints.Count = L1.Count then
    begin
      IsEqual := True;
      for i := 0 to L1.Count - 1 do
      begin
        L2.DelimitedText := L1[i];
        LL.Lat := LL.StringToReal(L2[0]);
        LL.Lng := LL.StringToReal(L2[1]);
        if not LL.IsEqual(Items[i].FLatLng) then
        begin
          IsEqual := False;
          Break;
        end;
      end;

      if IsEqual then
      begin
        if Assigned(LL) then FreeAndNil(LL);
        if Assigned(L1) then FreeAndNil(L1);
        if Assigned(L2) then FreeAndNil(L2);
        Exit;
      end;
    end;

    TPolylines(Collection).FGMLinkedComponent.AutoUpdate := False;
    LinePoints.Clear;
    for i := 0 to L1.Count - 1 do
    begin
      L2.DelimitedText := L1[i];
      if L2.Count <> 2 then Break;
      AddLinePoint(L2[0], L2[1]);
    end;
  finally
    TPolylines(Collection).FGMLinkedComponent.AutoUpdate := AutoUpdate;
    if Assigned(LL) then FreeAndNil(LL);
    if Assigned(L1) then FreeAndNil(L1);
    if Assigned(L2) then FreeAndNil(L2);
  end;

  if Assigned(TGMBasePolyline(TPolylines(Collection).FGMLinkedComponent).OnPathChange) then
    TGMBasePolyline(TPolylines(Collection).FGMLinkedComponent).OnPathChange(TGMBasePolyline(TPolylines(Collection).FGMLinkedComponent), Index, Self);
end;

function TBasePolyline.InsertLinePoint(Index: Integer; Lat,
  Lng: Real): TLinePoint;
begin
  Result := LinePoints.Insert(Index);
  Result.Lat := Lat;
  Result.Lng := Lng;
end;

procedure TBasePolyline.MoveLinePoint(CurIndex, NewIndex: Integer);
begin
  LinePoints.Move(CurIndex, NewIndex);
end;

function TBasePolyline.PolylineToStr: string;
var
  i: Integer;
  Prec: Integer;
begin
  Result := '';

  if not Assigned(Collection) or
     not Assigned(TPolylines(Collection).FGMLinkedComponent) or
     not Assigned(TPolylines(Collection).FGMLinkedComponent.Map) then
    Exit;

  Prec := TPolylines(Collection).FGMLinkedComponent.Map.Precision;
  for i := 0 to FLinePoints.Count - 1 do
    if Result <> '' then
      Result := Result + ';' + FLinePoints[i].LatToStr(Prec) + '|' + FLinePoints[i].LngToStr(Prec)
    else
      Result := FLinePoints[i].LatToStr(Prec) + '|' + FLinePoints[i].LngToStr(Prec);
end;

procedure TBasePolyline.SetClickable(const Value: Boolean);
begin
  if FClickable = Value then Exit;

  FClickable := Value;

  ChangeProperties;
  if Assigned(TGMBasePolyline(TPolylines(Collection).FGMLinkedComponent).FOnClickableChange) then
    TGMBasePolyline(TPolylines(Collection).FGMLinkedComponent).FOnClickableChange(
                  TGMPolyline(TPolylines(Collection).FGMLinkedComponent),
                  Index,
                  Self);
end;

procedure TBasePolyline.SetEditable(const Value: Boolean);
begin
  if FEditable = Value then Exit;

  FEditable := Value;

  ChangeProperties;
  if Assigned(TGMBasePolyline(TPolylines(Collection).FGMLinkedComponent).FOnEditableChange) then
    TGMBasePolyline(TPolylines(Collection).FGMLinkedComponent).FOnEditableChange(
                  TGMPolyline(TPolylines(Collection).FGMLinkedComponent),
                  Index,
                  Self);
end;

procedure TBasePolyline.SetGeodesic(const Value: Boolean);
begin
  if FGeodesic = Value then Exit;

  FGeodesic := Value;

  ChangeProperties;
  if Assigned(TGMBasePolyline(TPolylines(Collection).FGMLinkedComponent).FOnGeodesicChange) then
    TGMBasePolyline(TPolylines(Collection).FGMLinkedComponent).FOnGeodesicChange(
                  TGMPolyline(TPolylines(Collection).FGMLinkedComponent),
                  Index,
                  Self);
end;

procedure TBasePolyline.SetIdxList(const Value: Cardinal);
begin
  inherited;

end;

procedure TBasePolyline.SetPath;
begin
  ChangeProperties;
end;

procedure TBasePolyline.SetStrokeColor(const Value: TColor);
begin
  if FStrokeColor = Value then Exit;

  FStrokeColor := Value;

  ChangeProperties;
  if Assigned(TGMBasePolyline(TPolylines(Collection).FGMLinkedComponent).FOnStrokeColorChange) then
    TGMBasePolyline(TPolylines(Collection).FGMLinkedComponent).FOnStrokeColorChange(
                  TGMPolyline(TPolylines(Collection).FGMLinkedComponent),
                  Index,
                  Self);
end;

procedure TBasePolyline.SetStrokeOpacity(const Value: Real);
begin
  if FStrokeOpacity = Value then Exit;

  FStrokeOpacity := Value;
  if FStrokeOpacity < 0 then FStrokeOpacity := 0;
  if FStrokeOpacity > 1 then FStrokeOpacity := 1;
  FStrokeOpacity := Trunc(FStrokeOpacity * 100) / 100;

  ChangeProperties;
  if Assigned(TGMBasePolyline(TPolylines(Collection).FGMLinkedComponent).FOnStrokeOpacityChange) then
    TGMBasePolyline(TPolylines(Collection).FGMLinkedComponent).FOnStrokeOpacityChange(
                  TGMPolyline(TPolylines(Collection).FGMLinkedComponent),
                  Index,
                  Self);
end;

procedure TBasePolyline.SetStrokeWeight(const Value: Integer);
begin
  if FStrokeWeight = Value then Exit;

  FStrokeWeight := Value;
  if FStrokeWeight < 1 then FStrokeWeight := 1;
  if FStrokeWeight > 10 then FStrokeWeight := 10;

  ChangeProperties;
  if Assigned(TGMBasePolyline(TPolylines(Collection).FGMLinkedComponent).FOnStrokeWeightChange) then
    TGMBasePolyline(TPolylines(Collection).FGMLinkedComponent).FOnStrokeWeightChange(
                  TGMPolyline(TPolylines(Collection).FGMLinkedComponent),
                  Index,
                  Self);
end;

procedure TBasePolyline.SetVisible(const Value: Boolean);
begin
  if FVisible = Value then Exit;

  FVisible := Value;

  ChangeProperties;
  if Assigned(TGMBasePolyline(TPolylines(Collection).FGMLinkedComponent).FOnVisibleChange) then
    TGMBasePolyline(TPolylines(Collection).FGMLinkedComponent).FOnVisibleChange(
                  TGMPolyline(TPolylines(Collection).FGMLinkedComponent),
                  Index,
                  Self);
end;

{ TGMBasePolyline }

procedure TGMBasePolyline.EventFired(EventType: TEventType;
  Params: array of const);
var
  LL: TLatLng;
begin
  inherited;

  if EventType = etInfoWinCloseClick then Exit;

  if High(Params) <> 2 then
    raise Exception.Create(GetTranslateText('Nmero de parmetros incorrecto', Map.Language));

  if (Params[0].VType <> vtExtended) or (Params[1].VType <> vtExtended) then
    raise Exception.Create(GetTranslateText('Tipo de parmetro incorrecto', Map.Language));

  if Params[2].VType <> vtInteger then
    raise Exception.Create(GetTranslateText('Tipo de parmetro incorrecto', Map.Language));

  if Params[2].VInteger > VisualObjects.Count - 1 then
    raise Exception.Create(GetTranslateText('Valor de parmetro incorrecto', Map.Language));

  LL := TLatLng.Create(ControlPrecision(Params[0].VExtended^, Map.Precision),
                       ControlPrecision(Params[1].VExtended^, Map.Precision));
  try
    case EventType of
      etPolylineClick: if Assigned(FOnClick) then FOnClick(Self, LL, Params[2].VInteger, Items[Params[2].VInteger]);
      etPolylineDblClick: if Assigned(FOnDblClick) then FOnDblClick(Self, LL, Params[2].VInteger, Items[Params[2].VInteger]);
      etPolylineMouseDown: if Assigned(FOnMouseDown) then FOnMouseDown(Self, LL, Params[2].VInteger, Items[Params[2].VInteger]);
      etPolylineMouseMove: if Assigned(FOnMouseMove) then FOnMouseMove(Self, LL, Params[2].VInteger, Items[Params[2].VInteger]);
      etPolylineMouseOut: if Assigned(FOnMouseOut) then FOnMouseOut(Self, LL, Params[2].VInteger, Items[Params[2].VInteger]);
      etPolylineMouseOver: if Assigned(FOnMouseOver) then FOnMouseOver(Self, LL, Params[2].VInteger, Items[Params[2].VInteger]);
      etPolylineMouseUp: if Assigned(FOnMouseUp) then FOnMouseUp(Self, LL, Params[2].VInteger, Items[Params[2].VInteger]);
      etPolylineRightClick: if Assigned(FOnRightClick) then FOnRightClick(Self, LL, Params[2].VInteger, Items[Params[2].VInteger]);
    end;
  finally
    FreeAndNil(LL);
  end;
end;

function TGMBasePolyline.GetItems(I: Integer): TBasePolyline;
begin
  Result := TBasePolyline(inherited Items[i]);
end;

procedure TGMBasePolyline.GetPath;
var
  i: Integer;
begin
  if not Assigned(Map) then Exit;

  for i := 0 to VisualObjects.Count - 1 do
    Items[i].GetPath;
end;

procedure TGMBasePolyline.SetPath;
var
  i: Integer;
begin
  if not Assigned(Map) then Exit;

  for i := 0 to VisualObjects.Count - 1 do
    Items[i].SetPath;
end;

end.
