{
TGMRectangle component

  ES: componente para crear rectngulos en un mapa de Google Maps mediante el
      componente TGMMap
  EN: component to create rectangles on Google Map map using the component TGMMap

=========================================================================
MODO DE USO/HOW TO USE

  ES: poner el componente en el formulario, linkarlo a un TGMMap y poner los
      rectngulos a mostrar
  EN: put the component into a form, link to a TGMMap and put the rectangles to
      show
=========================================================================
History:

ver 0.1.7
  ES:
    cambio: modificados mtodos Set y ShowElement para que usen el nuevo mtodo
      ChangeProperties heredado de TLinkedComponent
  EN:
    change: modified all Set and ShowElements methods to use the new method
      ChangeProperties inherited from TLinkedComponent

ver 0.1.5
  ES:
    nuevo: TRectangle -> aadido mtodo GetCenter
  EN:
    new: TRectangle -> GetCenter method added

ver 0.1.4
  ES:
    cambio: cambio en el JavaScript de algunos mtodos Set
  EN:
    change: JavaScript changed from some Set methods

ver 0.1.3
  ES: primera versin
  EN: first version
=========================================================================
IMPORTANTE PROGRAMADORES: Por favor, si tienes comentarios, mejoras,
  ampliaciones, errores y/o cualquier otro tipo de sugerencia, envame un correo a:
  gmlib@cadetill.com

IMPORTANT PROGRAMMERS: please, if you have comments, improvements, enlargements,
  errors and/or any another type of suggestion, please send me a mail to:
  gmlib@cadetill.com
=========================================================================

Copyright () 2012, by Xavier Martinez (cadetill)

@author Xavier Martinez (cadetill)
@web  http://www.cadetill.com
}
unit GMRectangle;

interface

uses
  Classes, Graphics,
  GMMap, GMLatLngBounds, GMLatLng;

type
  { ****************************************************************************
    internal class
  **************************************************************************** }
  TRectangle = class(TLinkedComponent)
  private
    FFillColor: TColor;
    FStrokeColor: TColor;
    FVisible: Boolean;
    FBounds: TLatLngBounds;
    FStrokeWeight: Integer;
    FFillOpacity: Real;
    FClickable: Boolean;
    FStrokeOpacity: Real;
    FEditable: Boolean;
    FIsBoundsUpdt: Boolean;
    procedure SetClickable(const Value: Boolean);
    procedure SetEditable(const Value: Boolean);
    procedure SetFillColor(const Value: TColor);
    procedure SetFillOpacity(const Value: Real);
    procedure SetStrokeColor(const Value: TColor);
    procedure SetStrokeOpacity(const Value: Real);
    procedure SetStrokeWeight(const Value: Integer);
    procedure SetVisible(const Value: Boolean);
    procedure OnChangeBounds(Sender: TObject);
  protected
    procedure SetIdxList(const Value: Cardinal); override;
    function ChangeProperties: Boolean; override;
  public
    constructor Create(Collection: TCollection); override;
    destructor Destroy; override;

    procedure Assign(Source: TPersistent); override;

    function GetCenter: TLatLng;
    procedure CenterMapTo;
  published
    property Bounds: TLatLngBounds read FBounds write FBounds;
    property Clickable: Boolean read FClickable write SetClickable;
    property Editable: Boolean read FEditable write SetEditable;
    property FillColor: TColor read FFillColor write SetFillColor;
    property FillOpacity: Real read FFillOpacity write SetFillOpacity; // 0 to 1
    property StrokeColor: TColor read FStrokeColor write SetStrokeColor;
    property StrokeOpacity: Real read FStrokeOpacity write SetStrokeOpacity; // 0 to 1
    property StrokeWeight: Integer read FStrokeWeight write SetStrokeWeight; // 1 to 10
    property Visible: Boolean read FVisible write SetVisible;
    property InfoWindow;
    property Text;
  end;

  { ****************************************************************************
    internal class
  **************************************************************************** }
  TRectangles = class(TLinkedComponents)
  private
    procedure SetItems(I: Integer; const Value: TRectangle);
    function GetItems(I: Integer): TRectangle;
  protected
    function GetOwner: TPersistent; override;
  public
    function Add: TRectangle;
    function Insert(Index: Integer): TRectangle;

    property Items[I: Integer]: TRectangle read GetItems write SetItems; default;
  end;

  { ****************************************************************************
    https://developers.google.com/maps/documentation/javascript/reference?hl=en#Rectangle
  **************************************************************************** }
  TGMRectangle = class(TGMLinkedComponent)
  private
    FOnStrokeColorChange: TLinkedComponentChange;
    FOnRightClick: TLatLngIdxEvent;
    FOnMouseDown: TLatLngIdxEvent;
    FOnVisibleChange: TLinkedComponentChange;
    FOnMouseMove: TLatLngIdxEvent;
    FOnStrokeWeightChange: TLinkedComponentChange;
    FOnMouseUp: TLatLngIdxEvent;
    FOnFillOpacityChange: TLinkedComponentChange;
    FOnClickableChange: TLinkedComponentChange;
    FOnMouseOut: TLatLngIdxEvent;
    FOnStrokeOpacityChange: TLinkedComponentChange;
    FOnEditableChange: TLinkedComponentChange;
    FOnBoundsChanged: TLinkedComponentChange;
    FOnDblClick: TLatLngIdxEvent;
    FOnMouseOver: TLatLngIdxEvent;
    FOnFillColorChange: TLinkedComponentChange;
    FOnClick: TLatLngIdxEvent;
  protected
    function GetAPIUrl: string; override;

    function GetItems(I: Integer): TRectangle;

    procedure ShowElements; override;
    procedure EventFired(EventType: TEventType; Params: array of const); override;
    function GetCollectionItemClass: TLinkedComponentClass; override;
    function GetCollectionClass: TLinkedComponentsClass; override;
  public
    function Add(SWLat: Real = 0; SWLng: Real = 0; NELat: Real = 0; NELng: Real = 0): TRectangle;

    property Items[I: Integer]: TRectangle read GetItems; default;
  published
    property VisualObjects;
    // eventos
    // events
        // both (map and properties)
    property OnBoundsChanged: TLinkedComponentChange read FOnBoundsChanged write FOnBoundsChanged;
        // from map
    property OnClick: TLatLngIdxEvent read FOnClick write FOnClick;
    property OnDblClick: TLatLngIdxEvent read FOnDblClick write FOnDblClick;
    property OnMouseDown: TLatLngIdxEvent read FOnMouseDown write FOnMouseDown;
    property OnMouseMove: TLatLngIdxEvent read FOnMouseMove write FOnMouseMove;
    property OnMouseOut: TLatLngIdxEvent read FOnMouseOut write FOnMouseOut;
    property OnMouseOver: TLatLngIdxEvent read FOnMouseOver write FOnMouseOver;
    property OnMouseUp: TLatLngIdxEvent read FOnMouseUp write FOnMouseUp;
    property OnRightClick: TLatLngIdxEvent read FOnRightClick write FOnRightClick;
        // from properties
    property OnClickableChange: TLinkedComponentChange read FOnClickableChange write FOnClickableChange;
    property OnEditableChange: TLinkedComponentChange read FOnEditableChange write FOnEditableChange;
    property OnFillColorChange: TLinkedComponentChange read FOnFillColorChange write FOnFillColorChange;
    property OnFillOpacityChange: TLinkedComponentChange read FOnFillOpacityChange write FOnFillOpacityChange;
    property OnStrokeColorChange: TLinkedComponentChange read FOnStrokeColorChange write FOnStrokeColorChange;
    property OnStrokeOpacityChange: TLinkedComponentChange read FOnStrokeOpacityChange write FOnStrokeOpacityChange;
    property OnStrokeWeightChange: TLinkedComponentChange read FOnStrokeWeightChange write FOnStrokeWeightChange;
    property OnVisibleChange: TLinkedComponentChange read FOnVisibleChange write FOnVisibleChange;
  end;

implementation

uses
  Lang, WebControl, GMConstants,
  SysUtils;

{ TGMRectangle }

function TGMRectangle.Add(SWLat, SWLng, NELat, NELng: Real): TRectangle;
begin
  Result := TRectangle(inherited Add);
  Result.Bounds.SW.Lat := SWLat;
  Result.Bounds.SW.Lng := SWLng;
  Result.Bounds.NE.Lat := NELat;
  Result.Bounds.NE.Lng := NELng;
end;

procedure TGMRectangle.EventFired(EventType: TEventType;
  Params: array of const);
var
  LL: TLatLng;
  OldEvent: TNotifyEvent;
begin
  inherited;

  if EventType = etInfoWinCloseClick then Exit;

  if EventType = etRectangleBoundsChange then
  begin
    if High(Params) <> 4 then
      raise Exception.Create(GetTranslateText('Nmero de parmetros incorrecto', Map.Language));

    if (Params[1].VType <> vtExtended) or (Params[2].VType <> vtExtended) or
       (Params[3].VType <> vtExtended) or (Params[4].VType <> vtExtended) then
      raise Exception.Create(GetTranslateText('Tipo de parmetro incorrecto', Map.Language));

    if Params[0].VType <> vtInteger then
      raise Exception.Create(GetTranslateText('Tipo de parmetro incorrecto', Map.Language));

    if Params[0].VInteger > VisualObjects.Count - 1 then
      raise Exception.Create(GetTranslateText('Valor de parmetro incorrecto', Map.Language));

    if Items[Params[0].VInteger].FIsBoundsUpdt then
    begin
      Items[Params[0].VInteger].FIsBoundsUpdt := False;
      Exit;
    end;

    OldEvent := Items[Params[0].VInteger].Bounds.NE.OnChange;
    Items[Params[0].VInteger].Bounds.NE.OnChange := nil;
    Items[Params[0].VInteger].Bounds.SW.OnChange := nil;

    Items[Params[0].VInteger].Bounds.NE.Lat := ControlPrecision(Params[1].VExtended^, Map.Precision);
    Items[Params[0].VInteger].Bounds.NE.Lng := ControlPrecision(Params[2].VExtended^, Map.Precision);
    Items[Params[0].VInteger].Bounds.SW.Lat := ControlPrecision(Params[3].VExtended^, Map.Precision);
    Items[Params[0].VInteger].Bounds.SW.Lng := ControlPrecision(Params[4].VExtended^, Map.Precision);
    if Assigned(FOnBoundsChanged) then FOnBoundsChanged(Self, Params[0].VInteger, Items[Params[0].VInteger]);

    Items[Params[0].VInteger].Bounds.NE.OnChange := OldEvent;
    Items[Params[0].VInteger].Bounds.SW.OnChange := OldEvent;

    Exit;
  end;

  if High(Params) <> 2 then
    raise Exception.Create(GetTranslateText('Nmero de parmetros incorrecto', Map.Language));

  if (Params[0].VType <> vtExtended) or (Params[1].VType <> vtExtended) then
    raise Exception.Create(GetTranslateText('Tipo de parmetro incorrecto', Map.Language));

  if Params[2].VType <> vtInteger then
    raise Exception.Create(GetTranslateText('Tipo de parmetro incorrecto', Map.Language));

  if Params[2].VInteger > VisualObjects.Count - 1 then
    raise Exception.Create(GetTranslateText('Valor de parmetro incorrecto', Map.Language));

  LL := TLatLng.Create(ControlPrecision(Params[0].VExtended^, Map.Precision),
                       ControlPrecision(Params[1].VExtended^, Map.Precision));
  try
    case EventType of
      etRectangleClick: if Assigned(FOnClick) then FOnClick(Self, LL, Params[2].VInteger, Items[Params[2].VInteger]);
      etRectangleDblClick: if Assigned(FOnDblClick) then FOnDblClick(Self, LL, Params[2].VInteger, Items[Params[2].VInteger]);
      etRectangleMouseDown: if Assigned(FOnMouseDown) then FOnMouseDown(Self, LL, Params[2].VInteger, Items[Params[2].VInteger]);
      etRectangleMouseMove: if Assigned(FOnMouseMove) then FOnMouseMove(Self, LL, Params[2].VInteger, Items[Params[2].VInteger]);
      etRectangleMouseOut: if Assigned(FOnMouseOut) then FOnMouseOut(Self, LL, Params[2].VInteger, Items[Params[2].VInteger]);
      etRectangleMouseOver: if Assigned(FOnMouseOver) then FOnMouseOver(Self, LL, Params[2].VInteger, Items[Params[2].VInteger]);
      etRectangleMouseUp: if Assigned(FOnMouseUp) then FOnMouseUp(Self, LL, Params[2].VInteger, Items[Params[2].VInteger]);
      etRectangleRightClick: if Assigned(FOnRightClick) then FOnRightClick(Self, LL, Params[2].VInteger, Items[Params[2].VInteger]);
    end;
  finally
    FreeAndNil(LL);
  end;
end;

function TGMRectangle.GetAPIUrl: string;
begin
  Result := 'https://developers.google.com/maps/documentation/javascript/reference?hl=en#Rectangle';
end;

function TGMRectangle.GetCollectionClass: TLinkedComponentsClass;
begin
  Result := TRectangles;
end;

function TGMRectangle.GetCollectionItemClass: TLinkedComponentClass;
begin
  Result := TRectangle;
end;

function TGMRectangle.GetItems(I: Integer): TRectangle;
begin
  Result := TRectangle(inherited Items[i]);
end;

procedure TGMRectangle.ShowElements;
var
  i: Integer;
begin
  if not ExecuteScript('DeleteObjects', IntToStr(IdxList)) then Exit;

  for i:= 0 to VisualObjects.Count - 1 do
    Items[i].ChangeProperties;
end;

{ TRectangles }

function TRectangles.Add: TRectangle;
begin
  Result := TRectangle(inherited Add);
end;

function TRectangles.GetItems(I: Integer): TRectangle;
begin
  Result := TRectangle(inherited Items[I]);
end;

function TRectangles.GetOwner: TPersistent;
begin
  Result := TGMRectangle(inherited GetOwner);
end;

function TRectangles.Insert(Index: Integer): TRectangle;
begin
  Result := TRectangle(inherited Insert(Index));
end;

procedure TRectangles.SetItems(I: Integer; const Value: TRectangle);
begin
  inherited SetItem(I, Value);
end;

{ TRectangle }

procedure TRectangle.Assign(Source: TPersistent);
begin
  inherited;

  if Source is TRectangle then
  begin
    Bounds.Assign(TRectangle(Source).Bounds);
    Clickable := TRectangle(Source).Clickable;
    Editable  := TRectangle(Source).Editable;
    FillColor := TRectangle(Source).FillColor;
    FillOpacity := TRectangle(Source).FillOpacity;
    StrokeColor := TRectangle(Source).StrokeColor;
    StrokeOpacity := TRectangle(Source).StrokeOpacity;
    StrokeWeight := TRectangle(Source).StrokeWeight;
    Visible := TRectangle(Source).Visible;
  end;
end;

procedure TRectangle.CenterMapTo;
begin
  if Assigned(Collection) and (Collection is TRectangles) and
     Assigned(TRectangles(Collection).FGMLinkedComponent) and
     Assigned(TRectangles(Collection).FGMLinkedComponent.Map) then
    TRectangles(Collection).FGMLinkedComponent.Map.SetCenter(GetCenter);
end;

function TRectangle.ChangeProperties: Boolean;
const
  StrParams = '%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s';
var
  Params: string;
begin
  inherited;

  Result := False;

  if not Assigned(Collection) or not(Collection is TRectangles) or
     not Assigned(TRectangles(Collection).FGMLinkedComponent) or
     not TGMRectangle(TRectangles(Collection).FGMLinkedComponent).AutoUpdate or
     not Assigned(TGMRectangle(TRectangles(Collection).FGMLinkedComponent).Map) or
     (csDesigning in TGMRectangle(TRectangles(Collection).FGMLinkedComponent).ComponentState) then
    Exit;

  Params := Format(StrParams, [
                    IntToStr(IdxList),
                    LowerCase(BoolToStr(Clickable, True)),
                    LowerCase(BoolToStr(Editable, True)),
                    QuotedStr(TGMRectangle(TRectangles(Collection).FGMLinkedComponent).Map.TColorToStr(FillColor)),
                    StringReplace(FloatToStr(FillOpacity), ',', '.', [rfReplaceAll]),
                    QuotedStr(TGMRectangle(TRectangles(Collection).FGMLinkedComponent).Map.TColorToStr(StrokeColor)),
                    StringReplace(FloatToStr(StrokeOpacity), ',', '.', [rfReplaceAll]),
                    IntToStr(StrokeWeight),
                    LowerCase(BoolToStr(Visible, True)),
                    Bounds.SW.LatToStr(TGMRectangle(TRectangles(Collection).FGMLinkedComponent).Map.Precision),
                    Bounds.SW.LngToStr(TGMRectangle(TRectangles(Collection).FGMLinkedComponent).Map.Precision),
                    Bounds.NE.LatToStr(TGMRectangle(TRectangles(Collection).FGMLinkedComponent).Map.Precision),
                    Bounds.NE.LngToStr(TGMRectangle(TRectangles(Collection).FGMLinkedComponent).Map.Precision),
                    IntToStr(Index),
                    QuotedStr(InfoWindow.GetConvertedString),
                    LowerCase(BoolToStr(InfoWindow.DisableAutoPan, True)),
                    IntToStr(InfoWindow.MaxWidth),
                    IntToStr(InfoWindow.PixelOffset.Height),
                    IntToStr(InfoWindow.PixelOffset.Width),
                    LowerCase(BoolToStr(InfoWindow.CloseOtherBeforeOpen, True))
                  ]);

  Result := TGMRectangle(TRectangles(Collection).FGMLinkedComponent).ExecuteScript('MakeRectangle', Params);
  TGMRectangle(TRectangles(Collection).FGMLinkedComponent).ErrorControl;
end;

constructor TRectangle.Create(Collection: TCollection);
begin
  inherited;

  FBounds := TLatLngBounds.Create;
  Bounds.NE.OnChange := OnChangeBounds;
  Bounds.SW.OnChange := OnChangeBounds;
  FClickable := True;
  FEditable := False;
  FFillOpacity := 0.5;
  FFillColor := clRed;
  FStrokeColor := clBlack;
  FStrokeOpacity := 1;
  FStrokeWeight := 2;
  FVisible := True;
  FIsBoundsUpdt := False;
end;

destructor TRectangle.Destroy;
begin
  if Assigned(FBounds) then FreeAndNil(FBounds);

  inherited;
end;

function TRectangle.GetCenter: TLatLng;
const
  StrParams = '%s,%s';
var
  WC: TWebControl;
  Params: string;
begin
  Result := nil;

  if not Assigned(Collection) or not (Collection is TRectangles) or
     not Assigned(TRectangles(Collection).FGMLinkedComponent) or
     not Assigned(TRectangles(Collection).FGMLinkedComponent.Map) then
    Exit;

  Params := Format(StrParams, [IntToStr(IdxList), IntToStr(Index)]);
  if not TGMRectangle(TRectangles(Collection).FGMLinkedComponent).ExecuteScript('RectangleGetCenter', Params) then
    Exit;

  WC := TWebControl.Create(TRectangles(Collection).FGMLinkedComponent.Map.WebBrowser);
  Result := TLatLng.Create;
  try
    Result.Lat := WC.GetFloatField(RectangleForm, RectangleFormLat);
    Result.Lng := WC.GetFloatField(RectangleForm, RectangleFormLng);
  finally
    FreeAndNil(WC);
  end;
end;

procedure TRectangle.OnChangeBounds(Sender: TObject);
begin
  ChangeProperties;
  if Assigned(TGMRectangle(TRectangles(Collection).FGMLinkedComponent).FOnBoundsChanged) then
    TGMRectangle(TRectangles(Collection).FGMLinkedComponent).FOnBoundsChanged(
                  TGMRectangle(TRectangles(Collection).FGMLinkedComponent),
                  Index,
                  Self);
end;

procedure TRectangle.SetClickable(const Value: Boolean);
begin
  if FClickable = Value then Exit;

  FClickable := Value;

  ChangeProperties;
  if Assigned(TGMRectangle(TRectangles(Collection).FGMLinkedComponent).FOnClickableChange) then
    TGMRectangle(TRectangles(Collection).FGMLinkedComponent).FOnClickableChange(
                  TGMRectangle(TRectangles(Collection).FGMLinkedComponent),
                  Index,
                  Self);
end;

procedure TRectangle.SetEditable(const Value: Boolean);
begin
  if FEditable = Value then Exit;

  FEditable := Value;

  SetProperty('SetEditable', 'OnEditableChange', FEditable);
{  ChangeProperties;
  if Assigned(TGMRectangle(TRectangles(Collection).FGMLinkedComponent).FOnEditableChange) then
    TGMRectangle(TRectangles(Collection).FGMLinkedComponent).FOnEditableChange(
                  TGMRectangle(TRectangles(Collection).FGMLinkedComponent),
                  Index,
                  Self);}
end;

procedure TRectangle.SetFillColor(const Value: TColor);
begin
  if FFillColor = Value then Exit;

  FFillColor := Value;

  ChangeProperties;
  if Assigned(TGMRectangle(TRectangles(Collection).FGMLinkedComponent).FOnFillColorChange) then
    TGMRectangle(TRectangles(Collection).FGMLinkedComponent).FOnFillColorChange(
                  TGMRectangle(TRectangles(Collection).FGMLinkedComponent),
                  Index,
                  Self);
end;

procedure TRectangle.SetFillOpacity(const Value: Real);
begin
  if FFillOpacity = Value then Exit;

  FFillOpacity := Value;
  if FFillOpacity < 0 then FFillOpacity := 0;
  if FFillOpacity > 1 then FFillOpacity := 1;
  FFillOpacity := Trunc(FFillOpacity * 100) / 100;

  ChangeProperties;
  if Assigned(TGMRectangle(TRectangles(Collection).FGMLinkedComponent).FOnFillOpacityChange) then
    TGMRectangle(TRectangles(Collection).FGMLinkedComponent).FOnFillOpacityChange(
                  TGMRectangle(TRectangles(Collection).FGMLinkedComponent),
                  Index,
                  Self);
end;

procedure TRectangle.SetIdxList(const Value: Cardinal);
begin
  inherited;

end;

procedure TRectangle.SetStrokeColor(const Value: TColor);
begin
  if FStrokeColor = Value then Exit;

  FStrokeColor := Value;

  ChangeProperties;
  if Assigned(TGMRectangle(TRectangles(Collection).FGMLinkedComponent).FOnStrokeColorChange) then
    TGMRectangle(TRectangles(Collection).FGMLinkedComponent).FOnStrokeColorChange(
                  TGMRectangle(TRectangles(Collection).FGMLinkedComponent),
                  Index,
                  Self);
end;

procedure TRectangle.SetStrokeOpacity(const Value: Real);
begin
  if FStrokeOpacity = Value then Exit;

  FStrokeOpacity := Value;
  if FStrokeOpacity < 0 then FStrokeOpacity := 0;
  if FStrokeOpacity > 1 then FStrokeOpacity := 1;
  FStrokeOpacity := Trunc(FStrokeOpacity * 100) / 100;

  ChangeProperties;
  if Assigned(TGMRectangle(TRectangles(Collection).FGMLinkedComponent).FOnStrokeOpacityChange) then
    TGMRectangle(TRectangles(Collection).FGMLinkedComponent).FOnStrokeOpacityChange(
                  TGMRectangle(TRectangles(Collection).FGMLinkedComponent),
                  Index,
                  Self);
end;

procedure TRectangle.SetStrokeWeight(const Value: Integer);
begin
  if FStrokeWeight = Value then Exit;

  FStrokeWeight := Value;
  if FStrokeWeight < 1 then FStrokeWeight := 1;
  if FStrokeWeight > 10 then FStrokeWeight := 10;

  ChangeProperties;
  if Assigned(TGMRectangle(TRectangles(Collection).FGMLinkedComponent).FOnStrokeWeightChange) then
    TGMRectangle(TRectangles(Collection).FGMLinkedComponent).FOnStrokeWeightChange(
                  TGMRectangle(TRectangles(Collection).FGMLinkedComponent),
                  Index,
                  Self);
end;

procedure TRectangle.SetVisible(const Value: Boolean);
begin
  if FVisible = Value then Exit;

  FVisible := Value;

  ChangeProperties;
  if Assigned(TGMRectangle(TRectangles(Collection).FGMLinkedComponent).FOnVisibleChange) then
    TGMRectangle(TRectangles(Collection).FGMLinkedComponent).FOnVisibleChange(
                  TGMRectangle(TRectangles(Collection).FGMLinkedComponent),
                  Index,
                  Self);
end;

end.
