{
TWebControl class

  ES: Clase para la manipulacin del contenido de un TWebBrowser
  EN: Class for management the contents of a TWebBrowser

=========================================================================
History:

ver 0.1.6
  ES:
    - nuevo: Se aade mtodo SaveToJPGFile
  EN:
    - new: added SaveToJPGFile method

ver 0.1:
  ES: primera versin
  EN: first version
=========================================================================
IMPORTANTE PROGRAMADORES: Por favor, si tienes comentarios, mejoras,
  ampliaciones, errores y/o cualquier otro tipo de sugerencia, envame un correo a:
  gmlib@cadetill.com

IMPORTANT PROGRAMMERS: please, if you have comments, improvements, enlargements,
  errors and/or any another type of suggestion, please send me a mail to:
  gmlib@cadetill.com
=========================================================================

Copyright () 2011, by Xavier Martinez (cadetill)

@author Xavier Martinez (cadetill)
@web  http://www.cadetill.com
}
unit WebControl;

interface

uses
  SHDocVw, MSHTML, Classes, SysUtils;

//  http://delphi.about.com/lr/twebbrowser/127799/1/
type
  TWebControl = class
  private
    FWebBrowser: TWebBrowser;
    FForms: TStrings;
    FFields: TStrings;
    FLinks: TStrings;
  protected
    property WebBrowser: TWebBrowser read FWebBrowser write FWebBrowser;
  public
    constructor Create(WebBrowser: TWebBrowser); virtual;
    destructor Destroy; override;

    // inicializa las variables
    procedure Clear;

    // captura todos los links de la pgina
    procedure WebLinks;

    // salvar la pgina web en formato HTML
    procedure WebSaveAsHTML(const FileName: string);
    // cargar cdigo HTML directamente en el TWebBrowser
    procedure WebLoadHTML(const HTMLCode: string);
    // Aadir cdigo HTML a una pgina (al final)
    procedure WebAddHTML(const HTMLCode: WideString);

    // captura los nombres de todos los forms que tenga la pgina web
    procedure WebFormNames;
    // captura todos los campos de un determinado formulario
    procedure WebFormFields(const FormName: string); overload;
    procedure WebFormFields(const FormIdx: Integer); overload;
    // establece un valor a un campo de un determinado formulario
    procedure WebFormSetFieldValue(const FormNumber: integer;
      const FieldName, NewValue: string); overload;
    procedure WebFormSetFieldValue(const FormName, FieldName, NewValue: string); overload;
    // acepta un formulario
    procedure WebFormSubmit(const FormNumber: integer);
    // devuelve un formulario
    function WebFormGet(const FormNumber: Integer): IHTMLFormElement;
    // devuelve el valor de un campo de un formulario
    function WebFormFieldValue(const FormNumber: integer; const FieldName: string): string;
    // devuelve el valor de un campo string
    function GetStringField(const aFormName, aField: string; DefaultValue: string = ''): string;
    // devuelve el valor de un campo integer
    function GetIntegerField(const aFormName, aField: string; DefaultValue: Integer = 0): Integer;
    // devuelve el valor de un campo float
    function GetFloatField(const aFormName, aField: string; DefaultValue: Real = 0): Real;
    // devuelve el valor de un campo booleano
    function GetBoolField(const aFormName, aField: string; DefaultValue: Boolean = False): Boolean;

    // devuelve el cdigo fuente de la pgina web
    function WebHTMLCode: string;
    // devuelve el contenido de la pgina web
    function WebContent: string;

    // http://delphi.about.com/cs/adptips2004/a/bltip0504_5.htm
    // http://delphi.about.com/od/vclusing/a/wb_scren_shot.htm
    // imprime pgina web sin dilogo
    procedure WebPrintWithoutDialog;
    // imprime pgina web con dilogo
    procedure WebPrintWithDialog;
    // vista previa de impresin
    procedure WebPrintPreview;
    // llama al dilogo del setup
    procedure WebPrintPageSetup;
    // guarda el contenido del TWebBrowser en una imagen JPG
    procedure SaveToJPGFile(FileName: TFileName = '');

    property Forms: TStrings read FForms write FForms;
    property Fields: TStrings read FFields write FForms;
    property Links: TStrings read FLinks write FLinks;
  end;

implementation

uses
  ActiveX, Forms, Variants, Types, Graphics, jpeg;

{ TWebControl }

procedure TWebControl.Clear;
begin
  FFields.Clear;
  FForms.Clear;
  FLinks.Clear;
end;

constructor TWebControl.Create(WebBrowser: TWebBrowser);
begin
  FWebBrowser := WebBrowser;
  FForms := TStringList.Create;
  FFields := TStringList.Create;
  FLinks := TStringList.Create;
end;

destructor TWebControl.Destroy;
begin
  if Assigned(FLinks) then FreeAndNil(FFields);
  if Assigned(FFields) then FreeAndNil(FFields);
  if Assigned(FForms) then FreeAndNil(FForms);

  inherited;
end;

function TWebControl.GetBoolField(const aFormName, aField: string;
  DefaultValue: Boolean): Boolean;
var
  Def: Integer;
begin
  if DefaultValue then Def := 1
  else Def := 0;

  Result := GetIntegerField(aFormName, aField, Def) = 1;
end;

function TWebControl.GetFloatField(const aFormName, aField: string;
  DefaultValue: Real): Real;
var
  Strg: string;
  Val: Extended;
begin
  Result := DefaultValue;

  WebFormNames;

  if Forms.IndexOf(aFormName) = -1 then Exit;

  WebFormFields(aFormName);

  if Fields.IndexOf(aField) <> -1 then
  begin
    Strg := Trim(WebFormFieldValue(Forms.IndexOf(aFormName), aField));
    if {$IF CompilerVersion < 23}DecimalSeparator{$ELSE}FormatSettings.DecimalSeparator{$IFEND} = ',' then
      Strg := StringReplace(Strg, '.', ',', [rfReplaceAll]);
    if (Strg <> '') and TryStrToFloat(Strg, Val) then
      Result := Val;
  end;
end;

function TWebControl.GetIntegerField(const aFormName, aField: string;
  DefaultValue: Integer): Integer;
var
  Strg: string;
  Val: Integer;
begin
  Result := DefaultValue;

  WebFormNames;

  if Forms.IndexOf(aFormName) = -1 then Exit;

  WebFormFields(aFormName);

  if Fields.IndexOf(aField) <> -1 then
  begin
    Strg := Trim(WebFormFieldValue(Forms.IndexOf(aFormName), aField));
    if (Strg <> '') and TryStrToInt(Strg, Val) then
      Result := Val;
  end;
end;

function TWebControl.GetStringField(const aFormName, aField: string;
  DefaultValue: string): string;
begin
  Result := DefaultValue;

  WebFormNames;

  if Forms.IndexOf(aFormName) = -1 then Exit;

  WebFormFields(aFormName);

  if Fields.IndexOf(aField) <> -1 then
    Result := Trim(WebFormFieldValue(Forms.IndexOf(aFormName), aField));
end;

procedure TWebControl.SaveToJPGFile(FileName: TFileName);
var
  viewObject: IViewObject;
  r: TRect;
  bitmap: Graphics.TBitmap;
begin
  FWebBrowser.Document.QueryInterface(IViewObject, viewObject);
  if Assigned(viewObject) then
  try
    bitmap := Graphics.TBitmap.Create;
    try
      r := Rect(0, 0, FWebBrowser.Width, FWebBrowser.Height);

      bitmap.Height := FWebBrowser.Height;
      bitmap.Width := FWebBrowser.Width;

      viewObject.Draw(DVASPECT_CONTENT, 1, nil, nil, Application.Handle, bitmap.Canvas.Handle, @r, nil, nil, 0);

      with TJPEGImage.Create do
      try
        Assign(bitmap);
        SaveToFile(FileName);
      finally
        Free;
      end;
    finally
      bitmap.Free;
    end;
  finally
    viewObject._Release;
  end;
end;

function TWebControl.WebHTMLCode: string;
begin
  Result := IHTMLDocument2(FWebBrowser.Document).body.innerHTML;
end;

procedure TWebControl.WebFormFields(const FormName: string);
begin
  WebFormFields(Forms.IndexOf(FormName));
end;

procedure TWebControl.WebFormFields(const FormIdx: Integer);
var
  Form: IHTMLFormElement;
  Fld: IHTMLElement;
  FName: string;
  Idx: integer;
begin
  Form := WebFormGet(FormIdx);

  for idx := 0 to form.length - 1 do
  begin
    Fld := Form.Item(Idx, '') as IHTMLElement;

    if Fld = nil then Continue;
    FName := Fld.id;
    if Fld.tagName = 'INPUT' then FName := (Fld as IHTMLInputElement).name;
    if Fld.tagName = 'SELECT' then FName := (Fld as IHTMLSelectElement).name;
    if Fld.tagName = 'TEXTAREA' then FName := (Fld as IHTMLTextAreaElement).name;

    Fields.Add(FName) ;
  end;
end;

function TWebControl.WebFormFieldValue(const FormNumber: integer;
  const FieldName: string): string;
var
  Form: IHTMLFormElement;
  Field: IHTMLElement;
begin
  Form := WebFormGet(FormNumber) ;
  Field := Form.Item(FieldName,'') as IHTMLElement;
  if Field = nil then Exit;

  if Field.tagName = 'INPUT' then Result := (Field as IHTMLInputElement).value;
  if Field.tagName = 'SELECT' then Result := (Field as IHTMLSelectElement).value;
  if Field.tagName = 'TEXTAREA' then Result := (Field as IHTMLTextAreaElement).value;
end;

function TWebControl.WebFormGet(const FormNumber: Integer): IHTMLFormElement;
var
  Forms: IHTMLElementCollection;
begin
  Forms := (WebBrowser.Document as IHTMLDocument2).forms as IHTMLElementCollection;
  Result := Forms.Item(FormNumber, '') as IHTMLFormElement
end;

procedure TWebControl.WebFormNames;
var
  Frms: IHTMLElementCollection;
  Frm: IHTMLFormElement;
  Idx: integer;
begin
  if not Assigned(FWebBrowser) then Exit;

  Frms := (FWebBrowser.Document as IHTMLDocument2).forms as IHTMLElementCollection;
  for Idx := 0 to Frms.Length - 1 do
  begin
    Frm := Frms.Item(Idx,0) as IHTMLFormElement;
    Forms.Add(Frm.Name) ;
  end;
end;

procedure TWebControl.WebFormSetFieldValue(const FormNumber: integer;
  const FieldName, NewValue: string);
var
  Form: IHTMLFormElement;
  Field: IHTMLElement;
begin
  Form := WebFormGet(FormNumber);
  Field := Form.Item(FieldName, '') as IHTMLElement;
  if Field = nil then Exit;

  if Field.tagName = 'INPUT' then (Field as IHTMLInputElement).value := NewValue;
  if Field.tagName = 'SELECT' then (Field as IHTMLSelectElement).value := NewValue;
  if Field.tagName = 'TEXTAREA' then (Field as IHTMLTextAreaElement).value := NewValue;
end;

procedure TWebControl.WebFormSetFieldValue(const FormName, FieldName,
  NewValue: string);
begin
  WebFormNames;

  WebFormSetFieldValue(Forms.IndexOf(FormName), FieldName, NewValue);
end;

procedure TWebControl.WebFormSubmit(const FormNumber: integer);
var
  form: IHTMLFormElement;
begin
  form := WebFormGet(FormNumber) ;

  form.submit;
end;

procedure TWebControl.WebLinks;
var
  Lnks: IHTMLElementCollection;
  Lnk: IHTMLElement;
  Idx: integer;
begin
  if not Assigned(FWebBrowser) then Exit;

  Links.Clear;

  Lnks := (FWebBrowser.Document as IHTMLDocument2).links;
  for Idx := 0 to Lnks.Length - 1 do
  begin
    Lnk := Lnks.Item(Idx,'') as IHTMLElement;
    Links.Add(Lnk.toString) ;
  end;
end;

procedure TWebControl.WebPrintPageSetup;
var
  vIn,
  vOut: OleVariant;
begin
  FWebBrowser.ControlInterface.ExecWB(OLECMDID_PAGESETUP, OLECMDEXECOPT_PROMPTUSER, vIn, vOut) ;
end;

procedure TWebControl.WebPrintPreview;
var
  vIn,
  vOut: OleVariant;
begin
  FWebBrowser.ControlInterface.ExecWB(OLECMDID_PRINTPREVIEW, OLECMDEXECOPT_DONTPROMPTUSER, vIn, vOut) ;
end;

procedure TWebControl.WebPrintWithDialog;
var
  vIn,
  vOut: OleVariant;
begin
  FWebBrowser.ControlInterface.ExecWB(OLECMDID_PRINT, OLECMDEXECOPT_PROMPTUSER, vIn, vOut) ;
end;

procedure TWebControl.WebPrintWithoutDialog;
var
  vIn,
  vOut: OleVariant;
begin
  FWebBrowser.ControlInterface.ExecWB(OLECMDID_PRINT, OLECMDEXECOPT_DONTPROMPTUSER, vIn, vOut) ;
end;

procedure TWebControl.WebSaveAsHTML(const FileName: string);
var
  PersistStream: IPersistStreamInit;
  Stream: IStream;
  FileStream: TFileStream;
begin
  if not Assigned(FWebBrowser) then Exit;
  if not Assigned(FWebBrowser.Document) then Exit;

  PersistStream := FWebBrowser.Document as IPersistStreamInit;
  FileStream := TFileStream.Create(FileName, fmCreate) ;
  try
    Stream := TStreamAdapter.Create(FileStream, soReference) as IStream;
    if Failed(PersistStream.Save(Stream, True)) then
      raise Exception.Create('SaveAs HTML fail!');
  finally
    FileStream.Free;
  end;
end;

function TWebControl.WebContent: string;
begin
  Result := (FWebBrowser.Document as IHTMLDocument2).body.innerText;
end;

procedure TWebControl.WebLoadHTML(const HTMLCode: string);
var
  sl: TStringList;
  ms: TMemoryStream;
begin
  if not Assigned(FWebBrowser) then Exit;

  WebBrowser.Navigate('about:blank') ;
  while WebBrowser.ReadyState < READYSTATE_INTERACTIVE do
    Application.ProcessMessages;

  if Assigned(WebBrowser.Document) then
  begin
    sl := TStringList.Create;
    ms := TMemoryStream.Create;
    try
      sl.Text := HTMLCode;
      sl.SaveToStream(ms);
      ms.Seek(0, 0);
      (WebBrowser.Document as IPersistStreamInit).Load(TStreamAdapter.Create(ms));
    finally
      if Assigned(ms) then FreeAndNil(ms);
      if Assigned(sl) then FreeAndNil(sl);
    end;
  end;
end;

procedure TWebControl.WebAddHTML(const HTMLCode: WideString);
var
  Range: IHTMLTxtRange;
begin
  if not Assigned(FWebBrowser) then Exit;

  Range := IHTMLBodyElement(IHTMLDocument2(FWebBrowser.Document).body).createTextRange;
  Range.Collapse(False) ;
  Range.PasteHTML(HTMLCode) ;
end;

end.
